#!/usr/bin/env python3
"""
This example illustrates how to interpolate 1-d distributions
nonparametrically as a function of multiple predictors
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 10 2019"

# Perform necessary imports
import npstat as ns
import matplotlib.pyplot as plt

# Interpolation grids used to connect parameter values with
# distributions. We will only interpolate four distributions
# assuming that they are at the corners of a unit square.
paramGridX = ns.GridAxis(0.0, 1.0)
paramGridY = paramGridX

# Create the object which will perform density interpolation.
# We can interpolate vertically or we can interpolate the quantile
# function (horizontally). Change the "interpolateVertically"
# setting below and see how the results change.
interpolateVertically = False
interpolator = ns.InterpolatedDistro1DNP((paramGridX, paramGridY),
                                         interpolateVertically)

# Assign distributions to each cell of the interpolation grid.
# In this example, the first parameter is associated with the
# change in the mean of the Gaussian distribution, and the second
# with the change in the width.
cells      = ((0, 0),      (0, 1),      (1, 0),     (1, 1))
muAndSigma = ((-1.0, 0.4), (-1.0, 0.8), (1.0, 0.4), (1.0, 0.8))
for cell, musigma in zip(cells, muAndSigma):
    interpolator.setGridDistro(cell, ns.Gauss1D(*musigma))

# Cycle over some parameter values and plot interpolated distributions
xparams = ns.EquidistantInLinearSpace(0.0, 1.0, 3)
yparams = xparams

fig = plt.figure()
for ix, paramx in enumerate(xparams):
    for iy, paramy in enumerate(yparams):
        interpolator.setGridCoords((paramx, paramy))
        x, y = ns.scanDensity1D(interpolator, -3.5, 3.5, 350)
        ax = fig.add_subplot(3, 3, (2-iy)*3 + ix + 1)
        ax.plot(x, y)
        ax.set_ylim(-0.2, 1.2)
        ax.set_xlabel('X')
        ax.set_ylabel('Density')
        ax.set_title("p = ({:.1f}, {:.1f})".format(paramx, paramy))

# Display the window
fig.suptitle("Interpolating Location (Horizontal) and Scale (Vertical)")
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
