#!/usr/bin/env python3
"""
This example illustrates how to interpolate multivariate distributions
nonparametrically as a function of multiple predictors (a.k.a. multivariate
template morphing)
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 10 2019"

# Perform necessary imports
import npstat as ns
import matplotlib.pyplot as plt
from npstat_utils import BoxND, rotationMatrix2D, \
    discretizedCompositeND, discretizedDistributionND
from npstat_plotting import colormeshDensity2D
import numpy as np

# Number of cells to use for discretizations
nDiscrete = 200

# Are we going to interpolate by performing copula decomposition?
interpolateCopulas = False

# Interpolation grids used to connect parameter values with
# distributions. We will only interpolate four distributions
# assuming that they are at the corners of a unit square.
paramGridX = ns.GridAxis(0.0, 1.0)
paramGridY = paramGridX

# Create the object which will perform density interpolation.
# We can interpolate multivariate distributions by decomposing
# the distribution into the product of the copula and the
# marginals. In this case the copula is interpolated vertically
# and the marginals are interpolated horizontally. Set the
# parameter "interpolateCopulas" to True if you want this mode
# of interpolation. Alternatively, we can interpolate the
# inverse Rosenblatt transform (i.e., interpolate conditional
# quantile functions). Note that not all distributions support
# these interpolation modes: distributions whose copula can be
# interpolated must be an instance of "CompositeDistributionND"
# class while the distributions permitting inverse Rosenblatt
# transform return "true" when their method "mappedByQuantiles"
# is called. In general, copula interpolation works faster while
# the inverse Rosenblatt transform will probably result in
# interpolations which appear more intuitive.
interpolator = ns.GridInterpolatedDistribution((paramGridX, paramGridY),
                                               2, interpolateCopulas)

# Gaussian distribution shifted and scaled separately in each direction
g0 = ns.ScalableGaussND((0.0, 0.0), (0.4, 1.0))

# Assign distributions to each cell of the interpolation grid.
# The first parameter will change the location of the Gaussian
# distribution and the second will change the rotation angle
cells         = ((0, 0),     (0, 1),      (1, 0),     (1, 1))
shiftAndAngle = ((-1.0, 45), (-1.0, -45), (1.0, 45),  (1.0, -45))
for cell, sa in zip(cells, shiftAndAngle):
    # Construct rotated and shifted distributions
    rot = rotationMatrix2D(sa[1], True)
    gauss = ns.LinTransformedDistroND(g0, (sa[0], 0.0), rot)
    # Discretize these distributions. This allows for subsequent numerical
    # estimation of the copula and of the inverse Rosenblatt transform.
    box = BoxND((sa[0]-3, sa[0]+3), (-3, 3))
    if interpolateCopulas:
        discr = discretizedCompositeND(gauss, box, (nDiscrete, nDiscrete))
    else:
        # The following works slightly faster for inverse Rosenblatt.
        discr = discretizedDistributionND(gauss, box, (nDiscrete, nDiscrete))
    interpolator.setGridDistro(cell, discr)

# Cycle over some parameter values and plot interpolated distributions
xparams = ns.EquidistantInLinearSpace(0.0, 1.0, 3)
yparams = xparams

fig = plt.figure(figsize=(6, 6))
for ix, paramx in enumerate(xparams):
    for iy, paramy in enumerate(yparams):
        interpolator.setGridCoords((paramx, paramy))
        ax = fig.add_subplot(3, 3, (2-iy)*3 + ix + 1)
        colormeshDensity2D(ax, interpolator, BoxND((-3, 3), (-3, 3)),
                           (nDiscrete, nDiscrete))
        ax.set_title("p = ({:.1f}, {:.1f})".format(paramx, paramy))

# Display the window
if interpolateCopulas:
    fig.suptitle("Density Morphing with Vertical Copula Interpolation")
else:
    fig.suptitle("Density Morphing by Inverse Rosenblatt Transform")
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
