#!/usr/bin/env python3
"""
This example illustrates how to perform univariate expectation-maximization
unfolding (a.k.a. D'Agostini unfolding) regularized by smoothing instead of
early stopping of iterations. The mathematical details of the method can be
found in arXiv:1408.6500.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 14 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import BoxND, HistoND, Matrix
from npstat_plotting import colormeshNumpyArray, \
    uncertaintyBandPolygon, plotHistoOutline1D
import matplotlib.pyplot as plt
import sys, math
import numpy as np

# Some parameters for the script
sampleSize = 1000    # Average number of points to generate (sample size)
xmin = -7.0          # Left edge of the sample histogram
xmax = 7.0           # Right edge of the sample histogram
numBins = 70         # Number of bins for the sample histogram. Should be
                     # much larger than the number of effective degrees of
                     # freedom of the model (this can be checked later).
nDiscrete = 350      # Number of points to use for density discretization
resolution = 1.0     # Response matrix resolution (homoscedastic here)
symbetaPower = -1    # Power parameter of the symmetric beta weight function
                     # w.r.t. which the orthogonal polynomial system will
                     # be generated. Use -1 to employ a Gaussian instead.
filterDegree = 0     # Polynomial degree for local expansion. 0 is safe to
                     # use because smoothing matrices built in this manner
                     # will not have eigenvalues larger than 1. This is
                     # required to guarantee convergence of expectation-
                     # maximization iterations for arbitrary input data.
                     # For your data, higher degrees might actually work
                     # better if unfolding iterations do converge.
bw0 = resolution/4.0 # Initial guess for the smoothing bandwidth
useTraceNDoF = 0     # Using entropic or trace-based effective degrees
                     # of freedom for AIC (use 0 or 1 here, not boolean).

# Density to unfold
mix = ns.DistributionMix1D()
mix.add(ns.TruncatedDistribution1D(ns.Gauss1D(-2,1), xmin, xmax), 0.2)
mix.add(ns.TruncatedDistribution1D(ns.Gauss1D(2,1), xmin, xmax), 0.5)
mix.add(ns.Uniform1D(xmin, xmax-xmin), 0.3)

# Generate a random sample according to the Poisson intensity
# determined by the mean number of points and the density to unfold
rng = ns.MersenneTwister()
nPoints = ns.Poisson1D(sampleSize).random(rng)
sample = mix.generate(rng, nPoints)

# Smear this sample (using just a homoscedastic uncertainty)
smearedSample = [ns.Gauss1D(x, resolution).random(rng) for x in sample]

# Histogram the smeared sample. To work easily with the
# unfolding code, the bins should be of type "double".
h1 = HistoND("double", ns.HistoAxis(numBins, xmin, xmax, "Y"),
             "Observed Space", "Counts")
h1.fill(smearedSample, 1)

# Construct the response matrix. Here, we use a lot more bins in the
# unfolded space than in the smeared space. This avoids uncontrolled
# regularization by binning. While in this example the response matrix
# construction employs NPStat function "gaussianResponseMatrix" (which
# is nearly exact), in practice this matrix will have to be built by
# density estimation.
responseMatrix = ns.gaussianResponseMatrix(
    xmin, xmax, nDiscrete, xmin, xmax, numBins,
    ns.DoubleSame(), ns.DoubleConstValue1(resolution))

# Some objects we will need down the road
bh = ns.BoundaryHandling("BM_FOLD")
binwidth = (xmax - xmin)/nDiscrete
bins = ns.arrayNDToNumpy(h1.binContents(), False)

# The object which will perform unfolding. It needs a filter object as
# a constructor argument. We are going to use a dummy one for a moment.
unf = ns.SmoothedEMUnfold1D(responseMatrix,
                            ns.DummyLocalPolyFilter1D(nDiscrete), True)

# Choose the unfolding bandwidth by EAIC. Increase the range of
# scanned bandwidth values or adjust bw0 if it turns out that
# the best bandwidth is on the boundary.
bwvaues = ns.EquidistantInLogSpace(bw0/2.0, bw0*2.0, 50)
bestAIC = sys.float_info.max
bestBW = None
bestFilter = None
bestNDoF = None
for bw in bwvaues:
    # Create the "real" smoothing filter and pass it to the
    # unfolding object.
    filter = ns.symbetaLOrPEFilter1D(symbetaPower, bw, filterDegree,
                                     nDiscrete, binwidth, bh)
    unf.setFilter(filter)
    # In the next call, do not fill the covariance matrix
    # of unfolded results. This works somewhat faster.
    unfolded = unf.unfoldNoCov(bins)
    folded = unf.fold(unfolded)
    # The following assumes Poisson uncertainties in the
    # folded space. A different calculation would be needed
    # in case the uncertainties are multinomial or if they
    # are specified by a covariance matrix.
    logli = ns.poissonLogLikelihood(folded, bins)
    nDoF = unf.modelNDoF()[useTraceNDoF]
    aic = ns.aicc(nDoF, logli, numBins);
    print("Bandwidth is {:.4f}, the AIC is {:.3f}".format(bw, aic))
    if aic < bestAIC:
        bestAIC, bestBW, bestFilter, bestNDoF = aic, bw, filter, nDoF
print("Best AIC bandwidth is {:.4f}".format(bestBW))
print("Optimal number of degrees of freedom is {:.2f}".format(bestNDoF))

# Unfold with the optimal filter, filling the result covariance matrix
# this time
covmat = Matrix("double")
unf.setFilter(bestFilter)
unfolded = unf.unfold(bins, covmat)

# We are done with unfolding. The rest of this script is just
# plotting results. Plot the results in the folded space first.
fig = plt.figure(figsize=(8, 4))
ax = fig.add_subplot(121)
plotHistoOutline1D(ax, h1, label='Input Histogram')
xbins = h1.axis(0).binCenters()
ax.plot(xbins, unf.fold(unfolded), color='r', label='Reconstructed')
ax.legend(loc=2)

# Plot unfolded results
ax = fig.add_subplot(122)
xbins = ns.HistoAxis(nDiscrete, xmin, xmax).binCenters()
yerrs = [math.sqrt(x) for x in covmat.mainDiagonal()]
band = uncertaintyBandPolygon(xbins, unfolded-yerrs, unfolded+yerrs)
ax.fill(*band, facecolor='skyblue', label='Uncertainty band')
x, y = ns.scanDensity1D(mix, xmin, xmax, nDiscrete+1)
ax.plot(x, y*sampleSize*binwidth, color='k', label='True intensity')
ax.set_xlabel('X')
ax.set_ylabel('Poisson Intensity')
ax.set_title('Unfolded Space')
ax.legend(loc=2)

# Display the figure
fig.suptitle("Expectation-Maximization Unfolding with Smoothing")
fig.tight_layout(rect=[0, 0, 1, 0.95])
plt.show()
