#!/usr/bin/env python3
"""
This example illustrates how to perform constant-bandwidth univariate
kernel density estimation (KDE) efficiently. The code discrerizes the
data and then utilizes the fast Fourier transform (FFT) for convolutions.
Note that this type of density estimation can suffer from boundary bias.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.1"
__date__ ="Sep 20 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import FFTKDE1D, histoPluginBandwidth1D, \
                                      BinnedLSCVCalc1D, BinnedRLCVCalc1D
import matplotlib.pyplot as plt
from npstat_plotting import plotHistoOutline1D

# Some parameters for the script
sampleSize = 500     # Number of points to generate (sample size)
xmin = -3.5          # Left edge of the sample histogram
xmax = 3.5           # Right edge of the sample histogram
numBins = 70         # Number of bins for the sample histogram
nDiscrete = 1024     # Number of cells for sample discretization.
                     # Should be a power of 2 (this results in the
                     # most efficient application of FFT).
symbetaPower = -1    # Power parameter of the symmetric beta kernel.
                     # Specify -1 to use a Gaussian kernel instead.
filterDegree = 2     # Polynomial degree used to generate high-order kernels.
                     # In general, even numbers make the most sense here.
                     # Kernel order will be filterDegree + 2.
allowNegative = True # Specify whether the density estimate is allowed
                     # to be negative somewhere (this is possible in case
                     # filterDegree > 0)

# We will use a 1-d Gassian mixture to generate the random sample.
# See the "brute_force_kde_1d.py" script for more comments on this.
component1 = ns.GaussianMixtureEntry(3.0, -1.0, 0.5)
component2 = ns.GaussianMixtureEntry(7.0,  1.0, 0.5)
mixture = ns.GaussianMixture1D(0.0, 1.0, (component1, component2))

# Generate a random sample according to this density
rng = ns.MersenneTwister()
sample = mixture.generate(rng, sampleSize)

# Histogram for visualizing the sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Histogram for sample discretization. Note that the axis range is
# increased, to catch all possible tails. In general, such an increase
# may or may not make sense, depending on situation.
h2 = HistoND("int", ns.HistoAxis(nDiscrete, 1.5*xmin, 1.5*xmax, "X"),
             "KDE Discretization Histogram", "Counts")
h2.fill(sample, 1)

# Object which will perform the KDE
smoother = FFTKDE1D(h2, allowNegative)

# Make an initial guess for the KDE bandwidth. This guess will use
# a robust measure of the sample scale, estimated from the histogram.
bw0 = histoPluginBandwidth1D(h2, sampleSize, filterDegree, symbetaPower)
print("Plug-in bandwidth is {:.4f}".format(bw0))
yPlugin = smoother.estimateDensity(h2, (filterDegree, bw0, symbetaPower))

# Add a number of kernels with different bandwidth values
# for cross-validation calculations. In general, filterDegree
# parameter can be simultaneously optimized as well.
for bw in ns.EquidistantInLogSpace(bw0/4.0, bw0*2.0, 100):
    smoother.addFilter((filterDegree, bw, symbetaPower))

# Run the least squares cross-validation (LSCV)
lscvCalc = BinnedLSCVCalc1D()
bestLSCV, paramsLSCV, yLSCV = smoother.bestFilterByCV(h2, lscvCalc)
lscvFilterDegree, lscvBandwidth, lscvSymbetaPower = paramsLSCV
print("LSCV bandwidth is {:.4f}".format(lscvBandwidth))

# Run the regularized likelihood cross-validation (RLCV).
# For the description of this method, see section 4.2 in
# https://doi.org/10.1080/10485252.2017.1371715
regularisationParameter = 0.5
rlcvCalc = BinnedRLCVCalc1D(regularisationParameter)
bestRLCV, paramsRLCV, yRLCV = smoother.bestFilterByCV(h2, rlcvCalc)
rlcvFilterDegree, rlcvBandwidth, rlcvSymbetaPower = paramsRLCV
print("RLCV bandwidth is {:.4f}".format(rlcvBandwidth))

# We are done with density estimation at this point. The rest of the script
# is just plotting results. To begin with, create a Matplotlib plot window.
fig = plt.figure()
ax = fig.add_subplot(111)

# Plot the sample histogram
plotHistoOutline1D(ax, h1, ylabel='Density estimate',
                   title='KDE by FFT', ymin=-3)

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
verticalScale = h1.binVolume()*sampleSize
xdens, ydens = ns.scanDensity1D(mixture, xmin, xmax, nDiscrete+1)
ax.plot(xdens, ydens*verticalScale, 'g', label='True density')

# Overlay the KDE results with various bandwidth values
xkde = h2.axis(0).binCenters()
colors     = ('c',       'r',      'b'   )
labels     = ('Plug-in', 'RLCV',   'LSCV')
results    = (yPlugin,   yRLCV,    yLSCV )
for color, l, y in zip(colors, labels, results):
    ax.plot(xkde, y*verticalScale, color, label="{} estimate".format(l))

# Add plot legend
plt.legend(loc=2)

# Display the window
plt.show()
