#!/usr/bin/env python3
"""
This example illustrates how to perform local polynomial least squares
regression in 1-d with NPStat
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Oct 08 2019"

# Perform necessary imports
import npstat as ns
from npstat_local_regression import LocalQuadraticLeastSquaresND
import matplotlib.pyplot as plt

# Some parameters for the script
sampleSize = 50        # Number of points to generate (sample size)
regressionDegree = 1   # Degree of local polynomial used in regression.
                       # Can be 0, 1, or 2.
coeffs = (0, -2, 0, 2) # Coefficients for the orthonormal Legendre polynomial
                       # series used to represent the true regression mean
nDiscrete = 1000       # Number of points for various discretizations

# We will model the true mean of the regression curve
# with orthonormal Legendre polynomials on [-1, 1].
class TrueMean:
    def __init__(self, coeffs):
        self.coeffs = coeffs
    def __call__(self, x):
        poly = ns.LegendreOrthoPoly1D()
        return poly.series(self.coeffs, x)

# Model for the standard deviation. It will vary linearly with x
# from smin to smax in the x range [-1, 1].
class TrueStdev:
    def __init__(self, smin, smax):
        self._map = ns.LinearMapper1d(-1.0, smin, 1.0, smax)
    def __call__(self, x):
        return self._map(x)

# Generate a sample of x coordinates flat on [-1, 1]
rng = ns.MersenneTwister()
uniform = ns.Uniform1D(-1.0, 2.0)
xcoords = uniform.generate(rng, sampleSize)

# We will need to use the error definitions later (in this example, the
# sample is heteroscedastic), so we will make a separate list of them.
stdev = TrueStdev(1.5, 0.5)
yerrs = [stdev(x) for x in xcoords]

# Generate y values for the regression sample
mean = TrueMean(coeffs)
yvalues = [ns.Gauss1D(mean(x), sig).random(rng) \
           for x, sig in zip(xcoords, yerrs)]

# The local regression object. In this example, we will use
# the Gaussian kernel to localize the regression region.
localReg = LocalQuadraticLeastSquaresND(regressionDegree, ns.Gauss1D(0, 1),
                                        xcoords, yvalues, yerrs)

# Try a number of different bandwidth settings and select the best one
# by several bandwidth selection methods. Note that the CPU time consumed
# by AIC-based bandwidth selectors scales, roughly, in proportion to
# the sample size cubed, so these selectors can not be used with large
# samples. The CPU time needed for cross-validation scales, roughly,
# in proportion to the sample size squared.
bwSet = ns.EquidistantInLogSpace(0.05, 0.5, 100)
xdense = ns.EquidistantInLinearSpace(-1.0, 1.0, nDiscrete)

# Best least squares cross-validation bandwidth
bestBwLSCV = localReg.selectLSCVBandwidth(bwSet)
lscvCurve = [localReg.fit((x,)) for x in xdense]
print("LSCV bandwidth is {:.5f}".format(*bestBwLSCV))

# Best "entropic AIC" cross-validation bandwidth
bestBwEAIC = localReg.selectEAICBandwidth(bwSet)
eaicCurve = [localReg.fit((x,)) for x in xdense]
print("EAIC bandwidth is {:.5f}".format(*bestBwEAIC))

# Best "trace AIC" cross-validation bandwidth
bestBwTAIC = localReg.selectTAICBandwidth(bwSet)
taicCurve = [localReg.fit((x,)) for x in xdense]
print("TAIC bandwidth is {:.5f}".format(*bestBwTAIC))

# Curve for displaying the true mean
ydense = [mean(x) for x in xdense]

# Plot various things
plt.errorbar(xcoords, yvalues, yerr=yerrs, color='k', ecolor='k',
             marker='o', ls="", capsize=3)
plt.plot(xdense, ydense, 'g', label="True mean")
plt.plot(xdense, lscvCurve, 'r', label="LSCV curve")
plt.plot(xdense, eaicCurve, 'b', label="EAIC curve")
plt.plot(xdense, taicCurve, 'm', label="TAIC curve")

# Display the window
plt.xlabel('X')
plt.ylabel('Y')
plt.legend(loc=1)
if regressionDegree == 0:
    plt.title('Local Average Regression (Nadaraya–Watson)')
elif regressionDegree == 1:
    plt.title('Local Linear Regression')
elif regressionDegree == 2:
    plt.title('Local Quadratic Regression')
plt.show()
