#!/usr/bin/env python3
"""
This example illustrates how to perform univariate density estimation
utilizing local orthogonal polynomial expansion (LOrPE). For details of
the method, see the article by D.P. Amali Dassanayake, Igor Volobouev, and
A. Alexandre Trindade, "Local orthogonal polynomial expansion for density
estimation", Journal of Nonparametric Statistics, 29:4, 806-830 (2017).
https://doi.org/10.1080/10485252.2017.1371715 (also arXiv:1505.00275).
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.1"
__date__ ="Sep 20 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
from npstat_density_estimation import FFTKDE1D, histoPluginBandwidth1D
import matplotlib.pyplot as plt

# Some parameters for the script
sampleSize = 1000   # Number of points to generate (sample size)
xmin = 0.0          # Left edge of the sample histogram
xmax = 5.0          # Right edge of the sample histogram
numBins = 100       # Number of bins for the sample histogram
nDiscrete = 1000    # Number of cells for sample discretization
symbetaPower = -1   # Power parameter of the symmetric beta weight function
                    # w.r.t. which the orthogonal polynomial system will
                    # be generated. Use -1 to employ a Gaussian instead.
filterDegree = 2    # Polynomial degree for the local expansion. This
                    # degree is just a guess here. It can also be
                    # optimized by cross-validation, as illustrated
                    # by the "lorpe_1d_cv.py" script.
bwFactor = 1.0      # LOrPE bandwidth will be set to the product of the
                    # plug-in bandwidth and this factor

# Here, we will use the exponential distribution truncated to [xmin, xmax]
# in order to generate the random sample
d1 = ns.TruncatedDistribution1D(ns.Exponential1D(0.0, 1.0), xmin, xmax)

# Generate a random sample according to this density
rng = ns.MersenneTwister()
sample = d1.generate(rng, sampleSize)

# Histogram for displaying the sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Sample discretization for LOrPE. For smoothing purposes, the bin width
# should be substantially smaller than any physical scale of interest.
h2 = HistoND("int", ns.HistoAxis(nDiscrete, xmin, xmax, "X"),
             "LOrPE Discretization Histogram", "Counts")
h2.fill(sample, 1)

# LOrPE boundary handling method. Simple methods (which require no additional
# tuning parameters) are:
#
#  "BM_TRUNCATE" -- Just truncate the weight function at the boundary,
#                   as described in the article by Dassanayake et. al.
#  "BM_STRETCH"  -- Stretch the weight function by a factor that smoothly
#                   changes from 1 to 2 as the boundary is approached.
#  "BM_FOLD"     -- Reflect (fold) the weight function at the boundary and
#                   add to the non-reflected part.
#  "BM_CONSTSQ"  -- Stretch the weight function near the boundary in such
#                   a way that the integral of the weight function squared
#                   inside the bounded interval is preserved.
#  "BM_FOLDSQ"   -- Both fold the weight function at the boundary and
#                   preserve the integral of the weight function squared.
#  "BM_CONSTVAR" -- Stretch the weight function near the boundary so that
#                   the variance of a variable distributed acording to
#                   this weight function remains constant.
#  "BM_FOLDVAR"  -- Both fold the weight function at the boundary and
#                   stretch so that the variance remains constant.
#
# It is not obvious a priori which boundary handling method will perform
# best for any particular case, so it might be useful to try several
# different ones.
bh = ns.BoundaryHandling("BM_FOLD")

# Perform density estimation with LOrPE using a convenient high-level
# driver function "lorpeSmooth1D". A more general way of using LOrPE
# is illustrated in the example script "lorpe_1d_cv.py".
ylorpe, bw = ns.lorpeSmooth1D(h2, symbetaPower, filterDegree, bh, bwFactor)
print("LOrPE bandwidth is {:.4f}".format(bw))

# Add vanilla KDE for comparison (no mirroring and no high order kernels)
kde = FFTKDE1D(h2, False, False)
bwkde = bwFactor*histoPluginBandwidth1D(h2, sampleSize, 0, symbetaPower)
yKDE = kde.estimateDensity(h2, (0, bwkde, symbetaPower))
print("KDE bandwidth is {:.4f}".format(bwkde))

# Plot the sample histogram
xh, yh = ns.histoOutline1D(h1)
plt.plot(xh, yh, 'k')

# Overlay the original density. Scale all densities so that their
# plots are compatible with the histogram height.
area = h1.integral()
xdens, ydens = ns.scanDensity1D(d1, xmin, xmax, nDiscrete+1)
plt.plot(xdens, ydens*area, 'g', label='True density')

# Overlay the KDE result
coords = h2.axis(0).binCenters()
plt.plot(coords, yKDE*area, 'b', label='KDE')

# Overlay the LOrPE result
plt.plot(coords, ylorpe*area, 'r', label='LOrPE')

# Add labels and comments
plt.xlabel('X')
plt.ylabel('Density estimate')
plt.title('Density Estimation by KDE and LOrPE')
plt.legend()

# Display the window
plt.show()
