#!/usr/bin/env python3
"""
This example illustrates how to perform LOrPE on unbinned samples.
The bandwidth is chosen by the plug-in method.
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="May 31 2022"

# Perform necessary imports
import npstat as ns
import numpy as np
from npstat_utils import HistoND
import matplotlib.pyplot as plt

# Some parameters for the script
sampleSize = 500    # Number of points to generate (sample size)
xmin = 0.0          # Left edge of the sample histogram
xmax = 5.0          # Right edge of the sample histogram
numBins = 50        # Number of bins for the sample histogram
symbetaPower = 4    # Power parameter of the symmetric beta weight function
                    # w.r.t. which the orthogonal polynomial system will
                    # be generated. Use -1 to employ a Gaussian instead.
filterDegree = 1.5  # Polynomial degree for the local expansion. This
                    # degree is just a guess here. Must be non-negative
                    # and does not have to be an integer. Polynomial degree
                    # of 0 corresponds to KDE with boundary kernels.
nDiscrete = 1000    # Number of intervals for displaying the results.
normalize = True    # Do we want to normalize the density estimate?
                    # Here, this only ensures that the integral is 1
                    # and does not check for negative densities.
nIntegSplits = 50   # Number of intervals to use for various quadratures.
nIntegPoints = 4    # Number of Gauss-Legendre integration points to use
                    # on each interval for various quadratures.

# Choose boundary handling method. Only two methods are currently
# supported for unbinned LOrPE:
#
#  "BM_TRUNCATE" -- Just truncate the weight function at the boundary.
#
#  "BM_FOLD"     -- Reflect (fold) the weight function at the boundary and
#                   add to the non-reflected part.
#
# It is not obvious a priori which boundary handling method will perform
# best for any particular case, so it might be useful to try different ones.
bh = ns.BoundaryHandling("BM_TRUNCATE")

# Here, we will use the exponential distribution truncated to [xmin, xmax]
# in order to generate the random sample
d1 = ns.TruncatedDistribution1D(ns.Exponential1D(0.0, 1.0), xmin, xmax)

# Generate a random sample according to this density
rng = ns.MersenneTwister()
sample = d1.generate(rng, sampleSize)

# Histogram for displaying the sample
h1 = HistoND("int", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Counts")
h1.fill(sample, 1)

# Create the object that will be used to estimate the density
lorpeKernel = ns.LOrPE1DSymbetaKernel(symbetaPower, filterDegree, xmin, xmax, bh)
lorpeDensityEstimator = ns.DoubleLOrPE1D(lorpeKernel, sample)

# Select a reasonable bandwidth
acc = ns.DoubleSampleAccumulator()
acc.accumulate(sample)
robustScale = acc.sigmaRange()
bw = ns.approxAmisePluginBwGauss(filterDegree, sampleSize, robustScale)
if symbetaPower >= 0:
    bw *= ns.approxSymbetaBandwidthRatio(symbetaPower, filterDegree)

# Normalize the estimate
if normalize:
    integral = lorpeDensityEstimator.normalizeDensityEstimate(
        nIntegSplits, nIntegPoints, bw)
    print("Density estimate integral before normalization is", integral)

# Density estimate by LOrPE
coords, yunbinned = ns.scanDensityEstimate1D(
    lorpeDensityEstimator, xmin, xmax, nDiscrete+1, bw)

# Determine the ISE
ise = lorpeDensityEstimator.integratedSquaredError(
    d1, nIntegSplits, nIntegPoints, bw)

print("LOrPE filter degree is {:.1f}, bandwidth is {:.4f}, ISE = {:.4g}".format(
    filterDegree, bw, ise))

# Plot the sample histogram
xh, yh = ns.histoOutline1D(h1)
plt.plot(xh, yh, 'k')

# Overlay the original density. Scale all densities so that
# their plots are compatible with the histogram height.
area = h1.integral()
xdens, ydens = ns.scanDensity1D(d1, xmin, xmax, nDiscrete+1)
plt.plot(xdens, ydens*area, 'g', label='True density')

# Overlay the unbinned LOrPE result
plt.plot(coords, yunbinned*area, 'r', label='LOrPE')

# Add labels and comments
plt.xlabel('X')
plt.ylabel('Density estimate')
plt.title('Unbinned Density Estimation by LOrPE')
plt.legend()

# Display the window
plt.show()
