#!/usr/bin/env python3
"""
This example illustrates how to reconstruct beckground density in
a semiparametric mixture using unbinned LOrPE
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="Aug 16 2022"

# Perform necessary imports
import math, time
import npstat as ns
import numpy as np
from npstat_utils import HistoND
from npstat_density_estimation import pluginBandwidthScalingTable
import matplotlib.pyplot as plt

# Some parameters for the script
sampleSize = 500    # Number of points to generate (sample size)
xmin = 0.0          # Left edge of the sample histogram
xmax = 5.0          # Right edge of the sample histogram
numBins = 50        # Number of bins for the sample histogram
symbetaPower = 4    # Power parameter of the symmetric beta weight function
                    # w.r.t. which the orthogonal polynomial system will
                    # be generated. Use -1 to employ a Gaussian instead.
filterDegree = 0    # The first guess for the best degree of LOrPE filter.
normalize = True    # Do we want to normalize the density estimate?
                    # Here, this only ensures that the integral is 1
                    # and does not check for negative densities.
nIntegSplits = 50   # Number of intervals to use for various quadratures.
nIntegPoints = 4    # Number of Gauss-Legendre integration points to use
                    # on each interval for various quadratures.
alpha = 0.3         # Signal fraction in the mix.
convTol = 1.0e-7    # Tolerance parameter for background estimate convergence.
convDensity = True  # Set this parameter to "True" if the convergence
                    # should be checked for background density, to "False"
                    # if the convergence should be checked for weights.
maxIter = 1000      # Maximum number of iterations for the Fredholm equation.
rngSeed = 19394489  # Seed for the random number generator. Set to 0
                    # in order to have different random samples every time.

# Choose the cross-validation method. We will do the following here. Only
# non-negative values of regularization parameter "rlcvAlpha" make sense
# for RLCV. So, if the parameter below is negative, we use LSCV, otherwise
# we use RLCV. Try different values of rlcvAlpha and see what happens.
rlcvAlpha = -1.0

# Choose boundary handling method. Only two methods are currently
# supported for unbinned LOrPE:
#
#  "BM_TRUNCATE" -- Just truncate the weight function at the boundary.
#
#  "BM_FOLD"     -- Reflect (fold) the weight function at the boundary and
#                   add to the non-reflected part.
#
# It is not obvious a priori which boundary handling method will perform
# best for any particular case, so it might be useful to try different ones.
bh = ns.BoundaryHandling("BM_TRUNCATE")

# Generator of random numbers
if rngSeed == 0:
    rng = ns.MersenneTwister()
else:
    rng = ns.MersenneTwister(rngSeed)

# Signal distribution
signal = ns.TruncatedDistribution1D(ns.Gauss1D(2.0, 0.3), xmin, xmax)

# Background distribution. Note wider width -- it is
# here to avoid negative background density estimates.
bg = ns.TruncatedDistribution1D(ns.Exponential1D(0.0, 1.5), xmin, xmax)

# The mixture
mix = ns.DistributionMix1D()
mix.add(signal, alpha)
mix.add(bg, 1.0 - alpha)

# Generate a sample of points on [xmin, xmax].
# Sort it so that these points can be used as
# coordinates for calculating weights.
unsortedCoords = mix.generate(rng, sampleSize)
sampleCoords = np.sort(unsortedCoords)

# Here, we will simply initialize the weights to 1 for every point.
unitWeights = np.ones(sampleSize, dtype=np.double)
weightedSample = list(zip(sampleCoords, unitWeights))

# Histogram for displaying the weighted sample. Note that
# the type of bins is double rather than integer.
h1 = HistoND("double", ns.HistoAxis(numBins, xmin, xmax, "X"),
             "Sample Histogram", "Weight")
h1.fill(weightedSample)

# The functor to use for cross-validation
if rlcvAlpha < 0.0:
    useRLCV = False
    lorpeFunctor = ns.LOrPE1DLSCVFunctor
else:
    useRLCV = True
    lorpeFunctor = ns.LOrPE1DRLCVFunctor

# Create the object which performs cross-validation. The function
# signature looks as follows (see also C++ header file LOrPE1DCV.hh):
#
#      LOrPE1DLSCVFunctor(symbetaPower,
#                         leftBoundary,
#                         rightBoundary,
#                         bh,
#                         bandwidthFunctor,
#                         localizingWeight,
#                         localizingWeightXmin,
#                         localizingWeightXmax,
#                         nIntegIntervals,
#                         nIntegPoints,
#                         normalizeLOrPEEstimate,
#                         sample)
#
# The meaning of the arguments is described in the script
# "lorpe_1d_features.py".

# The example callable to use for bandwidth dependence on the coordinate.
# Polynomial series are used here, in the monomial basis. The monomial
# coefficients are arranged in the order of increasing degree. This works
# faster than a functor defined by a python callable.
polyCoeffs = (1.0, 0.0, 1.0)
polyXShift = 2.0
bandwidthPoly = ns.Poly1DShifted(ns.Poly1D(polyCoeffs), polyXShift)
bandwidthFunctor = ns.Functor1Ref(bandwidthPoly)

# You can also try to uncomment the following in order
# to have no bandwidth dependence on the coordinate:
#
bandwidthFunctor = 1.0

# Example distribution used to localize cross-validation
localizingDistro = ns.Gauss1D(2.0, 0.6)
localizingWeightFcn = ns.DensityFunctor1D(localizingDistro)
localizingWeightXmin = localizingDistro.quantile(0.0)
localizingWeightXmax = localizingDistro.quantile(1.0)

# Create the LOrPE cross-validation object
# lorpe = lorpeFunctor(symbetaPower, xmin, xmax, bh, bandwidthFunctor,
#                     ns.Functor1Ref(localizingWeightFcn),
#                     localizingWeightXmin, localizingWeightXmax,
#                     nIntegSplits, nIntegPoints, normalize, weightedSample)

lorpe = lorpeFunctor(symbetaPower, xmin, xmax, bh, bandwidthFunctor,
                     1.0, xmin-1.0, xmax+1.0,
                     nIntegSplits, nIntegPoints, normalize, weightedSample)

# Check if we use RLCV and set the regularization parameter
# if this is the case
if useRLCV:
    cv.setRlcvAlpha(rlcvAlpha)

# Possible LOrPE code speed optimization options.
#
# The lorpe object will remember boundary-corrected local
# polynomials if the following method is called (uncomment
# the next line if desired):
# lorpe.useMemoisingKernels(True)

# The lorpe object will remember density normalizations
# if the following method is called:
# lorpe.memoizeNorms(True)

# The lorpe object will remember different kernels
# if the following method is called. This is only
# meaningful if the cross-validation runner can
# actually optimize not only the bandwidth but
# multiple filter degrees as well.
# lorpe.memoizeKernels(True)

# We now need to run "solveForLOrPEWeights" function to
# solve the mixture Fredholm equation. We will try to
# solve it for different values of alpha. This function
# has the following signature:
#
# solveForLOrPEWeights(lorpe, cv, signal, alpha, conv, maxIter)
#
# The following arguments must be provided:
#
#  lorpe    is an instance of LOrPE1DCV class, created normally
#           by LOrPE1DLSCVFunctor or LOrPE1DRLCVFunctor functions.
#
#  cv       is an instance of a class that runs the cross-validation
#           calculations and chooses the LOrPE filter degree and the
#           bandwidth. Currently, the classes that can be used for
#           this purpose are LOrPE1DFixedDegreeCVScanner,
#           LOrPE1DFixedDegreeCVRunner, LOrPE1DFixedDegreeCVPicker
#           (fixed LOrPE degrees), LOrPE1DVariableDegreeCVRunner,
#           and LOrPE1DVariableDegreeCVPicker (variable LOrPE degrees).
#
#  signal   is the signal distribution, an instance of AbsDistribution1D.
#
#  alpha    is the assumed signal fraction.
#
#  conv     is an instance of LOrPEWeightsLocalConvergence class.
#
#  maxIter  is the maximum number of allowed iterations.

# The function "pluginBandwidthScalingTable" generates a table
# of optimal bandwidth values as a function of filter degree
# using AMISE of a plug-in distribution. The plug-in distribution
# in this case is the standard normal. This, of course, is not
# an optimal approach for the exponential background. It is done
# here only because it is easy (the formulas are provided in
# the literature and implemented in npstat). Derivation of
# "minBwFactors", "maxBwFactors", and "guessBwFactors" needs
# to be improved.
bwScales = pluginBandwidthScalingTable(symbetaPower, sampleSize*(1 - alpha))
maxFilterDegree = 4
tableLength = maxFilterDegree*4 + 1
degrees = np.linspace(0.0, maxFilterDegree*1.0, tableLength)
bw0Min = 0.5
bw0Max = 5.0
bw0Guess = 1.0
scalingFactors = [bwScales(d)/bwScales(0.0) for d in degrees]
minBwFactors = [bw0Min*s for s in scalingFactors]
maxBwFactors = [bw0Max*s for s in scalingFactors]
guessBwFactors = [bw0Guess*s for s in scalingFactors]
initialStepSizeInDegreeGridCells = 2
initialStepSizeInFactorsGridCells = 5
nBwFactors = 101

# In the scheme used in this example, each bandwidth factor
# in the bandwidth grid will differ from the previous bandwidth
# by the factor exp(log(bw0Max/bw0Min)/(nBwFactors - 1))

for i, d in enumerate(degrees):
    print("Degree {}, min bwFactor = {:.4f}, max bwFactor = {:.4f}".format(
        d, minBwFactors[i], maxBwFactors[i]))

cv = ns.LOrPE1DVariableDegreeCVPicker(maxFilterDegree,
                                      minBwFactors,
                                      maxBwFactors,
                                      guessBwFactors,
                                      nBwFactors,
                                      initialStepSizeInFactorsGridCells,
                                      filterDegree,
                                      initialStepSizeInDegreeGridCells)
conv = ns.LOrPEWeightsLocalConvergence(convTol, 1, convDensity)
alphaSet = np.linspace(0.1, 0.5, 21)
cvDegs      = list()
cvBandwidth = list()
cvFcnValues = list()
logliValues = list()

# A cheat for faster convergence. Reinitialize
# the weights to their oracle values for every alpha.
signalValues = ns.scanDensity1D(signal, sampleCoords)
bgValues = ns.scanDensity1D(bg, sampleCoords)

# Measure execution time of the cycle
startTime = time.time()
for a in alphaSet:
    print("Solving Fredholm equation for alpha = {:.2f}".format(a))
    oracleWeights = bgValues/(a*signalValues + (1.0-a)*bgValues)
    lorpe.setWeights(oracleWeights)
    stats = ns.solveForLOrPEWeights(lorpe, cv, signal, a, conv, maxIter)
    if stats.converged():
        print("{} iterations, deg = {:.4f}, bwFactor = {:.4f}, boundary flags: {} {}".format(
            stats.nIterations(), stats.filterDegree(), stats.bwFactor(),
            stats.isOnDegreeBoundary(), stats.isOnBandwidthBoundary()))
        cvDegs.append(stats.filterDegree())
        cvBandwidth.append(stats.bwFactor())
        cvFcnValues.append(stats.cvFunction())

        # In the "semiMixLocalLogLikelihood" function call below,
        # the last three arguments define how the likelihood is
        # localized. Here, we use the same localization as for the
        # cross-validation, but it can, of course, be different.
        logli = ns.semiMixLocalLogLikelihood(
            lorpe.coords(), signal, a,
            lorpe.densityFunctor(stats.filterDegree(), stats.bwFactor()),
            lorpe.cvLocalizingWeightFunctor(),
            lorpe.localizingWeightXmin(),
            lorpe.localizingWeighXmax())
        print("Local log-likelihood is {:.6g}".format(logli))
        logliValues.append(logli)
    else:
        print("Iterations did not converge. Appending dummy bandwidth.")
        cvDegs.append(0.0)
        cvBandwidth.append(0.0)
        logliValues.append(0.0)

cycleTime = time.time() - startTime
print("Cycle time in seconds: {:.6g}".format(cycleTime))

# Plot the sample histogram
xh, yh = ns.histoOutline1D(h1)
plt.plot(xh, yh, 'k')
plt.xlabel(h1.axis(0).label())
plt.ylabel(h1.accumulatedDataLabel())
plt.title(h1.title())
plt.show()

# Show the filter degree dependence on alpha
plt.plot(alphaSet, cvDegs)
plt.xlabel('Signal Fraction')
plt.ylabel('CV Filter Degree')
plt.title('CV Filter Degree Dependence on the Assumed Signal Fraction')
plt.show()

# Show the bandwidth dependence on alpha
plt.plot(alphaSet, cvBandwidth)
plt.xlabel('Signal Fraction')
plt.ylabel('CV Bandwidth Factor')
plt.title('CV Bandwidth Dependence on the Assumed Signal Fraction')
plt.show()

# Show the cross-validation function dependence on alpha
plt.plot(alphaSet, cvFcnValues)
plt.xlabel('Signal Fraction')
plt.ylabel('CV Function')
plt.title('CV Function Dependence on the Assumed Signal Fraction')
plt.show()

# Show local log-likelihood dependence on alpha
plt.plot(alphaSet, logliValues)
plt.xlabel('Signal Fraction')
plt.ylabel('Local Log-Likelihood')
plt.title('Local Log-Likelihood Dependence on the Assumed Signal Fraction')
plt.show()
