#!/usr/bin/env python3
"""
This example illustrates how to fill NPStat 1-d histograms and plot them
using Matplotlib
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.1"
__date__ ="Sep 15 2019"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
import matplotlib.pyplot as plt

# Create the histogram axis (needed to construct the histogram below).
# Here, we use an axis with uniform binning. The axis constructor arguments
# are the number of bins, left histogram edge, right histogram edge, and
# the axis label.
histoAxis = ns.HistoAxis(20, -5.0, 5.0, "X")

# The following line creates a 1-d histogram with bins of type "int"
# (here, we use the C++ type naming convention). Other bin types supported
# out of the box include "unsigned char", "long long", "float", "double",
# "StatAccumulator", "WeightedStatAccumulator", "FloatSampleAccumulator",
# "DoubleSampleAccumulator", and "DoubleWeightedSampleAccumulator".
# Multivariate histograms are created in a similar manner but with
# a different constructor (see the example script "plot_histo_2d.py").
h1 = HistoND("int", histoAxis, "Example NPStat Histogram", "Counts")

# We will fill the histogram with Gaussian random numbers. In order to do
# that, we need a random number generator and a statistical distribution
# object which will be used to convert random numbers on [0, 1) into the
# random numbers distributed as we like. The default constructor of
# MersenneTwister generator picks up a different random seed every time,
# so that the histogram will be different every time you run this script.
rng = ns.MersenneTwister()
d1 = ns.Gauss1D(0, 1)

# Create a random sample with 1000 points
sample = d1.generate(rng, 1000)

# Fill the histogram. Here, each point contributes unit weight to
# the corresponding bin. The histogram "fill" method will consume
# the whole sample, point-by-point.
h1.fill(sample, 1)

# Create Matplotlib plot window
fig = plt.figure()
ax = fig.add_subplot(111)

# Generate xdata, ydata sequence for the outline drawn for 1-d histograms
x, y = ns.histoOutline1D(h1)

# The following line will draw a polygon filled with some default color
ax.fill(x, y)

# To draw a simple line instead of a polygon, do the following instead
# of the previous line
# ax.plot(x, y)

# Set the x and y ranges of the plot
ax.set_xlim(h1.axis(0).min(), h1.axis(0).max())
ax.set_ylim(0, 1.1*max(y))

# Set the axis labels and the plot title
plt.xlabel(h1.axis(0).label())
plt.ylabel(h1.accumulatedDataLabel())
plt.title(h1.title())

# Display the window. The code enters Matplotlib event loop here.
# You will be able to use the window interface to manipulate the
# figure (pan, zoom) and to save it in various graphical formats.
plt.show()
