#!/usr/bin/env python3
"""
This example illustrates various resampling techniques implemented in NPStat
"""

__author__="Igor Volobouev (i.volobouev@ttu.edu)"
__version__="1.0"
__date__ ="June 28 2022"

# Perform necessary imports
import npstat as ns
from npstat_utils import HistoND
import matplotlib.pyplot as plt

# Some parameters for the script
sampleSize = 2500   # Number of points to generate (sample size)
xmin = 0.0          # Left edge of the sample histogram
xmax = 5.0          # Right edge of the sample histogram
numBins = 50        # Number of bins for the sample histograms
nResamples = 10     # How many times to resample

# Generate a weighted sample of points on [xmin, xmax]. Here, we
# use a flat distribution to generate the points and some other
# distribution (a mixture of exponential and Gaussian) to assign
# the weights.
d1 = ns.Uniform1D(xmin, xmax - xmin)
rng = ns.MersenneTwister()
sampleTuple = d1.generate(rng, sampleSize)

# For subsequent use with DeltaMixture1D, the coordinates
# must be sorted
sampleCoords = list(sampleTuple)
sampleCoords.sort()

# Assign the weights
truncatedGauss = ns.TruncatedDistribution1D(ns.Gauss1D(2.0, 0.2), xmin, xmax)
truncatedExponential = ns.TruncatedDistribution1D(ns.Exponential1D(0.0, 1.0), xmin, xmax)
weightDistro = ns.DistributionMix1D()
weightDistro.add(truncatedGauss, 0.3)
weightDistro.add(truncatedExponential, 0.7)
weights = [weightDistro.density(x) for x in sampleCoords]

# Histogram for displaying the unweighted sample
xAxis = ns.HistoAxis(numBins, xmin, xmax, "X")
h1 = HistoND("int", xAxis, "Unweighted Sample Histogram", "Counts")
h1.fill(sampleTuple, 1)

# Histogram for displaying the weighted sample. Note that
# the type of bins is double rather than integer.
h2 = HistoND("double", xAxis, "Weighted Sample Histogram", "Weight")
weightedSample = list(zip(sampleCoords, weights))
h2.fill(weightedSample)

# The distribution representing the weighted sample. The first
# two arguments represent the additional shift and scale.
distroToResample = ns.DeltaMixture1D(0.0, 1.0, sampleCoords, weights)

# Resample unweighted sample with replacement. Histogram the resuts.
h3 = HistoND("int", xAxis, "Unweighted Resample Histogram", "Counts")
for i in range(nResamples):
    resampledPoints = ns.resampleWithReplacement(rng, sampleTuple, sampleSize)
    h3.fill(resampledPoints, 1)

# Resample weighted sample (with replacement)
h4 = HistoND("int", xAxis, "Weighted Resample Histogram", "Counts")
for i in range(nResamples):
    resampledPoints = distroToResample.generate(rng, sampleSize)
    h4.fill(resampledPoints, 1)

# Special resampling in which the points are assigned the weights
# that are inversely proportional to the frequency of point
# occurrence. This kind of resampling does not affect the mean
# but it does affect the variance. The variance will be the smallest
# where the weights peak.
h5 = HistoND("double", xAxis, "Special Resample Histogram 1", "Weight")
maxWeight = max(weights)
weightFcn1 = ns.AcceptanceFunctor1D(weightDistro, maxWeight)
unweightedDistroToResample = ns.DeltaMixture1D(0.0, 1.0, sampleCoords)
for i in range(nResamples):
    weigtedSample, stats = ns.sampleDistro1DWithWeight(
        unweightedDistroToResample, weightFcn1, rng, sampleSize, ns.SIZE_IS_N_GENERATED)
    h5.fill(weigtedSample)

# Similar special resampling. The variance will be the largest
# where the weights peak.
h6 = HistoND("double", xAxis, "Special Resample Histogram 2", "Weight")
minWeight = min(weights)
weightFcn2 = ns.InverseAcceptanceFunctor1D(weightDistro, minWeight)
for i in range(nResamples):
    weigtedSample, stats = ns.sampleDistro1DWithWeight(
        unweightedDistroToResample, weightFcn2, rng, sampleSize, ns.SIZE_IS_N_GENERATED)
    h6.fill(weigtedSample)

# Show all these histograms
for h in (h1, h2, h3, h4, h5, h6):
    xh, yh = ns.histoOutline1D(h)
    plt.plot(xh, yh, 'k')
    plt.xlabel(xAxis.label())
    plt.ylabel(h.accumulatedDataLabel())
    plt.title(h.title())
    plt.show()
