#ifndef NPSTAT_CLASSICALORTHOPOLY1DFROMWEIGHT_HH_
#define NPSTAT_CLASSICALORTHOPOLY1DFROMWEIGHT_HH_

/*!
// \file ClassicalOrthoPoly1DFromWeight.hh
//
// \brief Orthogonal polynomials on an interval with an arbitrary
//        user-provided weight function
//
// Author: I. Volobouev
//
// November 2020
*/

#include <cmath>
#include <vector>

#include "npstat/nm/AbsClassicalOrthoPoly1D.hh"
#include "npstat/nm/RecurrenceCoeffs.hh"

namespace npstat {
    template <class Functor1D>
    class ClassicalOrthoPoly1DFromWeight : public AbsClassicalOrthoPoly1D
    {
    public:
        /**
        // "nIntegPoints" parameter is the number of integration points
        // for the Fejer quadrature. Note that Fejer quadrature with n
        // points can integrate exactly polynomials of degree n-1 and
        // that this quadrature will be used for calculating various
        // inner products in which the density is used as the weight.
        // Therefore, "nIntegPoints" should be at least
        // 2*maxDegree + 1 + C, where "C" is the polynomial degree
        // that can be used to approximate the density reasonably well.
        */
        ClassicalOrthoPoly1DFromWeight(
            const Functor1D& fcn, unsigned maxDegree,
            unsigned nIntegPoints, double xmin, double xmax,
            OrthoPolyMethod m = OPOLY_STIELTJES);

        inline virtual ~ClassicalOrthoPoly1DFromWeight() {}

        inline virtual ClassicalOrthoPoly1DFromWeight* clone() const
            {return new ClassicalOrthoPoly1DFromWeight(*this);}

        inline virtual long double weight(const long double x) const
            {return fcn_(x)/wsum_;}
        inline virtual double xmin() const {return xmin_;}
        inline virtual double xmax() const {return xmax_;}
        inline virtual unsigned maxDegree() const {return maxdeg_;}

    protected:
        inline virtual std::pair<long double,long double>
        recurrenceCoeffs(const unsigned deg) const
        {
            const Private::Recurrence& r = rCoeffs_.at(deg);
            return std::pair<long double,long double>(r.alpha, r.sqrbeta);
        }

    private:
        Functor1D fcn_;
        std::vector<Private::Recurrence> rCoeffs_;
        long double wsum_;
        double xmin_;
        double xmax_;
        unsigned maxdeg_;
    };
}

#include "npstat/nm/ClassicalOrthoPoly1DFromWeight.icc"

#endif // NPSTAT_CLASSICALORTHOPOLY1DFROMWEIGHT_HH_
