#ifndef NPSTAT_CONVOLUTIONENGINEND_HH_
#define NPSTAT_CONVOLUTIONENGINEND_HH_

/*!
// \file ConvolutionEngineND.hh
//
// \brief Fast multidimensional convolutions via Fourier 
//        transforms (FFTW interface)
//
// Author: I. Volobouev
//
// December 2011
*/

#include <map>
#include <vector>

#include "npstat/nm/FourierImage.hh"

#ifdef SWIG
#include "npstat/wrap/arrayNDToNumpy.hh"
#include "npstat/wrap/numpyArrayUtils.hh"
#endif // SWIG

namespace npstat {
    /**
    // Class for performing multivariate convolutions by FFT (using FFTW
    // library). Internally, transforms are calculated in double precision.
    */
    class ConvolutionEngineND
    {
    public:
        /**
        // Constructor takes the shape of the data and FFTW optimization
        // flag as parameters. Make sure the data dimensions are sufficient
        // to avoid the convolution wrap-around.
        */
        ConvolutionEngineND(const unsigned *shape, unsigned rank,
                            unsigned optimization = FFTW_ESTIMATE);
        ~ConvolutionEngineND();

        /** Provide the filter for subsequent convolutions */
        template <typename Real>
        void setFilter(const Real* data, const unsigned *dataShape,
                       unsigned dataRank);

        /**
        // Convolve provided data with the filter previously established
        // by setFilter
        */
        template <typename Real1, typename Real2>
        void convolveWithFilter(const Real1* in, Real2* out,
                                const unsigned *dataShape, unsigned dataRank);

        /**
        // Provide a filter for subsequent convolutions.
        //
        // "id" is the user id for the filter which can be later reused
        // with the method "convolveWithDeposit".
        */
        template <typename Real>
        void depositFilter(unsigned long id,
                           const Real* data, const unsigned *dataShape,
                           unsigned dataRank);

        /**
        // Convolve provided data with one of the filters previously
        // established by the depositFilter method. The "id" argument
        // is used to identify a particular filter.
        */
        template <typename Real1, typename Real2>
        void convolveWithDeposit(unsigned long id,
                                 const Real1* in, Real2* out,
                                 const unsigned *dataShape, unsigned dataRank);

        /**
        // Convolve the filter established by the "setFilter" method
        // with a previously deposited filter
        */
        template <typename Real2>
        void convolveFilterAndDeposit(unsigned long id,  Real2* out,
                                      const unsigned *dataShape,
                                      unsigned dataRank);

        /**
        // Deconvolve the filter established by the "setFilter" method
        // with a previously deposited filter
        */
        template <typename Real2>
        void deconvolveFilterAndDeposit(unsigned long id, Real2* out,
                                        const unsigned *dataShape,
                                        unsigned dataRank);

        /**
        // Discard a filter previously deposited by depositFilter method.
        // This method returns "true" if the filter with the given id
        // was indeed discarded and "false" if such a filter was not found.
        */
        bool discardFilter(unsigned long id);

        /** Discard all filters deposited by depositFilter method */
        void clearFilterBank();

        /** Check if an array shape is compatible with this filter */
        bool isShapeCompatible(const unsigned *shape, unsigned rank) const;

        /** Expected data dimensionality */
        inline unsigned rank() const {return shape_.size();}

        /** Expected shape of the data */
        inline const std::vector<unsigned>& shape() const {return shape_;}

        /** Expected data length */
        inline unsigned long dataLength() const {return dataLen_;}

        // Some accessors
        inline bool isFilterSet() const {return validFilter_;}
        inline unsigned filterBankSize() const {return filterMap_.size();}

    private:
        typedef std::map<unsigned long,FourierImage*> FilterMap;

        ConvolutionEngineND();
        ConvolutionEngineND(const ConvolutionEngineND&);
        ConvolutionEngineND& operator=(const ConvolutionEngineND&);

        template <typename Real1, typename Real2>
        void convolveWithImage(const fftw_complex* image,
                               const Real1* in, Real2* out,
                               const unsigned *dataShape, unsigned dataRank);
        fftw_plan pf;
        fftw_plan pb;

        double* in;
        fftw_complex* out;
        fftw_complex* filterImage;

        std::vector<unsigned> shape_;
        unsigned long dataLen_;
        unsigned long cmplLen_;
        bool validFilter_;

        FilterMap filterMap_;

#ifdef SWIG
    public:
        inline void setFilter2(PyObject* input)
        {
            if (!isNumpyShapeCompatible(input, &shape_[0], shape_.size())) throw std::invalid_argument(
                "In npstat::ConvolutionEngineND::setFilter2: incompatible array shape");
            const int typenum = numpyArrayType(input);
            if (typenum == NumpyTypecode<double>::code)
                setFilter((double*)PyArray_DATA((PyArrayObject*)input), &shape_[0], shape_.size());
            else if (typenum == NumpyTypecode<float>::code)
                setFilter((float*)PyArray_DATA((PyArrayObject*)input), &shape_[0], shape_.size());
            else if (typenum == NumpyTypecode<int>::code)
                setFilter((int*)PyArray_DATA((PyArrayObject*)input), &shape_[0], shape_.size());
            else if (typenum == NumpyTypecode<long long>::code)
                setFilter((long long*)PyArray_DATA((PyArrayObject*)input), &shape_[0], shape_.size());
            else if (typenum == NumpyTypecode<long int>::code)
                setFilter((long int*)PyArray_DATA((PyArrayObject*)input), &shape_[0], shape_.size());
            else if (typenum == NumpyTypecode<unsigned char>::code)
                setFilter((unsigned char*)PyArray_DATA((PyArrayObject*)input), &shape_[0], shape_.size());
            else
                throw std::invalid_argument("In npstat::ConvolutionEngineND::setFilter2: "
                                            "unsupported input array type");
        }

        inline void depositFilter2(const unsigned long id,
                                   PyObject* input)
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code) throw std::invalid_argument(
                "In npstat::ConvolutionEngineND::depositFilter2: input is not an array of doubles");
            if (!isNumpyShapeCompatible(input, &shape_[0], shape_.size())) throw std::invalid_argument(
                "In npstat::ConvolutionEngineND::depositFilter2: incompatible array shape");
            double* in = (double*)PyArray_DATA((PyArrayObject*)input);
            depositFilter(id, in, &shape_[0], shape_.size());
        }

        inline PyObject* convolveWithFilter2(PyObject* input,
                                             const bool makeNonNegative = false)
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code) throw std::invalid_argument(
                "In npstat::ConvolutionEngineND::convolveWithFilter2: input is not an array of doubles");
            const unsigned dataRank = shape_.size();
            assert(dataRank <= CHAR_BIT*sizeof(unsigned long));
            if (!isNumpyShapeCompatible(input, &shape_[0], dataRank)) throw std::invalid_argument(
                "In npstat::ConvolutionEngineND::convolveWithFilter2: incompatible array shape");
            double* in = (double*)PyArray_DATA((PyArrayObject*)input);
            npy_intp sh[CHAR_BIT*sizeof(unsigned long)];
            for (unsigned i=0; i<dataRank; ++i)
                sh[i] = shape_[i];
            PyObject* xarr = PyArray_SimpleNew(dataRank, sh, typenum);
            if (xarr)
            {
                double* out = (double*)PyArray_DATA((PyArrayObject*)xarr);
                try
                {
                    convolveWithFilter(in, out, &shape_[0], dataRank);
                }
                catch (const std::exception& e)
                {
                    Py_DECREF(xarr);
                    throw e;
                }
                if (makeNonNegative)
                    for (unsigned long i=0; i<dataLen_; ++i)
                        if (out[i] < 0.0)
                            out[i] = 0.0;
            }
            return xarr;
        }

        inline PyObject* convolveWithDeposit2(const unsigned long id,
                                              PyObject* input,
                                              const bool makeNonNegative = false)
        {
            const int typenum = numpyArrayType(input);
            if (typenum != NumpyTypecode<double>::code) throw std::invalid_argument(
                "In npstat::ConvolutionEngineND::convolveWithDeposit2: input is not an array of doubles");
            const unsigned dataRank = shape_.size();
            assert(dataRank <= CHAR_BIT*sizeof(unsigned long));
            if (!isNumpyShapeCompatible(input, &shape_[0], dataRank)) throw std::invalid_argument(
                "In npstat::ConvolutionEngineND::convolveWithDeposit2: incompatible array shape");
            double* in = (double*)PyArray_DATA((PyArrayObject*)input);
            npy_intp sh[CHAR_BIT*sizeof(unsigned long)];
            for (unsigned i=0; i<dataRank; ++i)
                sh[i] = shape_[i];
            PyObject* xarr = PyArray_SimpleNew(dataRank, sh, typenum);
            if (xarr)
            {
                double* out = (double*)PyArray_DATA((PyArrayObject*)xarr);
                try
                {
                    convolveWithDeposit(id, in, out, &shape_[0], dataRank);
                }
                catch (const std::exception& e)
                {
                    Py_DECREF(xarr);
                    throw e;
                }
                if (makeNonNegative)
                    for (unsigned long i=0; i<dataLen_; ++i)
                        if (out[i] < 0.0)
                            out[i] = 0.0;
            }
            return xarr;
        }

        inline PyObject* convolveFilterAndDeposit2(const unsigned long id,
                                                   const bool makeNonNegative = false)
        {
            const int typenum = NumpyTypecode<double>::code;
            const unsigned dataRank = shape_.size();
            assert(dataRank <= CHAR_BIT*sizeof(unsigned long));
            npy_intp sh[CHAR_BIT*sizeof(unsigned long)];
            for (unsigned i=0; i<dataRank; ++i)
                sh[i] = shape_[i];
            PyObject* xarr = PyArray_SimpleNew(dataRank, sh, typenum);
            if (xarr)
            {
                double* out = (double*)PyArray_DATA((PyArrayObject*)xarr);
                try
                {
                    convolveFilterAndDeposit(id, out, &shape_[0], dataRank);
                }
                catch (const std::exception& e)
                {
                    Py_DECREF(xarr);
                    throw e;
                }
                if (makeNonNegative)
                    for (unsigned long i=0; i<dataLen_; ++i)
                        if (out[i] < 0.0)
                            out[i] = 0.0;
            }
            return xarr;
        }
#endif // SWIG
    };
}

#include "npstat/nm/ConvolutionEngineND.icc"

#endif // NPSTAT_CONVOLUTIONENGINEND_HH_
