#include <cassert>
#include <stdexcept>
#include <algorithm>
#include <sstream>

#include "npstat/nm/ScanExtremum1D.hh"
#include "npstat/nm/isMonotonous.hh"
#include "npstat/nm/MathUtils.hh"

static void validateScan(const double* xcoords, const unsigned long xLen,
                         const double* yvalues, const unsigned long yLen,
                         const char* where)
{
    if (xLen != yLen)
    {
        std::ostringstream os;
        os << "In npstat::" << where << ": "
           << "incompatible array sizes";
        throw std::invalid_argument(os.str());
    }
    if (xLen < 2UL)
    {
        std::ostringstream os;
        os << "In npstat::" << where << ": "
           << "insufficient scan length, must be at least 2";
        throw std::invalid_argument(os.str());
    }
    assert(xcoords);
    assert(yvalues);
    if (!npstat::isStrictlyIncreasing(xcoords, xcoords+xLen))
    {
        std::ostringstream os;
        os << "In npstat::" << where << ": "
           << "coordinates are not monotonously increasing";
        throw std::invalid_argument(os.str());
    }
}

namespace npstat {
    ScanExtremum1D findScanMinimum1D(const double* xcoords, const unsigned long xLen,
                                     const double* yvalues, const unsigned long yLen)
    {
        validateScan(xcoords, xLen, yvalues, yLen, "findScanMinimum1D");
        const auto minpos = std::min_element(yvalues, yvalues+yLen);
        const unsigned long imin = std::distance(yvalues, minpos);
        if (imin == 0UL || imin == yLen - 1UL)
            return ScanExtremum1D(xcoords[imin], yvalues[imin], true, true);
        if (yvalues[imin] == yvalues[imin-1] && yvalues[imin] == yvalues[imin+1])
            return ScanExtremum1D(xcoords[imin], yvalues[imin], true, false);

        double x, y;
        bool onTheBoundary = false;
        const bool isMinumum = parabolicExtremum(xcoords[imin-1], yvalues[imin-1],
                                                 xcoords[imin], yvalues[imin],
                                                 xcoords[imin+1], yvalues[imin+1],
                                                 &x, &y);
        assert(isMinumum);
        if (x < xcoords[imin-1])
        {
            x = xcoords[imin-1];
            y = yvalues[imin-1];
            onTheBoundary = imin-1 == 0UL;
        }
        if (x > xcoords[imin+1])
        {
            x = xcoords[imin+1];
            y = yvalues[imin+1];
            onTheBoundary = imin+1 == yLen - 1UL;
        }
        return ScanExtremum1D(x, y, true, onTheBoundary);
    }

    ScanExtremum1D findScanMaximum1D(const double* xcoords, const unsigned long xLen,
                                     const double* yvalues, const unsigned long yLen)
    {
        validateScan(xcoords, xLen, yvalues, yLen, "findScanMaximum1D");
        const auto maxpos = std::max_element(yvalues, yvalues+yLen);
        const unsigned long imax = std::distance(yvalues, maxpos);
        if (imax == 0UL || imax == yLen - 1UL)
            return ScanExtremum1D(xcoords[imax], yvalues[imax], false, true);
        if (yvalues[imax] == yvalues[imax-1] && yvalues[imax] == yvalues[imax+1])
            return ScanExtremum1D(xcoords[imax], yvalues[imax], false, false);

        double x, y;
        bool onTheBoundary = false;
        const bool isMinumum = parabolicExtremum(xcoords[imax-1], yvalues[imax-1],
                                                 xcoords[imax], yvalues[imax],
                                                 xcoords[imax+1], yvalues[imax+1],
                                                 &x, &y);
        assert(!isMinumum);
        if (x < xcoords[imax-1])
        {
            x = xcoords[imax-1];
            y = yvalues[imax-1];
            onTheBoundary = imax-1 == 0UL;
        }
        if (x > xcoords[imax+1])
        {
            x = xcoords[imax+1];
            y = yvalues[imax+1];
            onTheBoundary = imax+1 == yLen - 1UL;
        }
        return ScanExtremum1D(x, y, false, onTheBoundary);
    }
}
