#ifndef NPSTAT_COORDANDWEIGHT_HH_
#define NPSTAT_COORDANDWEIGHT_HH_

//======================================================================
// coordAndWeight.hh
//
// This is an internal header which is subject to change without notice.
// Application code should never call the functions declared/defined in
// this header directly.
//
// Author: I. Volobouev
//
// May 2022
//======================================================================

#include <vector>
#include <utility>
#include <cassert>
#include <stdexcept>

namespace npstat {
    namespace Private {
        template<typename Numeric>
        inline void coordAndWeight(const Numeric& r,
                                   double* x, double* w)
        {
            *x = r;
            *w = 1.0;
        }

        template<typename Num1, typename Num2>
        inline void coordAndWeight(const std::pair<Num1, Num2>& r,
                                   double* x, double* w)
        {
            *x = r.first;
            *w = r.second;
        }

        template<typename Numeric>
        struct WeightedCoord
        {
            inline WeightedCoord(const double x, const double /* w */)
                : value(static_cast<Numeric>(x)) {}
            Numeric value;
        };

        template<typename Num1, typename Num2>
        struct WeightedCoord<std::pair<Num1, Num2> >
        {
            inline WeightedCoord(const double x, const double w)
                : value(x, w) {}
            std::pair<Num1, Num2> value;
        };

        template<typename Numeric>
        struct SampleWeight
        {
            inline SampleWeight(const std::vector<Numeric>& vec)
                : value(vec.size()) {}
            double value;
        };

        template<typename Num1, typename Num2>
        struct SampleWeight<std::pair<Num1, Num2> >
        {
            inline SampleWeight(const std::vector<std::pair<Num1, Num2> >& vec)
            {
                long double sum = 0.0L;
                const unsigned long sz = vec.size();
                for (unsigned long i=0; i<sz; ++i)
                    sum += static_cast<long double>(vec[i].second);
                value = sum;
            }
            double value;
        };

        template<typename Numeric>
        struct SampleSquaredWeight
        {
            inline SampleSquaredWeight(const std::vector<Numeric>& vec)
                : value(vec.size()) {}
            double value;
        };

        template<typename Num1, typename Num2>
        struct SampleSquaredWeight<std::pair<Num1, Num2> >
        {
            inline SampleSquaredWeight(const std::vector<std::pair<Num1, Num2> >& vec)
            {
                long double sum = 0.0L;
                const unsigned long sz = vec.size();
                for (unsigned long i=0; i<sz; ++i)
                {
                    const long double w = static_cast<long double>(vec[i].second);
                    sum += w*w;
                }
                value = sum;
            }
            double value;
        };

        template<typename Numeric>
        struct SetPairSeconds
        {
            inline static void set(Numeric* /* arr */, const double* /* weights */,
                                   const unsigned long /* nWeights */)
            {
                throw std::runtime_error("In npstat::Private::SetPairSeconds::set: "
                                         "array to set is not of pair type");
            }
        };

        template<typename Num1, typename Num2>
        struct SetPairSeconds<std::pair<Num1, Num2> >
        {
            inline static void set(std::pair<Num1,Num2>* arr, const double* weights,
                                   const unsigned long nWeights)
            {
                if (nWeights)
                {
                    assert(arr);
                    assert(weights);
                    for (unsigned long i=0; i<nWeights; ++i, ++arr)
                        arr->second = *weights++;
                }
            }                   
        };
    }
}

#endif // NPSTAT_COORDANDWEIGHT_HH_
