#ifndef NPSTAT_RESAMPLEWITHREPLACEMENT_HH_
#define NPSTAT_RESAMPLEWITHREPLACEMENT_HH_

/*!
// \file resampleWithReplacement.hh
//
// \brief Resampling with replacement from a vector of points
//
// Author: I. Volobouev
//
// May 2022
*/

#include <vector>
#include <cassert>
#include <stdexcept>

#include "npstat/rng/AbsRandomGenerator.hh"

namespace npstat {
    /**
    // On output, vector "to" will be filled by a random sample
    // of elements of the vector "from", chosen with replacement,
    // according to the sample size requested. Note that, due to
    // the simplicity of the algorithm and for very large "from"
    // vectors, discreteness of random doubles on [0, 1) can affect
    // resampling uniformity.
    */
    template<typename T>
    void resampleWithReplacement(AbsRandomGenerator& gen,
                                 const std::vector<T>& from,
                                 std::vector<T>* to,
                                 const unsigned long sampleSizeWanted)
    {
        const unsigned long sz = from.size();
        if (!sz) throw std::invalid_argument(
            "In npstat::resampleWithReplacement: empty input sample");
        const double dsz = sz;

        assert(to);
        to->clear();
        to->reserve(sampleSizeWanted);

        for (unsigned long i=0; i<sampleSizeWanted; ++i)
        {
            unsigned long idx = sz;
            while (idx >= sz)
                idx = static_cast<unsigned long>(dsz*gen());
            to->push_back(from[idx]);
        }
    }

#ifdef SWIG
    template<typename T>
    inline std::vector<T> resampleWithReplacement2(
        AbsRandomGenerator& gen, const std::vector<T>& from,
        const unsigned long sampleSizeWanted)
    {
        std::vector<T> to;
        resampleWithReplacement(gen, from, &to, sampleSizeWanted);
        return to;
    }
#endif
}

#endif // NPSTAT_RESAMPLEWITHREPLACEMENT_HH_
