#include <algorithm>

#include "geners/binaryIO.hh"

#include "npstat/stat/DeltaMixture1D.hh"
#include "npstat/stat/distributionReadError.hh"

namespace npstat {
    bool DeltaMixture1D::isEqual(const AbsDistribution1D& o) const
    {
        const DeltaMixture1D& r = static_cast<const DeltaMixture1D&>(o);
        return AbsScalableDistribution1D::isEqual(r) &&
            locations_ == r.locations_ && weights_ == r.weights_;
    }

    void DeltaMixture1D::buildCdfTable()
    {
        const unsigned len = locations_.size();
        assert(len);
        assert(len == weights_.size());

        cdf_.clear();
        cdf_.resize(len);
        {
            long double sum = 0.0L;
            for (unsigned i=0; i<len; ++i)
            {
                sum += weights_[i];
                cdf_[i] = sum;
            }
            const double dsum = sum;
            assert(dsum > 0.0);
            for (unsigned i=0; i<len; ++i)
            {
                weights_[i] /= dsum;
                cdf_[i] /= dsum;
            }
        }

        exceedance_.clear();
        exceedance_.resize(len);
        {
            long double sum = 0.0L;
            for (long i = len-1U; i>=0L; --i)
            {
                sum += weights_[i];
                exceedance_[i] = sum;
            }
            const double dsum = sum;
            for (unsigned i=0; i<len; ++i)
                exceedance_[i] /= dsum;
        }
    }

    bool DeltaMixture1D::write(std::ostream& os) const
    {
        AbsScalableDistribution1D::write(os);
        gs::write_pod_vector(os, locations_);
        gs::write_pod_vector(os, weights_);
        return !os.fail();
    }

    DeltaMixture1D* DeltaMixture1D::read(const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(gs::ClassId::makeId<DeltaMixture1D>());
        current.ensureSameId(id);

        double location, scale;
        if (AbsScalableDistribution1D::read(in, &location, &scale))
        {
            std::vector<double> coords, weights;
            gs::read_pod_vector(in, &coords);
            gs::read_pod_vector(in, &weights);
            const unsigned len = coords.size();
            if (!in.fail() && len && len == weights.size())
                return new DeltaMixture1D(location, scale, &coords[0], &weights[0], len);
        }
        distributionReadError(in, classname());
        return 0;
    }

    DeltaMixture1D::DeltaMixture1D(const double location, const double scale,
                                   const std::vector<double>& params)
        : AbsScalableDistribution1D(location, scale)
    {
        const unsigned nParams = params.size();
        if (!nParams) throw std::invalid_argument(
            "In npstat::DeltaMixture1D constructor: "
            "must have a positive number of parameters");
        if (nParams % 2) throw std::invalid_argument(
            "In npstat::DeltaMixture1D constructor: "
            "must have an even number of parameters");
        const unsigned half = nParams/2;
        initialize(&params[0], &params[half], half);
    }

    double DeltaMixture1D::unscaledCdf(const double x) const
    {
        const unsigned ind = std::upper_bound(locations_.begin(), locations_.end(), x) - locations_.begin();
        if (ind)
            return cdf_[ind-1U];
        else
            return 0.0;
    }

    double DeltaMixture1D::unscaledExceedance(const double x) const
    {
        // Sum of cdf and exceedance should be 1
        const unsigned ind = std::upper_bound(locations_.begin(), locations_.end(), x) - locations_.begin();
        if (ind == exceedance_.size())
            return 0.0;
        else
            return exceedance_[ind];
    }

    double DeltaMixture1D::unscaledQuantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::DeltaMixture1D::unscaledQuantile: "
            "cdf argument outside of [0, 1] interval");
        if (r1 == 0.0)
            return locations_[0];
        else if (r1 == 1.0)
            return locations_.back();
        else
        {
            const unsigned ind = std::lower_bound(cdf_.begin(), cdf_.end(), r1) - cdf_.begin();
            return locations_.at(ind);
        }
    }

    double DeltaMixture1D::weightSum(const unsigned imin, const unsigned imax) const
    {
        long double sum = 0.0L;
        const unsigned sz = weights_.size();
        const unsigned upper = std::min(imax, sz);
        if (imin < upper)
        {
            const double* w = &weights_[imin];
            for (unsigned i=imin; i<upper; ++i)
                sum += *w++;
        }
        return sum;
    }

    void DeltaMixture1D::determineRange(
        const double a0, const bool includeLeftBoundary,
        const double b0, const bool includeRightBoundary,
        unsigned* imin, unsigned* imax) const
    {
        assert(a0 <= b0);
        const double s = this->scale();
        const double mu = this->location();
        const double a = (a0 - mu)/s;
        const double b = (b0 - mu)/s;
        const unsigned sz = locations_.size();

        *imin = std::lower_bound(locations_.begin(), locations_.end(), a) - locations_.begin();
        *imax = std::lower_bound(locations_.begin(), locations_.end(), b) - locations_.begin();
        if (!includeLeftBoundary)
            if (*imin < sz)
                if (locations_[*imin] == a)
                    ++*imin;
        if (includeRightBoundary)
            if (*imax < sz)
                if (locations_[*imax] == b)
                    ++*imax;
    }

    double DeltaMixture1D::integral(const double a, const bool includePointAtA,
                                    const double b, const bool includePointAtB) const
    {
        if (a > b)
            return -integral(b, includePointAtB, a, includePointAtA);
        unsigned imin, imax;
        determineRange(a, includePointAtA, b, includePointAtB, &imin, &imax);
        return weightSum(imin, imax);
    }

    long double DeltaMixture1D::moment(const long double center,
                                        const unsigned order) const
    {
        if (order == 0U)
            return 1.0L;

        const double s = this->scale();
        const double mu = this->location();
        const unsigned sz = locations_.size();
        assert(sz == weights_.size());
        long double sum = 0.0L;
        for (unsigned i=0; i<sz; ++i)
        {
            const long double delta = locations_[i]*s + mu - center;
            if (order > 1U)
                sum += weights_[i]*powl(delta, order);
            else
                sum += weights_[i]*delta;
        }
        return sum;
    }

    void DeltaMixture1D::moments(const long double center, const unsigned maxOrder,
                                  long double* o_moments) const
    {
        assert(o_moments);
        o_moments[0] = 1.0L;
        for (unsigned i=1; i<=maxOrder; ++i)
            o_moments[i] = 0.0L;
        if (maxOrder)
        {
            const double s = this->scale();
            const double mu = this->location();
            const unsigned sz = locations_.size();
            assert(sz == weights_.size());

            for (unsigned i=0; i<sz; ++i)
            {
                const long double delta = locations_[i]*s + mu - center;
                o_moments[1] += weights_[i]*delta;
                for (unsigned k=2; k<=maxOrder; ++k)
                    o_moments[k] += weights_[i]*powl(delta, k);
            }
        }
    }

    void DeltaMixture1D::centralMoments(const unsigned maxOrder,
                                        long double* o_moments) const
    {
        assert(o_moments);
        o_moments[0] = 1.0L;
        if (maxOrder)
        {
            const long double mean = moment(0.0L, 1U);
            if (maxOrder > 1U)
                moments(mean, maxOrder, o_moments);
            o_moments[1] = mean;
        }
    }
}
