#include <cmath>
#include <cassert>
#include <stdexcept>

#include "npstat/stat/DensityOrthoPoly1D.hh"
#include "npstat/nm/FejerQuadrature.hh"

namespace npstat {
    DensityOrthoPoly1D::DensityOrthoPoly1D(const AbsDistribution1D& distro,
                                           const unsigned maxDegree,
                                           const unsigned nIntegPoints,
                                           const OrthoPolyMethod m)
        : distro_(distro.clone()), maxdeg_(maxDegree)
    {
        if (maxdeg_)
        {
            if (!nIntegPoints) throw std::invalid_argument(
                "In npstat::DensityOrthoPoly1D constructor: the number of "
                "integration points must be positive");
            std::shared_ptr<const FejerQuadrature> quad =
                FejerQuadratureSet::quad(nIntegPoints);
            assert(quad->npoints() == nIntegPoints);
            initialize(*quad, m);
        }
        else
        {
            // Just believe that the distro is correctly
            // normalized. We should not be asked for the
            // recurrence coefficients in the future.
            rCoeffs_.reserve(1);
            rCoeffs_.push_back(Private::Recurrence(0.0L, 1.0L));
        }
    }

    void DensityOrthoPoly1D::initialize(const FejerQuadrature& quad,
                                        const OrthoPolyMethod m)
    {
        typedef std::pair<double,double> MeasurePt;

        const unsigned nIntegPoints = quad.npoints();
        const DensityFunctor1D weight(*distro_);
        const std::vector<MeasurePt>& measure = quad.weightedIntegrationPoints(
            weight, distro_->quantile(0.0), distro_->quantile(1.0));
        assert(measure.size() == nIntegPoints);

        long double wsum = 0.0L;
        for (unsigned i=0; i<nIntegPoints; ++i)
            wsum += measure[i].second;
        if (wsum <= 0.0L) throw std::invalid_argument(
            "In npstat::DensityOrthoPoly1D::initialize: "
            "total weight is not positive");
        Private::calcRecurrenceCoeffs(measure, maxdeg_, wsum, m, &rCoeffs_);
    }

    std::pair<long double,long double>
    DensityOrthoPoly1D::recurrenceCoeffs(const unsigned deg) const
    {
        const Private::Recurrence& r = rCoeffs_.at(deg);
        return std::pair<long double,long double>(r.alpha, r.sqrbeta);
    }

    DensityOrthoPoly1D::DensityOrthoPoly1D(const DensityOrthoPoly1D& r)
        : distro_(r.distro_->clone()),
          rCoeffs_(r.rCoeffs_),
          maxdeg_(r.maxdeg_)
    {
    }

    DensityOrthoPoly1D& DensityOrthoPoly1D::operator=(const DensityOrthoPoly1D& r)
    {
        if (&r != this)
        {
            distro_ = CPP11_auto_ptr<AbsDistribution1D>(r.distro_->clone());
            rCoeffs_ = r.rCoeffs_;
            maxdeg_ = r.maxdeg_;
        }
        return *this;
    }
}
