#ifndef NPSTAT_DENSITYORTHOPOLY1D_HH_
#define NPSTAT_DENSITYORTHOPOLY1D_HH_

/*!
// \file DensityOrthoPoly1D.hh
//
// \brief Continuous orthonormal polynomial series in one dimension
//        generated for a continuous density used as the weight
//
// Author: I. Volobouev
//
// July 2020
*/

#include <vector>

#include "geners/CPP11_auto_ptr.hh"

#include "npstat/nm/AbsClassicalOrthoPoly1D.hh"
#include "npstat/nm/RecurrenceCoeffs.hh"

#include "npstat/stat/AbsDistribution1D.hh"

namespace npstat {
    class FejerQuadrature;

    class DensityOrthoPoly1D : public AbsClassicalOrthoPoly1D
    {
    public:
        /**
        // "nIntegPoints" parameter is the number of integration points
        // for the Fejer quadrature. Note that Fejer quadrature with n
        // points can integrate exactly polynomials of degree n-1 and
        // that this quadrature will be used for calculating various
        // inner products in which the density is used as the weight.
        // Therefore, "nIntegPoints" should be at least
        // 2*maxDegree + 1 + C, where "C" is the polynomial degree
        // that can be used to approximate the density reasonably well.
        //
        // The support of the weight for the integration purpose will
        // be set to [distro.quantile(0.0), distro.quantile(1.0)]. If
        // this interval is very wide, the results may be meaningless
        // (monomials like x^n can exceed the floating point dynamic
        // range, etc).
        */
        DensityOrthoPoly1D(const AbsDistribution1D& distro,
                           unsigned maxDegree,
                           unsigned nIntegPoints,
                           OrthoPolyMethod m = OPOLY_STIELTJES);

        DensityOrthoPoly1D(const DensityOrthoPoly1D&);
        DensityOrthoPoly1D& operator=(const DensityOrthoPoly1D&);

        inline virtual ~DensityOrthoPoly1D() {}

        inline virtual DensityOrthoPoly1D* clone() const
            {return new DensityOrthoPoly1D(*this);}

        inline virtual long double weight(const long double x) const
            {return distro_->density(x);}
        inline virtual double xmin() const {return distro_->quantile(0.0);}
        inline virtual double xmax() const {return distro_->quantile(1.0);}
        inline virtual unsigned maxDegree() const {return maxdeg_;}

    protected:
        virtual std::pair<long double,long double>
        recurrenceCoeffs(unsigned deg) const;

    private:
        void initialize(const FejerQuadrature& quad, OrthoPolyMethod m);
        
        CPP11_auto_ptr<AbsDistribution1D> distro_;
        std::vector<Private::Recurrence> rCoeffs_;
        unsigned maxdeg_;
    };
}

#endif // NPSTAT_DENSITYORTHOPOLY1D_HH_
