#include <numeric>
#include <stdexcept>

#include "geners/GenericIO.hh"
#include "geners/CPP11_auto_ptr.hh"
#include "geners/IOException.hh"

#include "npstat/rng/AbsRandomGenerator.hh"

#include "npstat/stat/StatUtils.hh"
#include "npstat/stat/DistributionMixND.hh"

namespace npstat {
    DistributionMixND::DistributionMixND(const unsigned i_dim)
        : AbsDistributionND(i_dim),
          wsum_(0.0L),
          isNormalized_(false)
    {
    }

    DistributionMixND::DistributionMixND(const DistributionMixND& r)
        : AbsDistributionND(r),
          entries_(r.entries_.size(), 0),
          weights_(r.weights_),
          weightCdf_(r.weightCdf_),
          wsum_(r.wsum_),
          isNormalized_(r.isNormalized_)
    {
        const unsigned n = entries_.size();
        for (unsigned i=0; i<n; ++i)
            entries_[i] = r.entries_[i]->clone();
    }

    DistributionMixND& DistributionMixND::operator=(const DistributionMixND& r)
    {
        if (this != &r)
        {
            AbsDistributionND::operator=(r);
            {
                const unsigned n = entries_.size();
                for (unsigned i=0; i<n; ++i)
                    delete entries_[i];
            }
            entries_ = r.entries_;
            const unsigned n = entries_.size();
            for (unsigned i=0; i<n; ++i)
                entries_[i] = 0;
            for (unsigned i=0; i<n; ++i)
                entries_[i] = r.entries_[i]->clone();
            weights_ = r.weights_;
            weightCdf_ = r.weightCdf_;
            wsum_ = r.wsum_;
            isNormalized_ = r.isNormalized_;
        }
        return *this;
    }

    DistributionMixND::~DistributionMixND()
    {
        const unsigned n = entries_.size();
        for (unsigned i=0; i<n; ++i)
            delete entries_[i];
    }

    DistributionMixND& DistributionMixND::add(
        const AbsDistributionND& distro, const double weight)
    {
        if (weight < 0.0) throw std::invalid_argument(
            "In npstat::DistributionMixND::add: weights must be non-negative");
        entries_.push_back(distro.clone());
        weights_.push_back(weight);
        isNormalized_ = false;
        return *this;
    }

    void DistributionMixND::setWeights(const double* weights, const unsigned nWeights)
    {
        const unsigned nentries = entries_.size();
        if (nentries != nWeights) throw std::invalid_argument(
            "In npstat::DistributionMixND::setWeights: wrong # of weights");
        if (nWeights)
        {
            assert(weights);
            for (unsigned i=0; i<nWeights; ++i)
                if (weights[i] < 0.0) throw std::invalid_argument(
                    "In npstat::DistributionMixND::setWeights: "
                    "all weights must be non-negative");
            for (unsigned i=0; i<nWeights; ++i)
                weights_[i] = weights[i];
            isNormalized_ = false;
        }
    }

    double DistributionMixND::getWeight(const unsigned n) const
    {
        if (!isNormalized_)
            (const_cast<DistributionMixND*>(this))->normalize();
        return weights_.at(n)/wsum_;
    }

    void DistributionMixND::normalize()
    {
        const unsigned nentries = entries_.size();
        if (!nentries) throw std::runtime_error(
            "In npstat::DistributionMixND::normalize: no components in the mix");
        weightCdf_.clear();
        weightCdf_.reserve(nentries);
        const long double wnorm = std::accumulate(
            weights_.begin(), weights_.end(), 0.0L);
        if (wnorm <= 0.0L) throw std::runtime_error(
            "In npstat::DistributionMixND::normalize: sum of weights is not positive");
        wsum_ = 0.0L;
        for (unsigned i=0; i<nentries; ++i)
        {
            weightCdf_.push_back(static_cast<double>(wsum_/wnorm));
            wsum_ += weights_[i];
        }
        isNormalized_ = true;
    }

    bool DistributionMixND::isEqual(const AbsDistributionND& oth) const
    {
        const DistributionMixND& r = static_cast<const DistributionMixND&>(oth);
        if (dim_ != r.dim_)
            return false;
        const unsigned n = entries_.size();
        if (n != r.entries_.size())
            return false;
        if (n == 0)
            return true;
        for (unsigned i=0; i<n; ++i)
            if (getWeight(i) != r.getWeight(i))
                return false;
        for (unsigned i=0; i<n; ++i)
            if (*entries_[i] != *r.entries_[i])
                return false;
        return true;
    }

    double DistributionMixND::density(const double* x, unsigned i_dim) const
    {
       if (i_dim != dim_) throw std::invalid_argument(
            "In npstat::DistributionMixND::density: "
            "incompatible input point dimensionality");
        assert(x);
        if (!isNormalized_)
            (const_cast<DistributionMixND*>(this))->normalize();
        const unsigned nentries = entries_.size();
        long double sum = 0.0L;
        for (unsigned i=0; i<nentries; ++i)
            sum += weights_[i]*entries_[i]->density(x, i_dim);
        return sum/wsum_;
    }

    void DistributionMixND::unitMap(const double* /* rnd */,
                                    unsigned /* bufLen */,
                                    double* /* x */) const
    {
        throw std::runtime_error(
            "Function npstat::DistributionMixND::unitMap is not implemented");
    }

    bool DistributionMixND::write(std::ostream& os) const
    {
        gs::write_pod(os, dim_);
        gs::write_pod_vector(os, weights_);
        gs::write_pod_vector(os, weightCdf_);
        gs::write_pod(os, wsum_);
        unsigned char c = isNormalized_;
        gs::write_pod(os, c);
        return gs::write_item(os, entries_);
    }

    DistributionMixND* DistributionMixND::read(
        const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(
            gs::ClassId::makeId<DistributionMixND>());
        current.ensureSameId(id);

        unsigned dim = 0;
        gs::read_pod(in, &dim);
        if (in.fail()) throw gs::IOReadFailure(
            "In npstat::DistributionMixND::read: input stream failure");
        CPP11_auto_ptr<DistributionMixND> p(new DistributionMixND(dim));
        gs::read_pod_vector(in, &p->weights_);
        gs::read_pod_vector(in, &p->weightCdf_);
        gs::read_pod(in, &p->wsum_);
        unsigned char c;
        gs::read_pod(in, &c);
        p->isNormalized_ = c;
        gs::restore_item(in, &p->entries_);
        if (in.fail()) throw gs::IOReadFailure(
            "In npstat::DistributionMixND::read: input stream failure");
        return p.release();
    }

    unsigned DistributionMixND::random(AbsRandomGenerator& g,
                                       double* x, const unsigned lenX) const
    {
        if (lenX != dim_) throw std::invalid_argument(
            "In npstat::DistributionMixND::random: incompatible buffer size");
        assert(x);
        const unsigned gdim(g.dim());
        double wrand = 0.0;
        if (gdim == 1U)
            wrand = g();
        else if (gdim == dim_)
        {
            throw std::runtime_error(
                "In npstat::DistributionMixND::random: "
                "use of multidimensional random number generators is not yet supported");
        }
        else
            throw std::invalid_argument(
                "In npstat::DistributionMixND::random: "
                "incompatible random number generator dimensionality");
        if (!isNormalized_)
            (const_cast<DistributionMixND*>(this))->normalize();
        const unsigned nentries = entries_.size();
        const unsigned bin = quantileBinFromCdf(&weightCdf_[0], nentries, wrand);
        assert(bin < nentries);
        return 1U + entries_[bin]->random(g, x, lenX);
    }
}
