#include <cmath>
#include <cassert>
#include <stdexcept>

#include "geners/GenericIO.hh"
#include "geners/vectorIO.hh"

#include "npstat/nm/SpecialFunctions.hh"
#include "npstat/rng/convertToSphericalRandom.hh"

#include "npstat/stat/EllipticalDistribution.hh"
#include "npstat/stat/Distributions1D.hh"
#include "npstat/stat/distributionReadError.hh"

static double calculateGNorm(const npstat::AbsDistribution1D& gDensity,
                             const npstat::AbsDistribution1D& hDensity,
                             const unsigned dim)
{
    const double gmax = gDensity.quantile(1.0);
    const double hmax = hDensity.quantile(1.0);
    assert(gmax > 0.0 && hmax > 0.0);

    double rsq = 1.0;
    while (rsq >= gmax || rsq >= hmax)
        rsq /= 2.0;

    const double gd = gDensity.density(rsq);
    assert(gd > 0.0);
    const double hd = hDensity.density(rsq);
    assert(hd > 0.0);

    const double dover2 = dim/2.0;
    const double piu = M_PI*rsq;
    return hd*npstat::Gamma(dover2)*rsq/gd/pow(piu,dover2);
}

namespace npstat {
    EllipticalDistribution::EllipticalDistribution(
        const double* location, const unsigned i_dim,
        const Matrix<double>& transformationMatrix,
        const AbsDistribution1D& gDensity,
        const AbsDistribution1D& hDensity)
        : AbsDistributionND(i_dim),
          mu_(location, location+i_dim),
          A_(transformationMatrix),
          g_(0),
          h_(0),
          det_(0.0),
          gNorm_(0.0),
          buf_(i_dim)
    {
        assert(location);
        assert(i_dim);
        initialize(gDensity, hDensity);
    }

    void EllipticalDistribution::initialize(
        const AbsDistribution1D& gDensity,
        const AbsDistribution1D& hDensity)
    {
        if (!A_.isSquare()) throw std::invalid_argument(
            "In npstat::EllipticalDistribution::initialize: "
            "transformation matrix is not square");
        if (A_.nRows() != dim_) throw std::invalid_argument(
            "In npstat::EllipticalDistribution::initialize: "
            "incompatible dimensions of the shift and "
            "transformation matrix");
        if (gDensity.quantile(0.0) != 0.0) throw std::invalid_argument(
            "In npstat::EllipticalDistribution::initialize: "
            "inappropriate g distribution");
        if (hDensity.quantile(0.0) != 0.0) throw std::invalid_argument(
            "In npstat::EllipticalDistribution::initialize: "
            "inappropriate h distribution");

        det_ = std::abs(A_.det());
        assert(det_);

        const Matrix<double>& covmat = A_.timesT();
        InvCovmat_ = covmat.symPDInv();

        g_ = gDensity.clone();
        h_ = hDensity.clone();
        gNorm_ = calculateGNorm(gDensity, hDensity, dim_);
    }

    EllipticalDistribution::EllipticalDistribution(
        const EllipticalDistribution& r)
        : AbsDistributionND(r.dim_),
          mu_(r.mu_),
          A_(r.A_),
          InvCovmat_(r.InvCovmat_),
          g_(r.g_->clone()),
          h_(r.h_->clone()),
          det_(r.det_),
          gNorm_(r.gNorm_),
          buf_(r.dim_)
    {
    }

    EllipticalDistribution& EllipticalDistribution::operator=(
        const EllipticalDistribution& r)
    {
        if (&r != this)
        {
            AbsDistributionND::operator=(r);
            cleanup();
            mu_ = r.mu_;
            A_ = r.A_;
            InvCovmat_ = r.InvCovmat_;
            g_ = r.g_->clone();
            h_ = r.h_->clone();
            det_ = r.det_;
            gNorm_ = r.gNorm_;
            buf_.resize(r.dim_);
        }
        return *this;
    }

    bool EllipticalDistribution::isEqual(const AbsDistributionND& other) const
    {
        const EllipticalDistribution& r = 
            static_cast<const EllipticalDistribution&>(other);
        return mu_  == r.mu_ &&
                A_  == r.A_ &&
               *g_  == *r.g_ &&
               *h_  == *r.h_;
    }

    void EllipticalDistribution::cleanup()
    {
        A_.uninitialize();
        InvCovmat_.uninitialize();
        delete g_; g_ = 0;
        delete h_; h_ = 0;
    }

    double EllipticalDistribution::chiSquare(
        const double* x, const unsigned i_dim) const
    {
        assert(x);
        assert(i_dim == dim_);

        double* delta = &buf_[0];
        const double* mu = &mu_[0];
        for (unsigned i=0; i<i_dim; ++i)
            *delta++ = *x++ - *mu++;

        return InvCovmat_.bilinear(&buf_[0], i_dim);
    }

    double EllipticalDistribution::density(
        const double* x, const unsigned i_dim) const
    {
        return g_->density(chiSquare(x, i_dim))*gNorm_/det_;
    }

    void EllipticalDistribution::unitMap(
        const double* rnd, const unsigned bufLen, double* x) const
    {
        if (bufLen)
        {
            if (bufLen < dim_) throw std::invalid_argument(
                "In npstat::EllipticalDistribution::unitMap: insufficient "
                "length of the random buffer");
            if (bufLen > dim_) throw std::out_of_range(
                "In npstat::EllipticalDistribution::unitMap: "
                "input point dimensionality is out of range");
            assert(rnd);
            assert(x);

            double *direction = &buf_[0];
            const double rn = convertToSphericalRandom(rnd, dim_, direction);
            const double u = h_->quantile(rn);
            const double r = sqrt(u);
            for (unsigned i=0; i<dim_; ++i)
                direction[i] *= r;
            A_.timesVector(direction, dim_, x, dim_);
            for (unsigned i=0; i<dim_; ++i)
                x[i] += mu_[i];
        }
    }

    unsigned EllipticalDistribution::random(AbsRandomGenerator& g,
                                            double* x, const unsigned lenX) const
    {
        assert(x);
        assert(lenX >= dim_);

        Gauss1D normal(0.0, 1.0);
        unsigned callCount = 0U;
        double *direction = &buf_[0];
        long double sumsql = 0.0L;
        for (unsigned i=0; i<dim_; ++i)
        {
            callCount += normal.random(g, direction+i);
            sumsql += direction[i]*direction[i];
        }
        const double sumsq = sumsql;
        const double len = sqrt(sumsq);
        if (len)
        {
            for (unsigned i=0; i<dim_; ++i)
                direction[i] /= len;
        }
        else
        {
            direction[0] = 1.0;
            for (unsigned i=1U; i<dim_; ++i)
                direction[i] = 0.0;
        }
        const double u = h_->quantile(g());
        ++callCount;
        const double r = sqrt(u);
        for (unsigned i=0; i<dim_; ++i)
            direction[i] *= r;
        A_.timesVector(direction, dim_, x, dim_);
        for (unsigned i=0; i<dim_; ++i)
            x[i] += mu_[i];
        return callCount;
    }

    bool EllipticalDistribution::write(std::ostream& os) const
    {
        gs::write_pod_vector(os, mu_);
        gs::write_item(os, A_);
        return gs::write_obj(os, *g_) && gs::write_obj(os, *h_) && !os.fail();
    }

    EllipticalDistribution* EllipticalDistribution::read(const gs::ClassId& id,
                                                         std::istream& is)
    {
        static const gs::ClassId current(gs::ClassId::makeId<EllipticalDistribution>());
        current.ensureSameId(id);

        std::vector<double> mu;
        gs::read_pod_vector(is, &mu);
        Matrix<double> transform;
        gs::restore_item(is, &transform);
        CPP11_auto_ptr<AbsDistribution1D> g = gs::read_obj<AbsDistribution1D>(is);
        CPP11_auto_ptr<AbsDistribution1D> h = gs::read_obj<AbsDistribution1D>(is);
        if (is.fail())
        {
            distributionReadError(is, classname());
            return 0;
        }
        else
            return new EllipticalDistribution(&mu[0], mu.size(), transform, *g, *h);
    }
}
