#include <cmath>
#include <cassert>
#include <stdexcept>

#include "npstat/stat/ExpTiltedDistribution1D.hh"

namespace npstat {
    ExpTiltedDistribution1D::ExpTiltedDistribution1D()
        : cgf_(0), d_(0), s_(0.0), Ks_(0.0) {}

    ExpTiltedDistribution1D::ExpTiltedDistribution1D(
        const AbsCGF1D& cgf,
        const AbsDistribution1D& distro,
        const double s_tilt)
        : cgf_(cgf.clone()), d_(0), s_(0.0), Ks_(0.0)
    {
        setTilt(s_tilt);
        d_ = distro.clone();
    }

    ExpTiltedDistribution1D::ExpTiltedDistribution1D(
        const ExpTiltedDistribution1D& r)
        : cgf_(r.cgf_->clone()), d_(r.d_->clone()),
          s_(r.s_), Ks_(r.Ks_)
    {
    }

    ExpTiltedDistribution1D& ExpTiltedDistribution1D::operator=(
        const ExpTiltedDistribution1D& r)
    {
        if (this != &r)
        {
            AbsDistribution1D::operator=(r);
            cleanup();
            cgf_ = r.cgf_->clone();
            d_ = r.d_->clone();
            s_ = r.s_;
            Ks_ = r.Ks_;
        }
        return *this;
    }

    void ExpTiltedDistribution1D::cleanup()
    {
        delete d_; d_ = 0;
        delete cgf_; cgf_ = 0;
        s_ = 0.0;
        Ks_ = 0.0;
    }

    bool ExpTiltedDistribution1D::isEqual(const AbsDistribution1D& o) const
    {
        const ExpTiltedDistribution1D& r = 
            static_cast<const ExpTiltedDistribution1D&>(o);
        return *cgf_ == *r.cgf_ && *d_ == *r.d_ && s_ == r.s_ && Ks_ == r.Ks_;
    }

    bool ExpTiltedDistribution1D::write(std::ostream& /* os */) const
    {
        throw std::runtime_error("In npstat::ExpTiltedDistribution1D::write: "
                                 "method not implemented");
        return false;
    }

    ExpTiltedDistribution1D* ExpTiltedDistribution1D::read(
        const gs::ClassId& id, std::istream& /* is */)
    {
        static const gs::ClassId myClassId(
            gs::ClassId::makeId<ExpTiltedDistribution1D>());
        myClassId.ensureSameId(id);

        throw std::runtime_error("In npstat::ExpTiltedDistribution1D::read: "
                                 "method not implemented");
        return 0;
    }

    double ExpTiltedDistribution1D::density(const double x) const
    {
        return d_->density(x)*exp(s_*x - Ks_);
    }

    double ExpTiltedDistribution1D::cdf(const double /* x */) const
    {
        throw std::runtime_error("In npstat::ExpTiltedDistribution1D::cdf: "
                                 "method not implemented");
        return 0.0;
    }

    double ExpTiltedDistribution1D::exceedance(const double /* x */) const
    {
        throw std::runtime_error("In npstat::ExpTiltedDistribution1D::exceedance: "
                                 "method not implemented");
        return 0.0;
    }

    double ExpTiltedDistribution1D::quantile(const double /* r1 */) const
    {
        throw std::runtime_error("In npstat::ExpTiltedDistribution1D::quantile: "
                                 "method not implemented");
        return 0.0;
    }

    void ExpTiltedDistribution1D::setTilt(const double s)
    {
        if (!(cgf_->smin() < s && s < cgf_->smax())) throw std::invalid_argument(
            "In npstat::ExpTiltedDistribution1D::setTilt: "
            "tilt value outside of CGF support");
        s_ = s;
        Ks_ = (*cgf_)(s);
    }
}
