#include <cassert>
#include <stdexcept>
#include <algorithm>

#include "npstat/nm/Interval.hh"
#include "npstat/nm/findRootUsingBisections.hh"

#include "npstat/stat/FoldedDistribution1D.hh"
#include "npstat/stat/distributionReadError.hh"

#include "geners/binaryIO.hh"
#include "geners/CPP11_auto_ptr.hh"

namespace npstat {
    FoldedDistribution1D::FoldedDistribution1D(
        const AbsDistribution1D& distro, double xmin, double xmax,
        const unsigned maxFolds)
        : distro_(distro.clone()), maxFoldings_(maxFolds)
    {
        if (xmin > xmax)
            std::swap(xmin, xmax);
        if (xmin == xmax)
            throw std::invalid_argument(
                "In npstat::FoldedDistribution1D constructor: "
                "can not use the same value for the support limits");
        leftBoundary_ = xmin;
        rightBoundary_ = xmax;
        lo_ = distro.quantile(0.0);
        hi_ = distro.quantile(1.0);
        Interval<double> newSupport(xmin, xmax);
        Interval<double> oldSupport(lo_, hi_);
        if (!newSupport.overlapLength(oldSupport))
            throw std::invalid_argument(
                "In npstat::FoldedDistribution1D constructor: "
                "new and old supports do not overlap");

        const double hiReflected = 2.0*rightBoundary_ - hi_;
        xmin_ = std::max(std::min(lo_, hiReflected), leftBoundary_);

        const double loReflected = 2.0*leftBoundary_ - lo_;
        xmax_ = std::min(std::max(hi_, loReflected), rightBoundary_);

        reallyFolded_ = !(leftBoundary_ <= lo_ && rightBoundary_ >= hi_);
        if (!reallyFolded_)
        {
            assert(xmin_ == lo_);
            assert(xmax_ == hi_);
        }

        // Don't forget to comment out the following
        // reallyFolded_ = true;
    }

    FoldedDistribution1D::FoldedDistribution1D()
        : distro_(0), leftBoundary_(0.0), rightBoundary_(0.0),
          lo_(0.0), hi_(0.0), xmin_(0.0), xmax_(0.0),
          maxFoldings_(0), reallyFolded_(false)
    {
    }

    FoldedDistribution1D::FoldedDistribution1D(
        const FoldedDistribution1D& r)
        : distro_(r.distro_->clone()),
          leftBoundary_(r.leftBoundary_),
          rightBoundary_(r.rightBoundary_),
          lo_(r.lo_),
          hi_(r.hi_),
          xmin_(r.xmin_),
          xmax_(r.xmax_),
          maxFoldings_(r.maxFoldings_),
          reallyFolded_(r.reallyFolded_)
    {
    }

    FoldedDistribution1D& FoldedDistribution1D::operator=(
        const FoldedDistribution1D& r)
    {
        if (&r != this)
        {
            AbsDistribution1D* p = r.distro_->clone();
            delete distro_;
            distro_ = p;
            leftBoundary_ = r.leftBoundary_;
            rightBoundary_ = r.rightBoundary_;
            lo_ = r.lo_;
            hi_ = r.hi_;
            xmin_ = r.xmin_;
            xmax_ = r.xmax_;
            maxFoldings_ = r.maxFoldings_;
            reallyFolded_ = r.reallyFolded_;
        }
        return *this;
    }

    FoldedDistribution1D::~FoldedDistribution1D()
    {
        delete distro_;
    }

    bool FoldedDistribution1D::isEqual(const AbsDistribution1D& other) const
    {
        const FoldedDistribution1D& r = 
            static_cast<const FoldedDistribution1D&>(other);
        return *distro_ == *r.distro_ &&
            leftBoundary_ == r.leftBoundary_ &&
            rightBoundary_ == r.rightBoundary_ &&
            lo_ == r.lo_ &&
            hi_ == r.hi_ &&
            xmin_ == r.xmin_ &&
            xmax_ == r.xmax_ &&
            maxFoldings_ == r.maxFoldings_ &&
            reallyFolded_ == r.reallyFolded_;
    }

    double FoldedDistribution1D::density(const double x) const
    {
        if (x < xmin_ || x > xmax_)
            return 0.0;
        else if (reallyFolded_)
        {
            long double sum = distro_->density(x);

            const long double pStep = 2.0*(rightBoundary_ - x);
            const long double nStep = 2.0*(x - leftBoundary_);

            // Fold the upper tail
            long double coord = x;
            const long double ldhi = hi_;
            bool reachedBound = false;
            for (unsigned ifold=0; ifold<=maxFoldings_; ++ifold)
            {
                coord += pStep;
                if (coord > ldhi)
                {
                    reachedBound = true;
                    break;
                }
                sum += distro_->density(coord);
                coord += nStep;
                if (coord > ldhi)
                {
                    reachedBound = true;
                    break;
                }
                sum += distro_->density(coord);                
            }
            if (!reachedBound) throw std::runtime_error(
                "In FoldedDistribution1D::density: "
                "maximum number of folds exceeded for the upper tail");

            // Fold the lower tail
            coord = x;
            const long double ldlo = lo_;
            reachedBound = false;
            for (unsigned ifold=0; ifold<=maxFoldings_; ++ifold)
            {
                coord -= nStep;
                if (coord < ldlo)
                {
                    reachedBound = true;
                    break;
                }    
                sum += distro_->density(coord);
                coord -= pStep;
                if (coord < ldlo)
                {
                    reachedBound = true;
                    break;
                }
                sum += distro_->density(coord);
            }
            if (!reachedBound) throw std::runtime_error(
                "In FoldedDistribution1D::density: "
                "maximum number of folds exceeded for the lower tail");

            return sum;
        }
        else
            return distro_->density(x);
    }

    double FoldedDistribution1D::cdf(const double x) const
    {
        if (x <= xmin_)
            return 0.0;
        else if (x >= xmax_)
            return 1.0;
        else if (reallyFolded_)
        {
            long double cdfsum = 0.0L;

            const long double pStep = 2.0*(rightBoundary_ - x);
            const long double nStep = 2.0*(x - leftBoundary_);

            // Fold the upper tail
            long double coord = x;
            const long double ldhi = hi_;
            bool reachedBound = false;
            for (unsigned ifold=0; ifold<=maxFoldings_; ++ifold)
            {
                coord += pStep;
                if (coord > ldhi)
                {
                    reachedBound = true;
                    break;
                }
                const long double oldCdf = distro_->cdf(coord);
                coord += nStep;
                const long double cdfx = distro_->cdf(coord);
                cdfsum += (cdfx - oldCdf);
                if (coord > ldhi)
                {
                    reachedBound = true;
                    break;
                }
            }
            if (!reachedBound) throw std::runtime_error(
                "In FoldedDistribution1D::cdf: "
                "maximum number of folds exceeded for the upper tail");

            // Fold the lower tail
            coord = x;
            const long double ldlo = lo_;
            reachedBound = false;
            for (unsigned ifold=0; ifold<=maxFoldings_; ++ifold)
            {
                const long double oldCdf = distro_->cdf(coord);
                coord -= nStep;
                const long double cdfx = distro_->cdf(coord);
                cdfsum += (oldCdf - cdfx);
                if (coord < ldlo)
                {
                    reachedBound = true;
                    break;
                }    
                coord -= pStep;
                if (coord < ldlo)
                {
                    reachedBound = true;
                    break;
                }
            }
            if (!reachedBound) throw std::runtime_error(
                "In FoldedDistribution1D::cdf: "
                "maximum number of folds exceeded for the lower tail");

            return cdfsum;
        }
        else
            return distro_->cdf(x);
    }

    double FoldedDistribution1D::exceedance(const double x) const
    {
        if (x <= xmin_)
            return 1.0;
        else if (x >= xmax_)
            return 0.0;
        else if (reallyFolded_)
        {
            long double exsum = 0.0L;

            const long double pStep = 2.0*(rightBoundary_ - x);
            const long double nStep = 2.0*(x - leftBoundary_);

            // Fold the upper tail
            long double coord = x;
            const long double ldhi = hi_;
            bool reachedBound = false;
            for (unsigned ifold=0; ifold<=maxFoldings_; ++ifold)
            {
                const long double oldEx = distro_->exceedance(coord);
                coord += pStep;
                const long double ex = distro_->exceedance(coord);
                exsum += (oldEx - ex);
                if (coord > ldhi)
                {
                    reachedBound = true;
                    break;
                }
                coord += nStep;
                if (coord > ldhi)
                {
                    reachedBound = true;
                    break;
                }
            }
            if (!reachedBound) throw std::runtime_error(
                "In FoldedDistribution1D::exceedance: "
                "maximum number of folds exceeded for the upper tail");

            // Fold the lower tail
            coord = x;
            const long double ldlo = lo_;
            reachedBound = false;
            for (unsigned ifold=0; ifold<=maxFoldings_; ++ifold)
            {
                coord -= nStep;
                if (coord < ldlo)
                {
                    reachedBound = true;
                    break;
                }    
                const long double oldEx = distro_->exceedance(coord);
                coord -= pStep;
                const long double ex = distro_->exceedance(coord);
                exsum += (ex - oldEx);
                if (coord < ldlo)
                {
                    reachedBound = true;
                    break;
                }
            }
            if (!reachedBound) throw std::runtime_error(
                "In FoldedDistribution1D::exceedance: "
                "maximum number of folds exceeded for the lower tail");

            return exsum;
        }
        else
            return distro_->exceedance(x);
    }

    double FoldedDistribution1D::quantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::FoldedDistribution1D::quantile: "
            "cdf argument outside of [0, 1] interval");
        if (r1 == 0.0)
            return xmin_;
        else if (r1 == 1.0)
            return xmax_;
        else if (reallyFolded_)
        {
            double x;
            const bool status = findRootUsingBisections(
                CdfFunctor1D(*this), r1, xmin_, xmax_, 4.0*DBL_EPSILON, &x);
            assert(status);
            return x;
        }
        else
            return distro_->quantile(r1);
    }

    bool FoldedDistribution1D::write(std::ostream& os) const
    {
        const bool status = distro_->classId().write(os) && distro_->write(os);
        if (status)
        {
            gs::write_pod(os, leftBoundary_);
            gs::write_pod(os, rightBoundary_);
            gs::write_pod(os, lo_);
            gs::write_pod(os, hi_);
            gs::write_pod(os, xmin_);
            gs::write_pod(os, xmax_);
            gs::write_pod(os, maxFoldings_);
            gs::write_pod(os, reallyFolded_);
        }
        return status && !os.fail();
    }

    FoldedDistribution1D* FoldedDistribution1D::read(
        const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(
            gs::ClassId::makeId<FoldedDistribution1D>());
        current.ensureSameId(id);

        gs::ClassId clid(in, 1);
        CPP11_auto_ptr<AbsDistribution1D> d(AbsDistribution1D::read(clid, in));
        FoldedDistribution1D* distro = new FoldedDistribution1D();

        gs::read_pod(in, &distro->leftBoundary_);
        gs::read_pod(in, &distro->rightBoundary_);
        gs::read_pod(in, &distro->lo_);
        gs::read_pod(in, &distro->hi_);
        gs::read_pod(in, &distro->xmin_);
        gs::read_pod(in, &distro->xmax_);
        gs::read_pod(in, &distro->maxFoldings_);
        gs::read_pod(in, &distro->reallyFolded_);

        if (in.fail())
        {
            delete distro;
            distro = 0;
            distributionReadError(in, classname());
        }
        else
            distro->distro_ = d.release();
        return distro;
    }
}
