#ifndef NPSTAT_FOLDEDDISTRIBUTION1D_HH_
#define NPSTAT_FOLDEDDISTRIBUTION1D_HH_

/*!
// \file FoldedDistribution1D.hh
//
// \brief 1-d continuous statistical distributions folded on some interval
//
// Author: I. Volobouev
//
// May 2022
*/

#include "npstat/stat/AbsDistribution1D.hh"

namespace npstat {
    class FoldedDistribution1D : public AbsDistribution1D
    {
    public:
        /**
        // Constructor arguments are as follows:
        //
        //  distro     -- Distribution whose support we want to fold.
        //
        //  xmin, xmax -- New limits for the support.
        //
        //  maxFolds   -- Maximum number of folds allowed. If this number
        //                is exceeded, an exception will be thrown.
        */
        FoldedDistribution1D(const AbsDistribution1D& distro,
                             double xmin, double xmax,
                             unsigned maxFolds = 10000U);

        FoldedDistribution1D(const FoldedDistribution1D&);
        FoldedDistribution1D& operator=(const FoldedDistribution1D&);

        inline virtual FoldedDistribution1D* clone() const
            {return new FoldedDistribution1D(*this);}

        virtual ~FoldedDistribution1D();

        /** Distribution density */
        virtual double density(double x) const;

        /** Cumulative distribution function */
        virtual double cdf(double x) const;

        /** 1 - cdf, avoiding subtractive cancellation */
        virtual double exceedance(double x) const;

        /** The quantile function */
        virtual double quantile(double x) const;

        //@{
        /** Method needed for I/O */
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;
        //@}

        static inline const char* classname()
            {return "npstat::FoldedDistribution1D";}
        static inline unsigned version() {return 1;}
        static FoldedDistribution1D* read(
            const gs::ClassId& id, std::istream&);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        FoldedDistribution1D();

        AbsDistribution1D* distro_;
        double leftBoundary_;
        double rightBoundary_;
        double lo_;
        double hi_;
        double xmin_;
        double xmax_;
        unsigned maxFoldings_;
        bool reallyFolded_;
    };
}

#endif // NPSTAT_FOLDEDDISTRIBUTION1D_HH_
