#include <cassert>
#include <stdexcept>

#include "geners/GenericIO.hh"

#include "npstat/stat/InterpolatedDistro1D1P.hh"
#include "npstat/stat/InterpolatedDistribution1D.hh"
#include "npstat/stat/VerticallyInterpolatedDistribution1D.hh"

namespace npstat {
    InterpolatedDistro1D1P::InterpolatedDistro1D1P(
        const GridAxis& axis,
        const std::vector<CPP11_shared_ptr<const AbsDistribution1D> >& distros,
        const double initialParameterValue, const bool interpolateVerticall)
        : axis_(axis),
          distros_(distros),
          interpolator_(0),
          param_(0.0),
          vertical_(!interpolateVerticall),
          parameterSet_(false),
          haveAllDistros_(false)
    {
        const unsigned nDistros = axis_.nCoords();
        if (nDistros < 2U) throw std::invalid_argument(
            "In npstat::InterpolatedDistro1D1P constructor: "
            "not enough grid cells to interpolate");
        if (distros.size() != nDistros) throw std::invalid_argument(
            "In npstat::InterpolatedDistro1D1P constructor: "
            "number of provided distributions must be equal to the grid size");
        interpolateVertically(interpolateVerticall);
        setParameter(initialParameterValue);
    }

    InterpolatedDistro1D1P::InterpolatedDistro1D1P(
        const GridAxis& axis, const bool interpolateVerticall)
        : axis_(axis),
          distros_(axis.nCoords()),
          interpolator_(0),
          param_(0.0),
          vertical_(!interpolateVerticall),
          parameterSet_(false),
          haveAllDistros_(false)
    {
        if (axis.nCoords() < 2U) throw std::invalid_argument(
            "In npstat::InterpolatedDistro1D1P constructor: "
            "not enough grid cells to interpolate");
        interpolateVertically(interpolateVerticall);
    }

    void InterpolatedDistro1D1P::checkReady() const
    {
        if (!parameterSet_) throw std::runtime_error(
            "In npstat::InterpolatedDistro1D1P::checkReady: "
            "interpolation parameter is not set");
    }

    void InterpolatedDistro1D1P::checkDistros()
    {
        if (!haveAllDistros_)
        {
            haveAllDistros_ = true;
            const unsigned nDistros = distros_.size();
            for (unsigned i=0; i<nDistros; ++i)
                if (!distros_[i].get())
                {
                    haveAllDistros_ = false;
                    break;
                }
        }
    }

    InterpolatedDistro1D1P::InterpolatedDistro1D1P(
        const InterpolatedDistro1D1P& r)
        : axis_(r.axis_),
          distros_(r.distros_),
          interpolator_(0),
          param_(0.0),
          vertical_(!r.vertical_),
          parameterSet_(false),
          haveAllDistros_(r.haveAllDistros_)
    {
        assert(axis_.nCoords() == distros_.size());
        interpolateVertically(r.vertical_);
        if (r.parameterSet_)
            setParameter(r.param_);
    }

    InterpolatedDistro1D1P& InterpolatedDistro1D1P::operator=(
        const InterpolatedDistro1D1P& r)
    {
        if (this != &r)
        {
            axis_ = r.axis_;
            distros_ = r.distros_;
            haveAllDistros_ = r.haveAllDistros_;
            assert(axis_.nCoords() == distros_.size());
            param_ = 0.0;
            parameterSet_ = false;
            if (vertical_ != r.vertical_)
                interpolateVertically(r.vertical_);
            else
                interpolator_->clear();
            if (r.parameterSet_)
                setParameter(r.param_);
        }
        return *this;
    }

    void InterpolatedDistro1D1P::setParameter(const double value)
    {
        checkDistros();
        if (!haveAllDistros_)
            throw std::runtime_error("In npstat::InterpolatedDistro1D1P::setParameter: "
                                     "interpolation table construction is not finished");
        parameterSet_ = false;
        param_ = value;
        if (!interpolator_)
            rebuildInterpolator();
        interpolator_->clear();
        interpolator_->normalizeAutomatically(false);
        const std::pair<unsigned,double>& intv = axis_.getInterval(param_);
        interpolator_->add(*distros_[intv.first], intv.second);
        interpolator_->add(*distros_[intv.first+1], 1.0-intv.second);
        interpolator_->normalizeAutomatically(true);
        parameterSet_ = true;
    }

    void InterpolatedDistro1D1P::rebuildInterpolator()
    {
        delete interpolator_;
        interpolator_ = 0;
        if (vertical_)
            interpolator_ = new VerticallyInterpolatedDistribution1D(2U);
        else
            interpolator_ = new InterpolatedDistribution1D(2U);
    }

    void InterpolatedDistro1D1P::interpolateVertically(const bool b)
    {
        if (b != vertical_)
        {
            vertical_ = b;
            rebuildInterpolator();
            if (parameterSet_)
                setParameter(param_);
        }
    }

    bool InterpolatedDistro1D1P::isEqual(const AbsDistribution1D& otherBase) const
    {
        const InterpolatedDistro1D1P& r =
            static_cast<const InterpolatedDistro1D1P&>(otherBase);
        checkReady();
        r.checkReady();
        if (param_ != r.param_)
            return false;
        if (vertical_ != r.vertical_)
            return false;
        if (axis_ != r.axis_)
            return false;
        const unsigned sz = distros_.size();
        if (sz != r.distros_.size())
            return false;
        for (unsigned i=0; i<sz; ++i)
            if (*distros_[i] != *r.distros_[i])
                return false;
        return true;
    }

    bool InterpolatedDistro1D1P::write(std::ostream& os) const
    {
        checkReady();
        const bool status = gs::write_obj(os, axis_) &&
                            gs::write_item(os, distros_);
        if (status)
        {
            gs::write_pod(os, param_);
            unsigned char c = vertical_;
            gs::write_pod(os, c);
        }
        return status && !os.fail();
    }

    InterpolatedDistro1D1P* InterpolatedDistro1D1P::read(const gs::ClassId& id,
                                                         std::istream& in)
    {
        static const gs::ClassId current(
            gs::ClassId::makeId<InterpolatedDistro1D1P>());
        current.ensureSameId(id);

        CPP11_auto_ptr<GridAxis> paxis = gs::read_obj<GridAxis>(in);
        std::vector<CPP11_shared_ptr<const AbsDistribution1D> > distros;
        gs::restore_item(in, &distros);
        double p;
        gs::read_pod(in, &p);
        unsigned char c;
        gs::read_pod(in, &c);

        if (in.fail()) throw gs::IOReadFailure(
            "In npstat::InterpolatedDistro1D1P::read: input stream failure");
        return new InterpolatedDistro1D1P(*paxis, distros, p, c);
    }

    void InterpolatedDistro1D1P::setGridDistro(const unsigned cell,
                                               const AbsDistribution1D& distro)
    {
        CPP11_shared_ptr<const AbsDistribution1D> ptr(distro.clone());
        distros_.at(cell) = ptr;
    }
}
