#ifndef NPSTAT_LORPE1DFIXEDDEGREECVSCANNER_HH_
#define NPSTAT_LORPE1DFIXEDDEGREECVSCANNER_HH_

/*!
// \file LOrPE1DFixedDegreeCVScanner.hh
//
// \brief Optimization of LOrPE bandwidth choice by cross-validation
//        by simply scanning a predefined set of bandwidth values
//
// Author: I. Volobouev
//
// August 2022
*/

#include <cmath>
#include <cassert>
#include <vector>
#include <stdexcept>

#include "npstat/nm/isMonotonous.hh"
#include "npstat/nm/ScanExtremum1D.hh"

#include "npstat/stat/LOrPE1DCVResult.hh"

namespace npstat {
    class LOrPE1DFixedDegreeCVScanner
    {
    public:
        inline LOrPE1DFixedDegreeCVScanner(const double degree,
                                           const std::vector<double>& bandwidthFactors)
            : bandwidthFactors_(bandwidthFactors),
              logBw_(bandwidthFactors.size()),
              fixedDegree_(degree)
        {
            if (fixedDegree_ < 0.0) throw std::invalid_argument(
                "In npstat::LOrPE1DFixedDegreeCVScanner constructor: "
                "degree argument must be non-negative");
            const unsigned nscan = bandwidthFactors_.size();
            if (nscan < 2U) throw std::invalid_argument(
                "In npstat::LOrPE1DFixedDegreeCVScanner constructor: "
                "insufficient number of bandwidth factors");
            if (!isStrictlyIncreasing(bandwidthFactors_.begin(), bandwidthFactors_.end()))
                throw std::invalid_argument(
                    "In npstat::LOrPE1DFixedDegreeCVScanner constructor: "
                    "sequence of bandwidth factors must be strictly increasing");
            for (unsigned i=0; i<nscan; ++i)
            {
                if (bandwidthFactors_[i] <= 0.0) throw std::invalid_argument(
                    "In npstat::LOrPE1DFixedDegreeCVScanner constructor: "
                    "all bandwidth factors must be positive");
                logBw_[i] = std::log(bandwidthFactors_[i]);
            }
        }

        inline unsigned nBwFactors() const {return bandwidthFactors_.size();}
        inline double getBwFactor(const unsigned i) const {return bandwidthFactors_.at(i);}

        template<class Lorpe>
        inline LOrPE1DCVResult crossValidate(Lorpe& lorpe) const
        {
            const unsigned nscan = bandwidthFactors_.size();
            std::vector<double> cvValues(nscan);
            for (unsigned i=0; i<nscan; ++i)
                cvValues[i] = lorpe(fixedDegree_, bandwidthFactors_[i]);
            const ScanExtremum1D& scanMax = findScanMaximum1D(&logBw_[0], nscan,
                                                              &cvValues[0], nscan);
            return LOrPE1DCVResult(
                fixedDegree_, std::exp(scanMax.location()), scanMax.value(),
                true, scanMax.isOnTheBoundary());
        }

    private:
        std::vector<double> bandwidthFactors_;
        std::vector<double> logBw_;
        double fixedDegree_;
    };
}

#endif // NPSTAT_LORPE1DFIXEDDEGREECVSCANNER_HH_
