#include <cmath>
#include <string>

#include "npstat/nm/ClassicalOrthoPolys1D.hh"

#include "npstat/stat/Distributions1D.hh"
#include "npstat/stat/TruncatedDistribution1D.hh"
#include "npstat/stat/FoldedDistribution1D.hh"
#include "npstat/stat/BoundaryMethod.hh"
#include "npstat/stat/DensityOrthoPoly1D.hh"

#include "npstat/stat/LOrPE1DSymbetaKernel.hh"

namespace npstat {
    LOrPE1DSymbetaKernel::LOrPE1DSymbetaKernel(
        const int symbetaPower, const double i_filterDegree,
        const double i_leftBoundary, const double i_rightBoundary,
        const BoundaryHandling& bh, const bool memoizeBoundaryPolys)
        : power_(symbetaPower),
          filterDegree_(i_filterDegree),
          leftBoundary_(i_leftBoundary),
          rightBoundary_(i_rightBoundary),
          lastShrinkage_(1.0),
          norm_(1.0),
          kernelRange_(1.0),
          lastCenterValue_(-DBL_MAX),
          bh_(bh),
          poly_(0),
          memoizingBoundaryPolys_(memoizeBoundaryPolys),
          usingBoundaryCorr_(false)
    {
        if (i_filterDegree < 0.0) throw std::invalid_argument(
            "In npstat::LOrPE1DSymbetaKernel constructor: "
            "filter degree can not be negative");
        if (!(leftBoundary_ < rightBoundary_)) throw std::invalid_argument(
            "In npstat::LOrPE1DSymbetaKernel constructor: "
            "invalid boundary specification");

        maxdeg_ = std::floor(filterDegree_);
        if (static_cast<double>(maxdeg_) != filterDegree_)
        {
            lastShrinkage_ = std::sqrt(filterDegree_ - maxdeg_);
            ++maxdeg_;
        }

        if (symbetaPower < 0)
        {
            // This is a cutoff range (somewhat arbitrary but reasonable)
            // for the Gaussian kernel with stdev of 1
            kernelRange_ = 10.0;
            poly_ = new HermiteProbOrthoPoly1D();
        }
        else if (symbetaPower == 0)
            poly_ = new LegendreOrthoPoly1D();
        else
            poly_ = new JacobiOrthoPoly1D(symbetaPower, symbetaPower);

        const unsigned maxdegPlus1 = maxdeg_ + 1U;
        polyAt0_.resize(maxdegPlus1);
        polyBuf_.resize(2U*maxdegPlus1);

        poly_->allpoly(0.0L, &polyAt0_[0], maxdeg_);
    }

    LOrPE1DSymbetaKernel::LOrPE1DSymbetaKernel(const LOrPE1DSymbetaKernel& r)
        : poly_(0)
    {
        copyInternals(r);
    }

    LOrPE1DSymbetaKernel& LOrPE1DSymbetaKernel::operator=(
        const LOrPE1DSymbetaKernel& r)
    {
        if (this != &r)
        {
            deallocateThings();
            copyInternals(r);
        }
        return *this;
    }

    void LOrPE1DSymbetaKernel::deallocateThings()
    {
        delete poly_; poly_ = 0;
    }

    void LOrPE1DSymbetaKernel::setNormFactor(const double normfactor)
    {
        if (normfactor <= 0.0) throw std::invalid_argument(
            "In npstat::LOrPE1DSymbetaKernel::setNormFactor: "
            "normalization factor must be positive");
        norm_ = normfactor;
    }

    void LOrPE1DSymbetaKernel::copyInternals(const LOrPE1DSymbetaKernel& r)
    {
        assert(!poly_);
        poly_ = r.poly_->clone();

        power_ = r.power_;
        maxdeg_ = r.maxdeg_;
        filterDegree_ = r.filterDegree_;
        leftBoundary_ = r.leftBoundary_;
        rightBoundary_ = r.rightBoundary_;
        lastShrinkage_ = r.lastShrinkage_;
        norm_ = r.norm_;
        kernelRange_ = r.kernelRange_;
        lastCenterValue_ = r.lastCenterValue_;
        bh_ = r.bh_;
        polyAt0_ = r.polyAt0_;
        memoizingBoundaryPolys_ = r.memoizingBoundaryPolys_;
        polyBuf_ = r.polyBuf_;
        usingBoundaryCorr_ = r.usingBoundaryCorr_;
        bpMap_ = r.bpMap_;
    }

    double LOrPE1DSymbetaKernel::kernelValue(const AbsClassicalOrthoPoly1D& poly,
                                             const long double* poly0,
                                             const double x) const
    {
        long double sum = 1.0L;
        if (filterDegree_)
        {
            long double* polyx = &polyBuf_[0];
            poly.allpoly(x, polyx, maxdeg_);
            for (unsigned i=1U; i<maxdeg_; ++i)
                sum += poly0[i]*polyx[i];
            sum += poly0[maxdeg_]*polyx[maxdeg_]*lastShrinkage_;
        }
        return sum*poly.weight(x);
    }

    double LOrPE1DSymbetaKernel::getLastKernelAt0() const
    {
        if (lastCenterValue_ == -DBL_MAX) throw std::runtime_error(
            "In npstat::LOrPE1DSymbetaKernel::getLastKernelAt0: please "
            "call \"lorpe\" function first with a usable sample");
        return lastCenterValue_;
    }

    bool LOrPE1DSymbetaKernel::usedBoundaryCorrection() const
    {
        if (lastCenterValue_ == -DBL_MAX) throw std::runtime_error(
            "In npstat::LOrPE1DSymbetaKernel::usedBoundaryCorrection: please "
            "call \"lorpe\" function first with a usable sample");
        return usingBoundaryCorr_;
    }

    bool LOrPE1DSymbetaKernel::isBoundaryCorrectionNeeded(
        const double x, const double h) const
    {
        if (h <= 0.0) throw std::invalid_argument(
            "In npstat::LOrPE1DSymbetaKernel::isBoundaryCorrectionNeeded: "
            "bandwidth must be positive");
        const double xlolim = x - kernelRange_*h;
        const double xuplim = x + kernelRange_*h;
        return xlolim < leftBoundary_ || xuplim > rightBoundary_;
    }

    AbsClassicalOrthoPoly1D* LOrPE1DSymbetaKernel::buildBoundaryPolyOnTheHeap(
        const double x, const double h) const
    {
        if (!isBoundaryCorrectionNeeded(x, h)) throw std::invalid_argument(
            "In npstat::LOrPE1DSymbetaKernel::buildBoundaryPoly: boundary "
            "correction is not needed for this coordinate and bandwidth");

        const unsigned polyModelMaxdeg = 1000;
        unsigned polyDegForKernel;
        CPP11_auto_ptr<AbsDistribution1D> baseKernel;
        if (power_ < 0)
        {
            baseKernel = CPP11_auto_ptr<AbsDistribution1D>(
                new TruncatedGauss1D(0.0, 1.0, kernelRange_));
            polyDegForKernel = polyModelMaxdeg;
        }
        else
        {
            assert(kernelRange_ == 1.0);
            baseKernel = CPP11_auto_ptr<AbsDistribution1D>(
                new SymmetricBeta1D(0.0, 1.0, power_));
            polyDegForKernel = 2*power_;
        }

        const unsigned nParams = bh_.nParameters();
        const double lb = (leftBoundary_ - x)/h;
        const double rb = (rightBoundary_ - x)/h;

        AbsClassicalOrthoPoly1D* bpoly = 0;
        switch (bh_.methodId())
        {
        case Private::BM_TRUNCATE:
        {
            assert(!nParams);
            TruncatedDistribution1D tr(*baseKernel, lb, rb);

            // DensityOrthoPoly1D is using Fejer quadrature.
            // With n points it can integrate exactly polynomials
            // of degree n-1.
            const unsigned nInteg = 2U*maxdeg_ + 1U + polyDegForKernel;
            bpoly = new DensityOrthoPoly1D(tr, maxdeg_, nInteg);
        }
        break;

        case Private::BM_FOLD:
        {
            assert(!nParams);
            FoldedDistribution1D fd(*baseKernel, lb, rb);

            // Folded symmetric beta is no longer polynomial.
            // Derivatives jump where folded support boundaries fall.
            const unsigned nInteg = 2U*maxdeg_ + 1U + polyModelMaxdeg;
            bpoly = new DensityOrthoPoly1D(fd, maxdeg_, nInteg);
        }
        break;

        default:
            std::string message("In npstat::LOrPE1DSymbetaKernel::buildBoundaryPoly: "
                                "boundary handling method \"");
            message += bh_.methodName();
            message += "\" is not yet supported";
            throw std::runtime_error(message);
        }

        return bpoly;
    }

    bool LOrPE1DSymbetaKernel::operator==(const LOrPE1DSymbetaKernel& r) const
    {
        return power_ == r.power_ &&
            maxdeg_ == r.maxdeg_ &&
            filterDegree_ == r.filterDegree_ &&
            leftBoundary_ == r.leftBoundary_ &&
            rightBoundary_ == r.rightBoundary_ &&
            lastShrinkage_ == r.lastShrinkage_ &&
            norm_ == r.norm_ &&
            kernelRange_ == r.kernelRange_ &&
            bh_ == r.bh_;
    }
}
