#include <set>
#include <cfloat>

#include "geners/mapIO.hh"
#include "geners/IOException.hh"

#include "npstat/stat/MemoizingSymbetaFilterProvider.hh"

#define getit(itemptr) do {\
    if (itemptr) *itemptr = it->first.get_ ## itemptr ();\
} while(0);


namespace npstat {
    CPP11_shared_ptr<const LocalPolyFilter1D> 
    MemoizingSymbetaFilterProvider::provideFilter(
        const int symbetaPower, const double bandwidth, const double degree,
        const unsigned nbins, const double binwidth,
        const BoundaryHandling& bm,
        const unsigned excludedBin, const bool excludeCentralPoint)
    {
        AllSymbetaParams1D par(symbetaPower, bandwidth, degree,
                               nbins, binwidth, bm,
                               excludedBin, excludeCentralPoint);
        FilterMap::const_iterator it = filterMap_.find(par);
        if (it != filterMap_.end())
            return it->second;

        FilterPtr ptr = SimpleSymbetaFilterProvider::provideFilter(
            symbetaPower, bandwidth, degree,
            nbins, binwidth, bm,
            excludedBin, excludeCentralPoint);
        if (this->isMemoizing())
            filterMap_[par] = ptr;
        return ptr;
    }

    bool MemoizingSymbetaFilterProvider::write(std::ostream& of) const
    {
        unsigned char memzing = this->isMemoizing();
        gs::write_pod(of, memzing);
        return gs::write_item(of, filterMap_) && !of.fail();
    }

    bool MemoizingSymbetaFilterProvider::operator==(
        const MemoizingSymbetaFilterProvider& r) const
    {
        if (this->isMemoizing() != r.isMemoizing())
            return false;
        if (filterMap_.size() != r.filterMap_.size())
            return false;
        FilterMap::const_iterator ri = r.filterMap_.begin();
        for (FilterMap::const_iterator it = filterMap_.begin();
             it != filterMap_.end(); ++it, ++ri)
        {
            if (it->first != ri->first)
                return false;
            if (*it->second != *ri->second)
                return false;
        }
        return true;
    }

    void MemoizingSymbetaFilterProvider::restore(
        const gs::ClassId& id, std::istream& in,
        MemoizingSymbetaFilterProvider* ptr)
    {
        static const gs::ClassId myClassId(
            gs::ClassId::makeId<MemoizingSymbetaFilterProvider>());
        myClassId.ensureSameId(id);

        assert(ptr);

        unsigned char memzing = 0;
        gs::read_pod(in, &memzing);
        if (memzing)
            ptr->startMemoizing();
        else
            ptr->stopMemoizing();
        return gs::restore_item(in, &ptr->filterMap_);
    }

    void MemoizingSymbetaFilterProvider::firstMemoizedInfo(
        int* symbetaPower, double* bandwidth,
        double* degree, unsigned* nbins, double* binwidth,
        BoundaryHandling* bm, unsigned* excludedBin,
        bool* excludeCentralPoint) const
    {
        if (filterMap_.empty()) throw std::runtime_error(
            "In npstat::MemoizingSymbetaFilterProvider::firstMemoizedInfo: "
            "no items memoized");
        FilterMap::const_iterator it = filterMap_.begin();
        getit(symbetaPower);
        getit(bandwidth);
        getit(degree);
        getit(nbins);
        getit(binwidth);
        getit(bm);
        getit(excludedBin);
        getit(excludeCentralPoint);
    }

    std::vector<double>
    MemoizingSymbetaFilterProvider::knownBandwidthValues() const
    {
        std::set<double> bws;
        for (FilterMap::const_iterator it = filterMap_.begin();
             it != filterMap_.end(); ++it)
            bws.insert(it->first.get_bandwidth());
        return std::vector<double>(bws.begin(), bws.end());
    }
}
