#include <cassert>
#include <stdexcept>
#include <sstream>
#include <climits>

#include "npstat/nm/findRootUsingBisections.hh"

#include "npstat/stat/PolynomialDistro1D.hh"
#include "npstat/stat/distributionReadError.hh"

namespace npstat {
    PolynomialDistro1D::PolynomialDistro1D(
        const double location, const double scale,
        const double* coeffs, const unsigned maxdeg,
        const unsigned nCheck)
        : AbsScalableDistribution1D(location, scale),
          glq_(GaussLegendreQuadrature::minimalExactRule(maxdeg))
    {
        setupCoeffs(coeffs, maxdeg);
        checkPositive(nCheck);
    }

    PolynomialDistro1D::PolynomialDistro1D(
        const double location, const double scale,
        const std::vector<double>& coeffs,
        const unsigned nCheck)
        : AbsScalableDistribution1D(location, scale),
          glq_(GaussLegendreQuadrature::minimalExactRule(coeffs.size()))
    {
        setupCoeffs(coeffs.empty() ? (const double*)0 : &coeffs[0],
                    coeffs.size());
        checkPositive(nCheck);
    }

    void PolynomialDistro1D::setupCoeffs(const double* coeffs,
                                         const unsigned maxdeg)
    {
        if (maxdeg)
            assert(coeffs);
        allCoeffs_.resize(maxdeg + 1U);
        allCoeffs_[0] = 1.0;
        for (unsigned i=0; i<maxdeg; ++i)
            allCoeffs_[i+1U] = coeffs[i];
    }

    void PolynomialDistro1D::checkPositive(const unsigned nCheck) const
    {
        if (nCheck < 2U) throw std::invalid_argument(
            "In npstat::PolynomialDistro1D::checkPositive: "
            "insufficient number of density check points");
        const unsigned nM1 = nCheck - 1U;
        const double step = 1.0/nM1;
        for (unsigned i=0; i<nCheck; ++i)
        {
            double x = step*i;
            if (i == nM1)
                x = 1.0;
            if (unscaledDensity(x) <= 0.0)
            {
                std::ostringstream os;
                os.precision(17);
                os << "In npstat::PolynomialDistro1D::checkPositive: "
                   << "density is not positive at x = " << x;
                throw std::invalid_argument(os.str());
            }
        }
    }

    double PolynomialDistro1D::unscaledDensity(const double x) const
    {
        if (x < 0.0 || x > 1.0)
            return 0.0;
        else
            return poly_.series(&allCoeffs_[0], allCoeffs_.size()-1, x);
    }

    double PolynomialDistro1D::unscaledCdf(const double x) const
    {
        if (x <= 0.0)
            return 0.0;
        else if (x >= 1.0)
            return 1.0;
        else
            return glq_.integrate(UnscaledDensityFunctor1D(*this), 0.0, x);
    }

    double PolynomialDistro1D::unscaledQuantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::PolynomialDistro1D::unscaledQuantile: "
            "cdf argument outside of [0, 1] interval");
        if (r1 == 0.0)
            return 0.0;
        if (r1 == 1.0)
            return 1.0;
        double q;
        const bool status = findRootUsingBisections(
            UnscaledCdfFunctor1D(*this), r1, 0.0, 1.0, 4.0*DBL_EPSILON, &q);
        assert(status);
        return q;
    }

    double PolynomialDistro1D::unscaledExceedance(const double x) const
    {
        if (x <= 0.0)
            return 1.0;
        else if (x >= 1.0)
            return 0.0;
        else
            return glq_.integrate(UnscaledDensityFunctor1D(*this), x, 1.0);
    }

    bool PolynomialDistro1D::isEqual(const AbsDistribution1D& other) const
    {
        const PolynomialDistro1D& r = static_cast<const PolynomialDistro1D&>(other);
        return AbsScalableDistribution1D::isEqual(r) && allCoeffs_ == r.allCoeffs_;
    }

    bool PolynomialDistro1D::write(std::ostream& os) const
    {
        AbsScalableDistribution1D::write(os);
        gs::write_pod_vector(os, allCoeffs_);
        return !os.fail();
    }

    PolynomialDistro1D* PolynomialDistro1D::read(const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(gs::ClassId::makeId<PolynomialDistro1D>());
        current.ensureSameId(id);

        double location, scale;
        if (AbsScalableDistribution1D::read(in, &location, &scale))
        {
            std::vector<double> table;
            gs::read_pod_vector(in, &table);
            if (!in.fail() && table.size())
                return new PolynomialDistro1D(location, scale,
                                              &table[1], table.size()-1, 2U);
        }
        distributionReadError(in, classname());
        return 0;
    }
}
