#ifndef NPSTAT_POLYNOMIALDISTRO1D_HH_
#define NPSTAT_POLYNOMIALDISTRO1D_HH_

/*!
// \file PolynomialDistro1D.hh
//
// \brief Statistical distribution constructed using polynomial series
//
// Author: I. Volobouev
//
// November 2020
*/

#include "npstat/nm/GaussLegendreQuadrature.hh"
#include "npstat/nm/ClassicalOrthoPolys1D.hh"

#include "npstat/stat/Distribution1DFactory.hh"

namespace npstat {
    /**
    // A distribution whose density is a series in terms of 
    // orthonormal Legendre polynomials on [0, 1]. The constructor
    // parameter "nCheck" gives the number of points for which the
    // code checks that the density is positive. The series
    // coefficients start with degree 1 (note, not 0).
    */
    class PolynomialDistro1D : public AbsScalableDistribution1D
    {
    public:
        PolynomialDistro1D(double location, double scale,
                           const double* coeffs, unsigned maxdeg,
                           unsigned nCheck = 10001);

        PolynomialDistro1D(double location, double scale,
                           const std::vector<double>& coeffs,
                           unsigned nCheck = 10001);

        inline virtual PolynomialDistro1D* clone() const
            {return new PolynomialDistro1D(*this);}

        inline virtual ~PolynomialDistro1D() {}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname()
            {return "npstat::PolynomialDistro1D";}
        static inline unsigned version() {return 1;}
        static PolynomialDistro1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<PolynomialDistro1D>;

        inline static int nParameters() {return -1;}

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledQuantile(double x) const;
        double unscaledExceedance(double x) const;

        void checkPositive(unsigned nCheck) const;
        void setupCoeffs(const double* coeffs, unsigned maxdeg);

        ShiftedLegendreOrthoPoly1D poly_;
        GaussLegendreQuadrature glq_;
        std::vector<double> allCoeffs_;
    };
}

#endif // NPSTAT_POLYNOMIALDISTRO1D_HH_
