#include <cmath>
#include <cfloat>
#include <stdexcept>
#include <utility>
#include <map>

#include "geners/binaryIO.hh"

#include "npstat/nm/SpecialFunctions.hh"
#include "npstat/nm/MathUtils.hh"
#include "npstat/nm/binomialCoefficient.hh"

#include "npstat/stat/PowerHalfCauchy1D.hh"
#include "npstat/stat/distributionReadError.hh"

inline static double quickpow(const double x, const double p)
{
    double r;
    if (p == 0.0)
        r = 1.0;
    else if (p == 1.0)
        r = x;
    else if (p == 2.0)
        r = x*x;
    else
        r = pow(x, p);
    return r;
}

static long double cdf_0_1(const long double x)
{
    return atanl(x);
}

static long double cdf_0_2(const long double x)
{
    return (x/(1 + x*x) + atanl(x))/2.0;
}

static long double cdf_1_2(const long double x)
{
    const long double x_2 = x*x;
    return x_2/(2 + 2*x_2);
}

static long double cdf_2_2(const long double x)
{
    return (-(x/(1 + x*x)) + atanl(x))/2.0;
}

static long double cdf_0_3(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;

    return ((x*(5 + 3*x_2))/(onepx2*onepx2) + 3*atanl(x))/8;
}

static long double cdf_1_3(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;

    return 0.25 - 1/(4*(onepx2*onepx2));
}

static long double cdf_2_3(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;

    return ((x*(-1 + x_2))/(onepx2*onepx2) + atanl(x))/8;
}

static long double cdf_3_3(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;

    return x_4/(4*(onepx2*onepx2));
}

static long double cdf_4_3(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;

    return -(x*(3 + 5*x_2))/(8*(onepx2*onepx2)) + (3*atanl(x))/8;
}

static long double cdf_0_4(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_3 = onepx2*onepx2*onepx2;

    return ((x*(33 + 40*x_2 + 15*x_4))/onepx2_3 + 15*atanl(x))/48;
}

static long double cdf_1_4(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_3 = onepx2*onepx2*onepx2;

    return 1/6.0L - 1/(6*onepx2_3);
}

static long double cdf_2_4(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_3 = onepx2*onepx2*onepx2;

    return ((x*(-3 + 8*x_2 + 3*x_4))/onepx2_3 + 3*atanl(x))/48;
}

static long double cdf_3_4(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_3 = onepx2*onepx2*onepx2;

    return (x_4*(3 + x_2))/(12*onepx2_3);
}

static long double cdf_4_4(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_3 = onepx2*onepx2*onepx2;

    return ((x*(-3 + x_2)*(1 + 3*x_2))/onepx2_3 + 3*atanl(x))/48;
}

static long double cdf_5_4(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_3 = onepx2*onepx2*onepx2;

    return x_2*x_2*x_2/(6*onepx2_3);
}

static long double cdf_6_4(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_3 = onepx2*onepx2*onepx2;

    return -(x*(15 + 40*x_2 + 33*x_4))/(48*onepx2_3) + (5*atanl(x))/16;
}

static long double cdf_0_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return ((x*(279 + 7*x_2*(73 + 55*x_2 + 15*x_4)))/onepx2_4 + 105*atanl(x))/384;
}

static long double cdf_1_5(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return 0.125 - 1/(8*onepx2_4);
}

static long double cdf_2_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return ((x*(-15 + 73*x_2 + 55*x_4 + 15*x_6))/onepx2_4 + 15*atanl(x))/384;
}

static long double cdf_3_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return (x_4*(6 + 4*x_2 + x_4))/(24*onepx2_4);
}

static long double cdf_4_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return (((-1 + x)*x*(1 + x)*(3 + 14*x_2 + 3*x_4))/onepx2_4 + 3*atanl(x))/128;
}

static long double cdf_5_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return (x_6*(4 + x_2))/(24*onepx2_4);
}

static long double cdf_6_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return ((x*(-15 - 55*x_2 - 73*x_4 + 15*x_6))/onepx2_4 + 15*atanl(x))/384;
}

static long double cdf_7_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_8 = x_4*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return x_8/(8*onepx2_4);
}

static long double cdf_8_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;

    return (-((x*(105 + 385*x_2 + 511*x_4 + 279*x_6))/onepx2_4) + 105*atanl(x))/384;
}

static long double cdf_0_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return ((x*(965 + 3*x_2*(790 + 896*x_2 + 490*x_4 + 105*x_6)))/onepx2_5 + 315*atanl(x))/1280;
}

static long double cdf_1_6(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return (1 - 1/onepx2_5)/10;
}

static long double cdf_2_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double x_8 = x_4*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return ((x*(-105 + 790*x_2 + 896*x_4 + 490*x_6 + 105*x_8))/onepx2_5 + 105*atanl(x))/3840;
}

static long double cdf_3_6(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_4 = onepx2_2*onepx2_2;
    const long double onepx2_5 = onepx2_4*onepx2;

    return (1 + 4/onepx2_5 - 5/onepx2_4)/40;
}

static long double cdf_4_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double x_8 = x_4*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return ((x*(-15 - 70*x_2 + 128*x_4 + 70*x_6 + 15*x_8))/onepx2_5 + 15*atanl(x))/1280;
}

static long double cdf_5_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return (x_6*(10 + 5*x_2 + x_4))/(60*onepx2_5);
}

static long double cdf_6_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double x_8 = x_4*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return ((x*(-15 - 70*x_2 - 128*x_4 + 70*x_6 + 15*x_8))/onepx2_5 + 15*atanl(x))/1280;
}

static long double cdf_7_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_8 = x_4*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return (x_8*(5 + x_2))/(40*onepx2_5);
}

static long double cdf_8_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double x_8 = x_4*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return ((x*(-105 - 490*x_2 - 896*x_4 - 790*x_6 + 105*x_8))/onepx2_5 + 105*atanl(x))/3840;
}

static long double cdf_9_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double x_10 = x_4*x_6;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return x_10/(10*onepx2_5);
}

static long double cdf_10_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double x_6 = x_2*x_4;
    const long double x_8 = x_4*x_4;
    const long double onepx2 = 1.0 + x_2;
    const long double onepx2_2 = onepx2*onepx2;
    const long double onepx2_5 = onepx2_2*onepx2_2*onepx2;

    return (-((x*(315 + 1470*x_2 + 2688*x_4 + 2370*x_6 + 965*x_8))/onepx2_5) + 315*atanl(x))/1280;
}

typedef long double (*Ptr)(long double);

namespace {
    struct Cdfmap : public std::map<std::pair<unsigned, unsigned>, Ptr>
    {
        typedef std::pair<unsigned, unsigned> P;

        inline Cdfmap()
        {
            (*this)[P(0,1)] = &cdf_0_1;

            (*this)[P(0,2)] = &cdf_0_2;
            (*this)[P(1,2)] = &cdf_1_2;
            (*this)[P(2,2)] = &cdf_2_2;

            (*this)[P(0,3)] = &cdf_0_3;
            (*this)[P(1,3)] = &cdf_1_3;
            (*this)[P(2,3)] = &cdf_2_3;
            (*this)[P(3,3)] = &cdf_3_3;
            (*this)[P(4,3)] = &cdf_4_3;

            (*this)[P(0,4)] = &cdf_0_4;
            (*this)[P(1,4)] = &cdf_1_4;
            (*this)[P(2,4)] = &cdf_2_4;
            (*this)[P(3,4)] = &cdf_3_4;
            (*this)[P(4,4)] = &cdf_4_4;
            (*this)[P(5,4)] = &cdf_5_4;
            (*this)[P(6,4)] = &cdf_6_4;

            (*this)[P(0,5)] = &cdf_0_5;
            (*this)[P(1,5)] = &cdf_1_5;
            (*this)[P(2,5)] = &cdf_2_5;
            (*this)[P(3,5)] = &cdf_3_5;
            (*this)[P(4,5)] = &cdf_4_5;
            (*this)[P(5,5)] = &cdf_5_5;
            (*this)[P(6,5)] = &cdf_6_5;
            (*this)[P(7,5)] = &cdf_7_5;
            (*this)[P(8,5)] = &cdf_8_5;

            (*this)[P(0,6)] = &cdf_0_6;
            (*this)[P(1,6)] = &cdf_1_6;
            (*this)[P(2,6)] = &cdf_2_6;
            (*this)[P(3,6)] = &cdf_3_6;
            (*this)[P(4,6)] = &cdf_4_6;
            (*this)[P(5,6)] = &cdf_5_6;
            (*this)[P(6,6)] = &cdf_6_6;
            (*this)[P(7,6)] = &cdf_7_6;
            (*this)[P(8,6)] = &cdf_8_6;
            (*this)[P(9,6)] = &cdf_9_6;
            (*this)[P(10,6)] = &cdf_10_6;
        }
    };
}

static Ptr findCdf(const unsigned m, const unsigned n)
{
    static const Cdfmap cdfmap;

    Cdfmap::const_iterator it = cdfmap.find(std::make_pair(m,n));
    if (it == cdfmap.end())
        return 0;
    else
        return it->second;
}

namespace npstat {
    bool PowerHalfCauchy1D::write(std::ostream& os) const
    {
        AbsScalableDistribution1D::write(os);
        gs::write_pod(os, m_);
        gs::write_pod(os, n_);
        return !os.fail();
    }

    PowerHalfCauchy1D* PowerHalfCauchy1D::read(const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(gs::ClassId::makeId<PowerHalfCauchy1D>());
        current.ensureSameId(id);

        double location, scale;
        unsigned a, b;
        if (AbsScalableDistribution1D::read(in, &location, &scale))
        {
            gs::read_pod(in, &a);
            gs::read_pod(in, &b);
            if (!in.fail())
                return new PowerHalfCauchy1D(location, scale, a, b);
        }
        distributionReadError(in, classname());
        return 0;
    }

    bool PowerHalfCauchy1D::isEqual(const AbsDistribution1D& otherBase) const
    {
        const PowerHalfCauchy1D& r = static_cast<const PowerHalfCauchy1D&>(otherBase);
        return AbsScalableDistribution1D::isEqual(r) && m_ == r.m_ && n_ == r.n_;
    }

    void PowerHalfCauchy1D::initialize()
    {
        const unsigned nCdfTerms = 21;
        const double cdfSwitchValue = 0.25;

        norm_ = calculateNorm();
        xmax_ = calculateXMax();
        cdf_  = findCdf(m_, n_);
        if (!cdf_) throw std::invalid_argument(
            "In npstat::PowerHalfCauchy1D::initialize: "
            "unsupported combination of power parameters");

        cdfCoeffs_.reserve(nCdfTerms);
        for (unsigned k=0; k<nCdfTerms; ++k)
            cdfCoeffs_.push_back(ldBinomialCoefficient(n_+k-1U, k)/(m_+1U+2*k));
        cdfSwitchX_ = cdfSwitchValue;

        rmax_ = unscaledCdf(xmax_);
        assert(std::isnormal(rmax_));
        if (rmax_ > 1.0)
            rmax_ = 1.0;

        const double xtry = cdfSwitchX_*0.9999999;
        const double cdf1 = norm_ * (*cdf_)(xtry);
        const double cdf2 = unscaledCdf(xtry);
        assert(std::abs(cdf1 - cdf2)*2.0/(cdf1 + cdf2) < 1.0e-11);
    }

    PowerHalfCauchy1D::PowerHalfCauchy1D(const double location, const double scale,
                                         const unsigned i_m, const unsigned i_n)
        : AbsScalableDistribution1D(location, scale),
          m_(i_m),
          n_(i_n)
    {
        initialize();
    }

    PowerHalfCauchy1D::PowerHalfCauchy1D(const double location,
                                         const double scale,
                                         const std::vector<double>& params)
        : AbsScalableDistribution1D(location, scale),
          m_(params[0]),
          n_(params[1])
    {
        if (static_cast<double>(m_) != params[0] ||
            static_cast<double>(n_) != params[1])
            throw std::invalid_argument(
                "In npstat::PowerHalfCauchy1D constructor: "
                "power parameters must be non-negative integers");
        initialize();
    }

    double PowerHalfCauchy1D::calculateNorm() const
    {
        if (!(2U*n_ > 1U + m_)) throw std::invalid_argument(
            "In npstat::PowerHalfCauchy1D::calculateNorm: invalid power parameters");
        const double tmp = (1U + m_)/2.0;
        return 2.0*Gamma(n_)/Gamma(tmp)/Gamma(n_ - tmp);
    }

    double PowerHalfCauchy1D::calculateXMax() const
    {
        return pow(DBL_MAX, 0.5/n_);
    }

    double PowerHalfCauchy1D::unscaledDensity(const double x) const
    {
        if (x < 0.0 || x >= xmax_)
            return 0.0;
        else
            return norm_*quickpow(x, m_)/quickpow(1.0 + x*x, n_);
    }

    double PowerHalfCauchy1D::unscaledExceedance(const double x) const
    {
        return 1.0 - unscaledCdf(x);
    }

    double PowerHalfCauchy1D::unscaledCdf(const double x) const
    {
        if (x >= xmax_)
            return 1.0;
        else if (x <= 0.0)
            return 0.0;
        else
        {
            if (x <= cdfSwitchX_)
            {
                const long double ldx = x;
                const long double mxsq = -ldx*ldx;
                return norm_ * powl(ldx, m_+1U)*polySeriesSum(&cdfCoeffs_[0], cdfCoeffs_.size()-1U, mxsq);
            }
            else
                return norm_ * (*cdf_)(x);
        }
    }

    double PowerHalfCauchy1D::unscaledQuantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::PowerHalfCauchy1D::unscaledQuantile: "
            "cdf argument outside of [0, 1] interval");
        if (r1 == 0.0)
            return 0.0;
        else if (r1 >= rmax_)
            return xmax_;
        else
        {
            double qmin = 0.0;
            double qmax = xmax_;
            for (unsigned i=0; i<1000U; ++i)
            {
                const double x = (qmin + qmax)/2.0;
                if ((qmax - qmin)/x < 2.0*DBL_EPSILON)
                    return x;
                const double fval = unscaledCdf(x);
                if (fval == r1)
                    return x;
                else if (fval > r1)
                    qmax = x;
                else
                    qmin = x;
                if (qmax == qmin)
                    return qmin;
            }
            return (qmin + qmax)/2.0;
        }
    }
}
