#ifndef NPSTAT_POWERHALFCAUCHY1D_HH_
#define NPSTAT_POWERHALFCAUCHY1D_HH_

/*!
// \file PowerHalfCauchy1D.hh
//
// \brief Distribution with unscaled density proportional to x^m/(1 + x^2)^n
//        for x >= 0 (and 0 for x < 0)
//
// Author: I. Volobouev
//
// September 2022
*/

#include "npstat/stat/Distribution1DFactory.hh"

namespace npstat {
    class PowerHalfCauchy1D : public AbsScalableDistribution1D
    {
    public:
        /**
        // Must have 2*n > m + 1 (otherwise the density integral
        // will not converge)
        */
        PowerHalfCauchy1D(double location, double scale, unsigned m, unsigned n);

        inline virtual PowerHalfCauchy1D* clone() const
            {return new PowerHalfCauchy1D(*this);}

        inline virtual ~PowerHalfCauchy1D() {}

        inline unsigned m() const {return m_;}
        inline unsigned n() const {return n_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::PowerHalfCauchy1D";}
        static inline unsigned version() {return 1;}
        static PowerHalfCauchy1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<PowerHalfCauchy1D>;

        PowerHalfCauchy1D(double location, double scale,
                          const std::vector<double>& params);
        inline static int nParameters() {return 2;}

        void initialize();

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledExceedance(double x) const;
        double unscaledQuantile(double x) const;

        double calculateNorm() const;
        double calculateXMax() const;

        unsigned m_;
        unsigned n_;
        double norm_;
        double xmax_;
        long double (*cdf_)(long double);
        double rmax_;
        double cdfSwitchX_;
        std::vector<long double> cdfCoeffs_;
    };
}

#endif // NPSTAT_POWERHALFCAUCHY1D_HH_
