#include <cmath>
#include <cfloat>
#include <stdexcept>
#include <utility>
#include <map>

#include "geners/binaryIO.hh"

#include "npstat/nm/SpecialFunctions.hh"
#include "npstat/nm/MathUtils.hh"
#include "npstat/nm/binomialCoefficient.hh"

#include "npstat/stat/PowerRatio1D.hh"
#include "npstat/stat/distributionReadError.hh"

inline static double quickpow(const double x, const double p)
{
    double r;
    if (p == 0.0)
        r = 1.0;
    else if (p == 0.5)
        r = sqrt(x);
    else if (p == 1.0)
        r = x;
    else if (p == 1.5)
        r = x*sqrt(x);
    else if (p == 2.0)
        r = x*x;
    else
        r = pow(x, p);
    return r;
}

inline static long double atansqrl(const long double x)
{
    return atanl(sqrtl(x));
}

static long double cdf_m1_1(const long double x)
{
    return 2*atansqrl(x);
}

static long double cdf_m1_2(const long double x)
{
    return sqrtl(x)/(1 + x) + atansqrl(x);
}

static long double cdf_0_2(const long double x)
{
    return x/(1 + x);
}

static long double cdf_1_2(const long double x)
{
    return -(sqrtl(x)/(1 + x)) + atansqrl(x);
}

static long double cdf_m1_3(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;

    return ((sqrtl(x)*(5 + 3*x))/onepx_2 + 3*atansqrl(x))/4;
}

static long double cdf_0_3(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;

    return (x*(2 + x))/(2*onepx_2);
}

static long double cdf_1_3(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;

    return (((-1 + x)*sqrtl(x))/onepx_2 + atansqrl(x))/4;
}

static long double cdf_2_3(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;

    return x_2/(2*onepx_2);
}

static long double cdf_3_3(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;

    return -(sqrtl(x)*(3 + 5*x))/(4*onepx_2) + (3*atansqrl(x))/4;
}

static long double cdf_m1_4(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_3 = onepx*onepx*onepx;

    return (sqrtl(x)*(33 + 5*x*(8 + 3*x)))/(24*onepx_3) + (5*atansqrl(x))/8;
}

static long double cdf_0_4(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_3 = onepx*onepx*onepx;

    return 1/3.0L - 1/(3*onepx_3);
}

static long double cdf_1_4(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_3 = onepx*onepx*onepx;

    return ((sqrtl(x)*(3 + x)*(-1 + 3*x))/onepx_3 + 3*atansqrl(x))/24;
}

static long double cdf_2_4(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx = 1 + x;
    const long double onepx_3 = onepx*onepx*onepx;

    return (x_2*(3 + x))/(6*onepx_3);
}

static long double cdf_3_4(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_3 = onepx*onepx*onepx;

    return (((-3 + x)*sqrtl(x)*(1 + 3*x))/onepx_3 + 3*atansqrl(x))/24;
}

static long double cdf_4_4(const long double x)
{
    const long double x_3 = x*x*x;
    const long double onepx = 1 + x;
    const long double onepx_3 = onepx*onepx*onepx;

    return x_3/(3*onepx_3);
}

static long double cdf_5_4(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_3 = onepx*onepx*onepx;

    return -(sqrtl(x)*(15 + x*(40 + 33*x)))/(24*onepx_3) + (5*atansqrl(x))/8;
}

static long double cdf_m1_5(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return ((sqrtl(x)*(279 + 7*x*(73 + 5*x*(11 + 3*x))))/onepx_4 + 105*atansqrl(x))/192;
}

static long double cdf_0_5(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return 0.25 - 1/(4*onepx_4);
}

static long double cdf_1_5(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return ((sqrtl(x)*(-15 + x*(73 + 5*x*(11 + 3*x))))/onepx_4 + 15*atansqrl(x))/192;
}

static long double cdf_2_5(const long double x)
{
    const long double x_2 = x*x;
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return (x_2*(6 + x*(4 + x)))/(12*onepx_4);
}

static long double cdf_3_5(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return (((-1 + x)*sqrtl(x)*(3 + x*(14 + 3*x)))/onepx_4 + 3*atansqrl(x))/64;
}

static long double cdf_4_5(const long double x)
{
    const long double x_3 = x*x*x;
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return (x_3*(4 + x))/(12*onepx_4);
}

static long double cdf_5_5(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return ((sqrtl(x)*(-15 + x*(-55 + x*(-73 + 15*x))))/onepx_4 + 15*atansqrl(x))/192;
}

static long double cdf_6_5(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return x_4/(4*onepx_4);
}

static long double cdf_7_5(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;

    return -(sqrtl(x)*(105 + x*(385 + x*(511 + 279*x))))/(192*onepx_4) + (35*atansqrl(x))/64;
}

static long double cdf_m1_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return ((sqrtl(x)*(965 + 3*x*(790 + 7*x*(128 + 5*x*(14 + 3*x)))))/onepx_5 + 315*atansqrl(x))/640;
}

static long double cdf_0_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return 0.2 - 1/(5*onepx_5);
}

static long double cdf_1_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return ((sqrtl(x)*(-105 + x*(790 + 7*x*(128 + 5*x*(14 + 3*x)))))/onepx_5 + 105*atansqrl(x))/1920;
}

static long double cdf_2_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_4 = onepx_2*onepx_2;
    const long double onepx_5 = onepx_4*onepx;

    return (1 + 4/onepx_5 - 5/onepx_4)/20;
}

static long double cdf_3_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return ((sqrtl(x)*(-15 + x*(-70 + x*(128 + 5*x*(14 + 3*x)))))/onepx_5 + 15*atansqrl(x))/640;
}

static long double cdf_4_6(const long double x)
{
    const long double x_3 = x*x*x;
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return (x_3*(10 + x*(5 + x)))/(30*onepx_5);
}

static long double cdf_5_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return ((sqrtl(x)*(-15 + x*(-70 + x*(-128 + 5*x*(14 + 3*x)))))/onepx_5 + 15*atansqrl(x))/640;
}

static long double cdf_6_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_4 = x_2*x_2;
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return (x_4*(5 + x))/(20*onepx_5);
}

static long double cdf_7_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;
    return ((sqrtl(x)*(-105 + x*(-490 + x*(-896 + 5*x*(-158 + 21*x)))))/onepx_5 + 105*atansqrl(x))/1920;
}

static long double cdf_8_6(const long double x)
{
    const long double x_2 = x*x;
    const long double x_5 = x_2*x_2*x;
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return x_5/(5*onepx_5);
}

static long double cdf_9_6(const long double x)
{
    const long double onepx = 1 + x;
    const long double onepx_2 = onepx*onepx;
    const long double onepx_5 = onepx_2*onepx_2*onepx;

    return (-((sqrtl(x)*(315 + x*(1470 + x*(2688 + 5*x*(474 + 193*x)))))/onepx_5) + 315*atansqrl(x))/640;
}

typedef long double (*Ptr)(long double);

namespace {
    struct Cdfmap : public std::map<std::pair<int, unsigned>, Ptr>
    {
        typedef std::pair<int, unsigned> P;

        inline Cdfmap()
        {
            (*this)[P(-1,1)] = &cdf_m1_1;

            (*this)[P(-1,2)] = &cdf_m1_2;
            (*this)[P(0,2)] = &cdf_0_2;
            (*this)[P(1,2)] = &cdf_1_2;

            (*this)[P(-1,3)] = &cdf_m1_3;
            (*this)[P(0,3)] = &cdf_0_3;
            (*this)[P(1,3)] = &cdf_1_3;
            (*this)[P(2,3)] = &cdf_2_3;
            (*this)[P(3,3)] = &cdf_3_3;

            (*this)[P(-1,4)] = &cdf_m1_4;
            (*this)[P(0,4)] = &cdf_0_4;
            (*this)[P(1,4)] = &cdf_1_4;
            (*this)[P(2,4)] = &cdf_2_4;
            (*this)[P(3,4)] = &cdf_3_4;
            (*this)[P(4,4)] = &cdf_4_4;
            (*this)[P(5,4)] = &cdf_5_4;

            (*this)[P(-1,5)] = &cdf_m1_5;
            (*this)[P(0,5)] = &cdf_0_5;
            (*this)[P(1,5)] = &cdf_1_5;
            (*this)[P(2,5)] = &cdf_2_5;
            (*this)[P(3,5)] = &cdf_3_5;
            (*this)[P(4,5)] = &cdf_4_5;
            (*this)[P(5,5)] = &cdf_5_5;
            (*this)[P(6,5)] = &cdf_6_5;
            (*this)[P(7,5)] = &cdf_7_5;

            (*this)[P(-1,6)] = &cdf_m1_6;
            (*this)[P(0,6)] = &cdf_0_6;
            (*this)[P(1,6)] = &cdf_1_6;
            (*this)[P(2,6)] = &cdf_2_6;
            (*this)[P(3,6)] = &cdf_3_6;
            (*this)[P(4,6)] = &cdf_4_6;
            (*this)[P(5,6)] = &cdf_5_6;
            (*this)[P(6,6)] = &cdf_6_6;
            (*this)[P(7,6)] = &cdf_7_6;
            (*this)[P(8,6)] = &cdf_8_6;
            (*this)[P(9,6)] = &cdf_9_6;
        }
    };
}

static Ptr findCdf(const int twom, const unsigned n)
{
    static const Cdfmap cdfmap;

    Cdfmap::const_iterator it = cdfmap.find(std::make_pair(twom,n));
    if (it == cdfmap.end())
        return 0;
    else
        return it->second;
}

namespace npstat {
    bool PowerRatio1D::write(std::ostream& os) const
    {
        AbsScalableDistribution1D::write(os);
        gs::write_pod(os, twom_);
        gs::write_pod(os, n_);
        return !os.fail();
    }

    PowerRatio1D* PowerRatio1D::read(const gs::ClassId& id, std::istream& in)
    {
        static const gs::ClassId current(gs::ClassId::makeId<PowerRatio1D>());
        current.ensureSameId(id);

        double location, scale;
        int a;
        unsigned b;
        if (AbsScalableDistribution1D::read(in, &location, &scale))
        {
            gs::read_pod(in, &a);
            gs::read_pod(in, &b);
            if (!in.fail())
                return new PowerRatio1D(location, scale, a, b);
        }
        distributionReadError(in, classname());
        return 0;
    }

    bool PowerRatio1D::isEqual(const AbsDistribution1D& otherBase) const
    {
        const PowerRatio1D& r = static_cast<const PowerRatio1D&>(otherBase);
        return AbsScalableDistribution1D::isEqual(r) && twom_ == r.twom_ && n_ == r.n_;
    }

    void PowerRatio1D::initialize()
    {
        const unsigned nCdfTerms = 31;
        const double cdfSwitchValue = 0.1;

        norm_ = calculateNorm();
        xmax_ = calculateXMax();
        cdf_  = findCdf(twom_, n_);
        if (!cdf_) throw std::invalid_argument(
            "In npstat::PowerRatio1D::initialize: "
            "unsupported combination of power parameters");

        const double m = twom_/2.0;
        cdfCoeffs_.reserve(nCdfTerms);
        for (unsigned k=0; k<nCdfTerms; ++k)
            cdfCoeffs_.push_back(ldBinomialCoefficient(n_+k-1U, k)/(m + 1 + k));
        cdfSwitchX_ = cdfSwitchValue;

        rmax_ = unscaledCdf(xmax_);
        assert(std::isnormal(rmax_));
        if (rmax_ > 1.0)
            rmax_ = 1.0;

        const double xtry = cdfSwitchX_*0.9999999;
        const double cdf1 = norm_ * (*cdf_)(xtry);
        const double cdf2 = unscaledCdf(xtry);
        assert(std::abs(cdf1 - cdf2)*2.0/(cdf1 + cdf2) < 1.0e-12);
    }

    PowerRatio1D::PowerRatio1D(const double location, const double scale,
                               const int i_m, const unsigned i_n)
        : AbsScalableDistribution1D(location, scale),
          twom_(i_m),
          n_(i_n)
    {
        initialize();
    }

    PowerRatio1D::PowerRatio1D(const double location,
                               const double scale,
                               const std::vector<double>& params)
        : AbsScalableDistribution1D(location, scale),
          twom_(params[0]),
          n_(params[1])
    {
        if (static_cast<double>(twom_) != params[0] ||
            static_cast<double>(n_) != params[1])
            throw std::invalid_argument(
                "In npstat::PowerRatio1D constructor: "
                "power parameters must be integers");
        initialize();
    }

    double PowerRatio1D::calculateNorm() const
    {
        if (!(2U*n_ > 2U + twom_) || twom_ < -1) throw std::invalid_argument(
            "In npstat::PowerRatio1D::calculateNorm: invalid power parameters");
        const double m = twom_/2.0;
        return Gamma(n_)/Gamma(m + 1)/Gamma(n_ - m - 1);
    }

    double PowerRatio1D::calculateXMax() const
    {
        return pow(DBL_MAX, 1.0/n_);
    }

    double PowerRatio1D::unscaledDensity(const double x) const
    {
        if (x < 0.0 || x >= xmax_)
            return 0.0;
        else
            return norm_*quickpow(x, twom_/2.0)/quickpow(1.0 + x, n_);
    }

    double PowerRatio1D::unscaledExceedance(const double x) const
    {
        return 1.0 - unscaledCdf(x);
    }

    double PowerRatio1D::unscaledCdf(const double x) const
    {
        if (x >= xmax_)
            return 1.0;
        else if (x <= 0.0)
            return 0.0;
        else
        {
            if (x <= cdfSwitchX_)
            {
                const double m = twom_/2.0;
                const long double ldx = x;
                const long double mx = -ldx;
                return norm_ * powl(ldx, m+1)*polySeriesSum(&cdfCoeffs_[0], cdfCoeffs_.size()-1U, mx);
            }
            else
                return norm_ * (*cdf_)(x);
        }
    }

    double PowerRatio1D::unscaledQuantile(const double r1) const
    {
        if (!(r1 >= 0.0 && r1 <= 1.0)) throw std::domain_error(
            "In npstat::PowerRatio1D::unscaledQuantile: "
            "cdf argument outside of [0, 1] interval");
        if (r1 == 0.0)
            return 0.0;
        else if (r1 >= rmax_)
            return xmax_;
        else
        {
            double qmin = 0.0;
            double qmax = xmax_;
            for (unsigned i=0; i<1000U; ++i)
            {
                const double x = (qmin + qmax)/2.0;
                if ((qmax - qmin)/x < 2.0*DBL_EPSILON)
                    return x;
                const double fval = unscaledCdf(x);
                if (fval == r1)
                    return x;
                else if (fval > r1)
                    qmax = x;
                else
                    qmin = x;
                if (qmax == qmin)
                    return qmin;
            }
            return (qmin + qmax)/2.0;
        }
    }
}
