#ifndef NPSTAT_POWERRATIO1D_HH_
#define NPSTAT_POWERRATIO1D_HH_

/*!
// \file PowerRatio1D.hh
//
// \brief Distribution with unscaled density proportional to x^m/(1 + x)^n
//        for x >= 0 (and 0 for x < 0). m can be integer or half-integer,
//        starting with -1/2 and going up.
//
// Author: I. Volobouev
//
// September 2022
*/

#include "npstat/stat/Distribution1DFactory.hh"

namespace npstat {
    class PowerRatio1D : public AbsScalableDistribution1D
    {
    public:
        /** 
        // Parameter "twom" is 2*m. Must have 2*n > twom + 2 (otherwise
        // the density integral will not converge). twom must also be >= -1
        // (so that m >= -1/2).
        */
        PowerRatio1D(double location, double scale, int twom, unsigned n);

        inline virtual PowerRatio1D* clone() const
            {return new PowerRatio1D(*this);}

        inline virtual ~PowerRatio1D() {}

        inline unsigned twom() const {return twom_;}
        inline unsigned n() const {return n_;}

        // Methods needed for I/O
        virtual gs::ClassId classId() const {return gs::ClassId(*this);}
        virtual bool write(std::ostream& os) const;

        static inline const char* classname() {return "npstat::PowerRatio1D";}
        static inline unsigned version() {return 1;}
        static PowerRatio1D* read(const gs::ClassId& id, std::istream& in);

    protected:
        virtual bool isEqual(const AbsDistribution1D&) const;

    private:
        friend class ScalableDistribution1DFactory<PowerRatio1D>;

        PowerRatio1D(double location, double scale,
                     const std::vector<double>& params);
        inline static int nParameters() {return 2;}

        void initialize();

        double unscaledDensity(double x) const;
        double unscaledCdf(double x) const;
        double unscaledExceedance(double x) const;
        double unscaledQuantile(double x) const;

        double calculateNorm() const;
        double calculateXMax() const;

        int twom_;
        unsigned n_;
        double norm_;
        double xmax_;
        long double (*cdf_)(long double);
        double rmax_;
        double cdfSwitchX_;
        std::vector<long double> cdfCoeffs_;
    };
}

#endif // NPSTAT_POWERRATIO1D_HH_
