#include <cmath>
#include <stdexcept>

#include "geners/binaryIO.hh"
#include "npstat/stat/PowerTransform1D.hh"

inline static double quickpow(const double x, const double p)
{
    double r;
    if (p == 0.0)
        r = 1.0;
    else if (p == 0.5)
        r = sqrt(x);
    else if (p == 1.0)
        r = x;
    else if (p == 2.0)
        r = x*x;
    else if (p == -0.5)
        r = 1.0/sqrt(x);
    else if (p == -1.0)
        r = 1.0/x;
    else
        r = pow(x, p);
    return r;
}

namespace npstat {
    void PowerTransform1D::validatePower(const double p)
    {
        if (p == 0.0) throw std::invalid_argument(
            "In npstat::PowerTransform1D::validatePower: invalid power parameter");
    }

    void PowerTransform1D::validateScale(const double a)
    {
        if (a <= 0.0) throw std::invalid_argument(
            "In npstat::PowerTransform1D::validateScale: invalid scale parameter");
    }

    PowerTransform1D::PowerTransform1D(const double a, const double p)
        : AbsDistributionTransform1D(2U)
    {
        validateScale(a);
        validatePower(p);
        params_[0] = a;
        params_[1] = p;
    }

    double PowerTransform1D::transformForward(const double x, double* dydx) const
    {
        if (x < 0.0) throw std::invalid_argument(
            "In npstat::PowerTransform1D::transformForward: coordinate must be non-negative");
        const double p = params_[1];
        if (p < 0.0 && x == 0.0) throw std::invalid_argument(
            "In npstat::PowerTransform1D::transformForward: coordinate must be positive");
        const double a = params_[0];
        if (dydx)
        {
            if (p == -1.0)
                *dydx = a*log(x);
            else
                *dydx = a*p*quickpow(x, p-1.0);
        }
        return a*quickpow(x, p);
    }

    double PowerTransform1D::transformBack(const double y) const
    {
        if (y < 0.0) throw std::invalid_argument(
            "In npstat::PowerTransform1D::transformBack: coordinate must be non-negative");
        const double p = params_[1];
        if (p < 0.0 && y == 0.0) throw std::invalid_argument(
            "In npstat::PowerTransform1D::transformBack: coordinate must be positive");
        const double a = params_[0];
        return quickpow(y/a, 1.0/p);
    }

    bool PowerTransform1D::isEqual(const AbsDistributionTransform1D& o) const
    {
        const PowerTransform1D& r = static_cast<const PowerTransform1D&>(o);
        return params_[0] == r.params_[0] && params_[1] == r.params_[1];
    }

    bool PowerTransform1D::write(std::ostream& os) const
    {
        gs::write_pod(os, params_[0]);
        gs::write_pod(os, params_[1]);
        return !os.fail();
    }

    PowerTransform1D* PowerTransform1D::read(const gs::ClassId& id,
                                             std::istream& is)
    {
        static const gs::ClassId current(gs::ClassId::makeId<PowerTransform1D>());
        current.ensureSameId(id);

        double a, p;
        gs::read_pod(is, &a);
        gs::read_pod(is, &p);
        if (is.fail()) throw gs::IOReadFailure(
            "In npstat::PowerTransform1D::read: input stream failure");
        return new PowerTransform1D(a, p);
    }
}
