#include <cmath>
#include <cassert>
#include <limits>

#include "npstat/nm/MathUtils.hh"
#include "npstat/stat/SbMomentsCalculator.hh"

#define HG_MAXDEGREE 22

static long double calculate_mean(
    const unsigned maxdeg, const long double w, const long double q)
{
    long double coeff[HG_MAXDEGREE+1];
    coeff[0] = 0.L;
    coeff[1] = sqrtl(w);
    coeff[2] = -w*w;
    if (maxdeg >= 3) coeff[3] = powl(w,4.5L);
    if (maxdeg >= 4) coeff[4] = -powl(w,8.L);
    if (maxdeg >= 5) coeff[5] = powl(w,12.5L);
    if (maxdeg >= 6) coeff[6] = -powl(w,18.L);
    if (maxdeg >= 7) coeff[7] = powl(w,24.5L);
    if (maxdeg >= 8) coeff[8] = -powl(w,32.L);
    if (maxdeg >= 9) coeff[9] = powl(w,40.5L);
    if (maxdeg >= 10) coeff[10] = -powl(w,50.L);
    if (maxdeg >= 11) coeff[11] = powl(w,60.5L);
    if (maxdeg >= 12) coeff[12] = -powl(w,72.L);
    if (maxdeg >= 13) coeff[13] = powl(w,84.5L);
    if (maxdeg >= 14) coeff[14] = -powl(w,98.L);
    if (maxdeg >= 15) coeff[15] = powl(w,112.5L);
    if (maxdeg >= 16) coeff[16] = -powl(w,128.L);
    if (maxdeg >= 17) coeff[17] = powl(w,144.5L);
    if (maxdeg >= 18) coeff[18] = -powl(w,162.L);
    if (maxdeg >= 19) coeff[19] = powl(w,180.5L);
    if (maxdeg >= 20) coeff[20] = -powl(w,200.L);
    if (maxdeg >= 21) coeff[21] = powl(w,220.5L);
    if (maxdeg >= 22) coeff[22] = -powl(w,242.L);
    return npstat::polySeriesSum(coeff, maxdeg, q);
}

static long double calculate_variance(
    const unsigned maxdeg, const long double w, const long double q)
{
    long double coeff[HG_MAXDEGREE+1];
    coeff[0] = 0.L;
    coeff[1] = 0.L;
    coeff[2] = (-1 + w)*w;
    const long double wsq = w*w;
    const long double wquad = wsq*wsq;
    const long double wsix = wquad*wsq;
    if (maxdeg >= 3) coeff[3] = -2*powl(w,2.5)*(-1 + wsq);
    if (maxdeg >= 4) coeff[4] = wquad*(-1 - 2*w + 3*wquad);
    if (maxdeg >= 5) coeff[5] = 2*powl(w,6.5)*(1 + wsq - 2*wsix);
    if (maxdeg >= 6) coeff[6] = powl(w,9)*(-1 - 2*w - 2*wquad + 5*powl(w,9));
    if (maxdeg >= 7) coeff[7] = 2*powl(w,12.5)*
        (1 + wsq + wsix - 3*powl(w,12));
    if (maxdeg >= 8) coeff[8] = powl(w,16)*(-1 - 2*w - 2*wquad - 2*powl(w,9) + 
                                            7*powl(w,16));
    if (maxdeg >= 9) coeff[9] = 2*powl(w,20.5)*
        (1 + wsq + wsix + powl(w,12) - 4*powl(w,20));
    if (maxdeg >= 10) coeff[10] = powl(w,25)*
        (-1 - 2*w - 2*wquad - 2*powl(w,9) - 2*powl(w,16) + 9*powl(w,25));
    if (maxdeg >= 11) coeff[11] = 2*powl(w,30.5)*
        (1 + wsq + wsix + powl(w,12) + powl(w,20) - 5*powl(w,30));
    if (maxdeg >= 12) coeff[12] = powl(w,36)*
        (-1 - 2*w - 2*wquad - 2*powl(w,9) - 2*powl(w,16) - 2*powl(w,25) + 
         11*powl(w,36));
    if (maxdeg >= 13) coeff[13] = 2*powl(w,42.5)*
        (1 + wsq + wsix + powl(w,12) + powl(w,20) + powl(w,30) - 
         6*powl(w,42));
    if (maxdeg >= 14) coeff[14] = powl(w,49)*
        (-1 - 2*w - 2*wquad - 2*powl(w,9) - 2*powl(w,16) - 2*powl(w,25) - 
         2*powl(w,36) + 13*powl(w,49));
    if (maxdeg >= 15) coeff[15] = 2*powl(w,56.5)*
        (1 + wsq + wsix + powl(w,12) + powl(w,20) + powl(w,30) + 
         powl(w,42) - 7*powl(w,56));
    if (maxdeg >= 16) coeff[16] = powl(w,64)*
        (-1 - 2*w - 2*wquad - 2*powl(w,9) - 2*powl(w,16) - 2*powl(w,25) - 
         2*powl(w,36) - 2*powl(w,49) + 15*powl(w,64));
    if (maxdeg >= 17) coeff[17] = 2*powl(w,72.5)*
        (1 + wsq + wsix + powl(w,12) + powl(w,20) + powl(w,30) + 
         powl(w,42) + powl(w,56) - 8*powl(w,72));
    if (maxdeg >= 18) coeff[18] = powl(w,81)*
        (-1 - 2*w - 2*wquad - 2*powl(w,9) - 2*powl(w,16) - 2*powl(w,25) - 
         2*powl(w,36) - 2*powl(w,49) - 2*powl(w,64) + 17*powl(w,81));
    if (maxdeg >= 19) coeff[19] = 2*powl(w,90.5)*
        (1 + wsq + wsix + powl(w,12) + powl(w,20) + powl(w,30) + 
         powl(w,42) + powl(w,56) + powl(w,72) - 9*powl(w,90));
    if (maxdeg >= 20) coeff[20] = powl(w,100)*
        (-1 - 2*w - 2*wquad - 2*powl(w,9) - 2*powl(w,16) - 2*powl(w,25) - 
         2*powl(w,36) - 2*powl(w,49) - 2*powl(w,64) - 2*powl(w,81) + 
         19*powl(w,100));
    if (maxdeg >= 21) coeff[21] = 2*powl(w,110.5)*
        (1 + wsq + wsix + powl(w,12) + powl(w,20) + powl(w,30) + 
         powl(w,42) + powl(w,56) + powl(w,72) + powl(w,90) - 10*powl(w,110));
    if (maxdeg >= 22) coeff[22] = powl(w,121)*
        (-1 - 2*w - 2*wquad - 2*powl(w,9) - 2*powl(w,16) - 2*powl(w,25) - 
         2*powl(w,36) - 2*powl(w,49) - 2*powl(w,64) - 2*powl(w,81) - 
         2*powl(w,100) + 21*powl(w,121));
    return npstat::polySeriesSum(coeff, maxdeg, q);
}

static bool my_inf_nan(long double x)
{
    volatile long double temp = x;
    return ((temp - x) != 0.0L);
}

static unsigned best_poly_degree(
    const double** serCoeffs, const unsigned* serLen,
    const long double w, const long double q)
{
    long double qcoeff[HG_MAXDEGREE+1];

    for (unsigned i=2; i<=HG_MAXDEGREE; ++i)
    {
        qcoeff[i] = npstat::polySeriesSum(serCoeffs[i], serLen[i]-1, w);
        qcoeff[i] *= powl(w, i*1.5L+2.L);
        qcoeff[i] *= powl(q, i);
        qcoeff[i] = std::abs(qcoeff[i]);
        if (my_inf_nan(qcoeff[i]))
            return i - 1;
        if (i > 2)
        {
            if (qcoeff[i] < std::numeric_limits<long double>::min())
                return i - 1;
            if (qcoeff[i] > qcoeff[i - 1])
                return i - 1;
        }
    }
    return HG_MAXDEGREE;
}

static void calculate_series(
    const double** serCoeffs, const unsigned* serLen,
    const unsigned maxdeg, const long double w, const long double q,
    long double* res, long double* dresdw, long double* dresdq)
{
    long double qcoeff[HG_MAXDEGREE+1], qderiv[HG_MAXDEGREE+1];

    for (unsigned i=0; i<=maxdeg; ++i)
    {
        long double lead = 1.L, dlead = 0.L;
        if (i)
        {
            lead = powl(w, i*1.5L+2.L);
            dlead = (i*1.5L+2.L)*lead/w;
        }
        npstat::polyAndDeriv(serCoeffs[i], serLen[i]-1, w, qcoeff+i, qderiv+i);
        qderiv[i] = qderiv[i]*lead + qcoeff[i]*dlead;
        qcoeff[i] *= lead;
    }
    npstat::polyAndDeriv(qcoeff, maxdeg, q, res, dresdq);
    npstat::polyAndDeriv(qderiv, maxdeg, q, dresdw, qcoeff);
}


static void check_coeffs(const double** serCoeffs,
                         const unsigned* serLen)
{
    for (unsigned i=0; i<=HG_MAXDEGREE; ++i)
    {
        const unsigned len = serLen[i];
        const double* coeffs = serCoeffs[i];
        double sum = 0.0;
        for (unsigned j=0; j<len; ++j)
            sum += coeffs[j];
        assert(sum == 0.0);
    }
}

static bool coeffs_checked = false;

static const double hg_skewness0[] = {
    -4.0,
    0.0,
    3.0,
    1.0
};

static const double hg_skewness1[] = {
    12.0,
    6.0,
    -12.0,
    -6.0
};

static const double hg_skewness2[] = {
    30.0,
    21.0,
    -60.0,
    -66.0,
    18.0,
    45.0,
    12.0
};

static const double hg_skewness3[] = {
    60.0,
    60.0,
    -186.0,
    -288.0,
    74.0,
    372.0,
    168.0,
    -84.0,
    -96.0,
    -60.0,
    -20.0
};

static const double hg_skewness4[] = {
    105.0,
    150.0,
    -450.0,
    -927.0,
    222.0,
    1692.0,
    1026.0,
    -786.0,
    -1161.0,
    -420.0,
    18.0,
    171.0,
    150.0,
    90.0,
    90.0,
    30.0
};

static const double hg_skewness5[] = {
    168.0,
    336.0,
    -924.0,
    -2520.0,
    420.0,
    5970.0,
    4434.0,
    -4104.0,
    -7554.0,
    -2370.0,
    2502.0,
    2814.0,
    1422.0,
    252.0,
    -18.0,
    6.0,
    -198.0,
    -216.0,
    -126.0,
    -126.0,
    -126.0,
    -42.0
};

static const double hg_skewness6[] = {
    252.0,
    686.0,
    -1674.0,
    -6096.0,
    192.0,
    17682.0,
    16299.0,
    -15642.0,
    -35478.0,
    -11988.0,
    20964.0,
    24321.0,
    7614.0,
    -4812.0,
    -7086.0,
    -3921.0,
    -1466.0,
    -870.0,
    0.0,
    178.0,
    -282.0,
    -168.0,
    273.0,
    294.0,
    168.0,
    168.0,
    168.0,
    168.0,
    56.0
};

static const double hg_skewness7[] = {
    360.0,
    1296.0,
    -2724.0,
    -13476.0,
    -2184.0,
    45924.0,
    53064.0,
    -47142.0,
    -135858.0,
    -54702.0,
    109188.0,
    142938.0,
    33750.0,
    -63726.0,
    -72480.0,
    -32658.0,
    1692.0,
    12462.0,
    11922.0,
    9690.0,
    3570.0,
    -1086.0,
    -24.0,
    1044.0,
    48.0,
    -156.0,
    348.0,
    600.0,
    216.0,
    -360.0,
    -384.0,
    -216.0,
    -216.0,
    -216.0,
    -216.0,
    -216.0,
    -72.0
};

static const double hg_skewness8[] = {
    495.0,
    2295.0,
    -3996.0,
    -27672.0,
    -11358.0,
    107349.0,
    156327.0,
    -117168.0,
    -448854.0,
    -226758.0,
    438927.0,
    666792.0,
    142914.0,
    -441078.0,
    -486090.0,
    -154740.0,
    123144.0,
    185214.0,
    122406.0,
    54246.0,
    6159.0,
    -29637.0,
    -31743.0,
    -14118.0,
    -8331.0,
    -7521.0,
    -1311.0,
    3816.0,
    2853.0,
    -807.0,
    -2064.0,
    -396.0,
    252.0,
    -420.0,
    -756.0,
    -756.0,
    -270.0,
    459.0,
    486.0,
    270.0,
    270.0,
    270.0,
    270.0,
    270.0,
    270.0,
    90.0
};

static const double hg_skewness9[] = {
    660.0,
    3850.0,
    -5220.0,
    -53382.0,
    -38060.0,
    229434.0,
    422676.0,
    -239854.0,
    -1323204.0,
    -852810.0,
    1470358.0,
    2654868.0,
    627498.0,
    -2246940.0,
    -2589546.0,
    -567018.0,
    1242902.0,
    1466970.0,
    709860.0,
    3422.0,
    -297546.0,
    -357870.0,
    -288430.0,
    -123126.0,
    -8052.0,
    10296.0,
    27906.0,
    56856.0,
    51482.0,
    18612.0,
    -4086.0,
    -2278.0,
    4518.0,
    798.0,
    -5168.0,
    -4644.0,
    -1242.0,
    2314.0,
    3090.0,
    444.0,
    -366.0,
    498.0,
    930.0,
    930.0,
    930.0,
    330.0,
    -570.0,
    -600.0,
    -330.0,
    -330.0,
    -330.0,
    -330.0,
    -330.0,
    -330.0,
    -330.0,
    -110.0
};

static const double hg_skewness10[] = {
    858.0,
    6171.0,
    -5808.0,
    -97560.0,
    -104580.0,
    452478.0,
    1060014.0,
    -376884.0,
    -3551238.0,
    -2922150.0,
    4255974.0,
    9367752.0,
    2782209.0,
    -9352737.0,
    -11806938.0,
    -1938960.0,
    8063934.0,
    8823438.0,
    2973414.0,
    -2215644.0,
    -3602007.0,
    -2615247.0,
    -1222290.0,
    -43290.0,
    656658.0,
    659667.0,
    413430.0,
    349350.0,
    289803.0,
    87177.0,
    -92262.0,
    -113016.0,
    -51804.0,
    -33813.0,
    -59910.0,
    -56058.0,
    -16623.0,
    14094.0,
    20928.0,
    6306.0,
    -7323.0,
    -2760.0,
    5046.0,
    5106.0,
    2946.0,
    174.0,
    -3522.0,
    -3762.0,
    -492.0,
    498.0,
    -582.0,
    -1122.0,
    -1122.0,
    -1122.0,
    -1122.0,
    -396.0,
    693.0,
    726.0,
    396.0,
    396.0,
    396.0,
    396.0,
    396.0,
    396.0,
    396.0,
    396.0,
    132.0
};

static const double hg_skewness11[] = {
    1092.0,
    9516.0,
    -4686.0,
    -170016.0,
    -253746.0,
    826704.0,
    2486598.0,
    -297216.0,
    -8787834.0,
    -9183216.0,
    10808358.0,
    29957424.0,
    11742930.0,
    -33438432.0,
    -47798202.0,
    -7384518.0,
    40857888.0,
    44723460.0,
    10024908.0,
    -21385908.0,
    -26275062.0,
    -13577814.0,
    -533442.0,
    6091284.0,
    7489242.0,
    5385732.0,
    2023692.0,
    156732.0,
    -159906.0,
    -580236.0,
    -1174848.0,
    -1118178.0,
    -534018.0,
    -148320.0,
    -181200.0,
    -220392.0,
    -13380.0,
    213870.0,
    245574.0,
    133272.0,
    14046.0,
    -7878.0,
    36990.0,
    42342.0,
    10878.0,
    -13674.0,
    -27684.0,
    -24654.0,
    -1626.0,
    13650.0,
    5910.0,
    -5250.0,
    -6330.0,
    -3630.0,
    -1830.0,
    666.0,
    4236.0,
    4500.0,
    540.0,
    -648.0,
    672.0,
    1332.0,
    1332.0,
    1332.0,
    1332.0,
    1332.0,
    468.0,
    -828.0,
    -864.0,
    -468.0,
    -468.0,
    -468.0,
    -468.0,
    -468.0,
    -468.0,
    -468.0,
    -468.0,
    -468.0,
    -156.0
};

static const double hg_skewness12[] = {
    1365.0,
    14196.0,
    -78.0,
    -283971.0,
    -562931.0,
    1397418.0,
    5493786.0,
    766677.0,
    -20214966.0,
    -26673456.0,
    24054271.0,
    88081932.0,
    45703947.0,
    -105557892.0,
    -175522173.0,
    -32473188.0,
    174415087.0,
    200719128.0,
    30269139.0,
    -135656931.0,
    -150554658.0,
    -54800196.0,
    35179741.0,
    64372509.0,
    50031495.0,
    22287934.0,
    -2354235.0,
    -14200806.0,
    -12641134.0,
    -7754682.0,
    -5933457.0,
    -4209518.0,
    -628584.0,
    1954722.0,
    1671745.0,
    398028.0,
    459612.0,
    1423496.0,
    1602261.0,
    792363.0,
    -45834.0,
    -291714.0,
    -60306.0,
    85228.0,
    -84909.0,
    -269124.0,
    -291512.0,
    -205308.0,
    -44676.0,
    84616.0,
    71052.0,
    -14100.0,
    -43736.0,
    -15102.0,
    11028.0,
    21794.0,
    27363.0,
    20310.0,
    -3422.0,
    -19080.0,
    -8385.0,
    6410.0,
    7686.0,
    4386.0,
    2186.0,
    1086.0,
    -798.0,
    -5016.0,
    -5304.0,
    -588.0,
    816.0,
    -768.0,
    -1560.0,
    -1560.0,
    -1560.0,
    -1560.0,
    -1560.0,
    -1560.0,
    -546.0,
    975.0,
    1014.0,
    546.0,
    546.0,
    546.0,
    546.0,
    546.0,
    546.0,
    546.0,
    546.0,
    546.0,
    546.0,
    182.0
};

static const double hg_skewness13[] = {
    1680.0,
    20580.0,
    10764.0,
    -456456.0,
    -1165020.0,
    2165766.0,
    11495826.0,
    4871808.0,
    -43437846.0,
    -72151548.0,
    45736596.0,
    240442458.0,
    162870804.0,
    -298651284.0,
    -592771866.0,
    -148268712.0,
    652306566.0,
    818819832.0,
    97373508.0,
    -688809522.0,
    -743286906.0,
    -181835658.0,
    338956242.0,
    442539486.0,
    247638966.0,
    18649140.0,
    -114865266.0,
    -142510998.0,
    -95517444.0,
    -31516596.0,
    2268726.0,
    10544718.0,
    17960196.0,
    24056496.0,
    17854476.0,
    4711902.0,
    -1546536.0,
    1353114.0,
    3970188.0,
    1100406.0,
    -3061932.0,
    -3876864.0,
    -1822728.0,
    -71106.0,
    -381654.0,
    -1516332.0,
    -1659816.0,
    -869874.0,
    113244.0,
    809580.0,
    782406.0,
    231468.0,
    -100254.0,
    26142.0,
    230118.0,
    255468.0,
    177114.0,
    69924.0,
    -63834.0,
    -151596.0,
    -104274.0,
    19356.0,
    64752.0,
    28626.0,
    -5466.0,
    -18516.0,
    -20706.0,
    -24408.0,
    -19200.0,
    7416.0,
    24306.0,
    9906.0,
    -7686.0,
    -9174.0,
    -5214.0,
    -2574.0,
    -1254.0,
    -1254.0,
    942.0,
    5862.0,
    6174.0,
    636.0,
    -1002.0,
    870.0,
    1806.0,
    1806.0,
    1806.0,
    1806.0,
    1806.0,
    1806.0,
    1806.0,
    630.0,
    -1134.0,
    -1176.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -630.0,
    -210.0
};

static const double hg_skewness14[] = {
    2040.0,
    29100.0,
    31890.0,
    -708402.0,
    -2278692.0,
    3004476.0,
    22885731.0,
    16882728.0,
    -87363132.0,
    -182999592.0,
    67830615.0,
    613480605.0,
    533596728.0,
    -761670696.0,
    -1858660299.0,
    -641727960.0,
    2185983948.0,
    3082012248.0,
    388842177.0,
    -3001059876.0,
    -3304853028.0,
    -527362272.0,
    2146508700.0,
    2460147603.0,
    987043446.0,
    -503683908.0,
    -1083524949.0,
    -894236676.0,
    -389061750.0,
    68499132.0,
    267967092.0,
    235674495.0,
    152310888.0,
    106533066.0,
    55139472.0,
    -17707653.0,
    -58649766.0,
    -41763108.0,
    -8826411.0,
    -3726507.0,
    -17644302.0,
    -19918092.0,
    -5854542.0,
    6850953.0,
    6514890.0,
    -1448496.0,
    -5276475.0,
    -1911393.0,
    3201846.0,
    5865036.0,
    5010123.0,
    1636299.0,
    -853440.0,
    -441150.0,
    1053264.0,
    1368384.0,
    541794.0,
    -377736.0,
    -988911.0,
    -1207392.0,
    -847992.0,
    -103380.0,
    341586.0,
    182328.0,
    -115638.0,
    -190626.0,
    -131544.0,
    -66432.0,
    -5958.0,
    101574.0,
    185484.0,
    107688.0,
    -46026.0,
    -92742.0,
    -42756.0,
    2340.0,
    20010.0,
    20730.0,
    16770.0,
    22779.0,
    18426.0,
    -10812.0,
    -28518.0,
    -11553.0,
    9078.0,
    10794.0,
    6114.0,
    2994.0,
    1434.0,
    1434.0,
    1434.0,
    -1098.0,
    -6774.0,
    -7110.0,
    -684.0,
    1206.0,
    -978.0,
    -2070.0,
    -2070.0,
    -2070.0,
    -2070.0,
    -2070.0,
    -2070.0,
    -2070.0,
    -2070.0,
    -720.0,
    1305.0,
    1350.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    720.0,
    240.0
};

static const double hg_skewness15[] = {
    2448.0,
    40256.0,
    69048.0,
    -1064220.0,
    -4250800.0,
    3496164.0,
    43498260.0,
    47333962.0,
    -164252202.0,
    -437745210.0,
    46983976.0,
    1469672964.0,
    1619088264.0,
    -1744328160.0,
    -5448171468.0,
    -2547730650.0,
    6651198228.0,
    10802560386.0,
    1813685106.0,
    -11611543804.0,
    -13545058170.0,
    -1510506096.0,
    10997314726.0,
    11967461262.0,
    3306816792.0,
    -5043994664.0,
    -7139833584.0,
    -4323461574.0,
    -484695916.0,
    1970968794.0,
    2475869538.0,
    1609263984.0,
    513673128.0,
    -65926872.0,
    -257482816.0,
    -406163664.0,
    -473238162.0,
    -310808202.0,
    -49169634.0,
    67783722.0,
    18800104.0,
    -25453272.0,
    20623878.0,
    80859898.0,
    73494720.0,
    15653262.0,
    -24664758.0,
    -17598318.0,
    9758232.0,
    23356552.0,
    16692204.0,
    -230976.0,
    -12505214.0,
    -9703044.0,
    1297842.0,
    6053348.0,
    1651050.0,
    -4526016.0,
    -7248140.0,
    -6538116.0,
    -3593070.0,
    298564.0,
    2669232.0,
    1911564.0,
    -208890.0,
    -949998.0,
    -239850.0,
    519164.0,
    837024.0,
    1006914.0,
    1094116.0,
    642762.0,
    -258162.0,
    -707654.0,
    -390258.0,
    51666.0,
    192022.0,
    151566.0,
    72750.0,
    27598.0,
    -7098.0,
    -117174.0,
    -197060.0,
    -101730.0,
    67926.0,
    116646.0,
    55266.0,
    -306.0,
    -23706.0,
    -24486.0,
    -15126.0,
    -12646.0,
    -21684.0,
    -18798.0,
    12710.0,
    33066.0,
    13326.0,
    -10586.0,
    -12546.0,
    -7086.0,
    -3446.0,
    -1626.0,
    -1626.0,
    -1626.0,
    -1626.0,
    1266.0,
    7752.0,
    8112.0,
    732.0,
    -1428.0,
    1092.0,
    2352.0,
    2352.0,
    2352.0,
    2352.0,
    2352.0,
    2352.0,
    2352.0,
    2352.0,
    2352.0,
    816.0,
    -1488.0,
    -1536.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -816.0,
    -272.0
};

static const double hg_skewness16[] = {
    2907.0,
    54621.0,
    130152.0,
    -1550616.0,
    -7614039.0,
    2645238.0,
    79138023.0,
    117527946.0,
    -286993716.0,
    -992321007.0,
    -160585428.0,
    3314503260.0,
    4584551538.0,
    -3518731797.0,
    -15005395110.0,
    -9251750322.0,
    18499851249.0,
    35478205248.0,
    8485844901.0,
    -40689118518.0,
    -51826901238.0,
    -5191986600.0,
    48838054029.0,
    52978727424.0,
    9680313801.0,
    -32377167981.0,
    -38928444885.0,
    -17288828820.0,
    6641119176.0,
    17409437493.0,
    15090197283.0,
    6509410626.0,
    -1324875144.0,
    -4569247617.0,
    -4043314644.0,
    -2706553956.0,
    -1792685391.0,
    -710990190.0,
    580628013.0,
    1177905894.0,
    781610580.0,
    188418690.0,
    105719181.0,
    324789630.0,
    310975845.0,
    6850716.0,
    -242855961.0,
    -221773800.0,
    -46043133.0,
    65803773.0,
    46351791.0,
    -30738390.0,
    -78594789.0,
    -55625835.0,
    10698963.0,
    47569140.0,
    25523409.0,
    -15884607.0,
    -33924975.0,
    -24898446.0,
    -5330958.0,
    12152751.0,
    19712553.0,
    13338240.0,
    117390.0,
    -6130137.0,
    -1679334.0,
    4810992.0,
    6747084.0,
    5486826.0,
    3701190.0,
    1070940.0,
    -2785209.0,
    -5014785.0,
    -3242028.0,
    -9600.0,
    1190151.0,
    527139.0,
    -277308.0,
    -592896.0,
    -588483.0,
    -691296.0,
    -835158.0,
    -368010.0,
    545271.0,
    923817.0,
    494004.0,
    -55830.0,
    -255825.0,
    -222093.0,
    -113394.0,
    -27372.0,
    -7356.0,
    19452.0,
    124638.0,
    208368.0,
    104184.0,
    -88710.0,
    -141504.0,
    -67038.0,
    624.0,
    27714.0,
    28554.0,
    17634.0,
    9444.0,
    9144.0,
    22023.0,
    21306.0,
    -14760.0,
    -37950.0,
    -15225.0,
    12210.0,
    14430.0,
    8130.0,
    3930.0,
    1830.0,
    1830.0,
    1830.0,
    1830.0,
    1830.0,
    -1446.0,
    -8796.0,
    -9180.0,
    -780.0,
    1668.0,
    -1212.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -2652.0,
    -918.0,
    1683.0,
    1734.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    918.0,
    306.0
};

static const double hg_skewness17[] = {
    3420.0,
    72846.0,
    225828.0,
    -2194326.0,
    -13163448.0,
    -1608108.0,
    138045804.0,
    268162992.0,
    -459478920.0,
    -2140031304.0,
    -948723342.0,
    7042229754.0,
    12195526206.0,
    -5904923700.0,
    -38976612684.0,
    -30937812588.0,
    47089544604.0,
    109677940524.0,
    36866529180.0,
    -130670263728.0,
    -186467029806.0,
    -22834778088.0,
    194300147346.0,
    218061133932.0,
    26689863738.0,
    -168306560634.0,
    -187817049078.0,
    -59250700776.0,
    71485685754.0,
    111597327594.0,
    72603038430.0,
    10609284084.0,
    -31731252204.0,
    -40587020604.0,
    -26089694736.0,
    -8270070702.0,
    1425896364.0,
    5238116352.0,
    7786207578.0,
    8159215248.0,
    4675220832.0,
    123374148.0,
    -1559778714.0,
    -487843002.0,
    298396068.0,
    -542337072.0,
    -1597725996.0,
    -1418276886.0,
    -326992752.0,
    466177452.0,
    434899512.0,
    -2640264.0,
    -266991942.0,
    -156154590.0,
    162185250.0,
    348925806.0,
    222799248.0,
    -43858332.0,
    -178463406.0,
    -123074298.0,
    501570.0,
    83119830.0,
    94460964.0,
    46855908.0,
    -21214944.0,
    -53231904.0,
    -26401902.0,
    19467900.0,
    36307056.0,
    22791234.0,
    4054470.0,
    -9598032.0,
    -21266424.0,
    -26578896.0,
    -16427082.0,
    2092224.0,
    10332252.0,
    4586388.0,
    -3166080.0,
    -5327310.0,
    -3570540.0,
    -1593120.0,
    -705516.0,
    1018404.0,
    4380636.0,
    5892990.0,
    3291228.0,
    -514926.0,
    -2009214.0,
    -1333980.0,
    -294852.0,
    300846.0,
    400578.0,
    350820.0,
    539298.0,
    737088.0,
    287154.0,
    -672996.0,
    -1075788.0,
    -549486.0,
    116484.0,
    351312.0,
    293946.0,
    152196.0,
    35268.0,
    -17184.0,
    -16170.0,
    -17928.0,
    -124938.0,
    -223392.0,
    -107202.0,
    109764.0,
    166056.0,
    76686.0,
    -984.0,
    -32034.0,
    -32934.0,
    -20334.0,
    -10884.0,
    -4584.0,
    -7254.0,
    -24876.0,
    -23970.0,
    16962.0,
    43170.0,
    17250.0,
    -13950.0,
    -16446.0,
    -9246.0,
    -4446.0,
    -2046.0,
    -2046.0,
    -2046.0,
    -2046.0,
    -2046.0,
    -2046.0,
    1638.0,
    9906.0,
    10314.0,
    828.0,
    -1926.0,
    1338.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    2970.0,
    1026.0,
    -1890.0,
    -1944.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -1026.0,
    -342.0
};

static const double hg_skewness18[] = {
    3990.0,
    95665.0,
    370044.0,
    -3018390.0,
    -22055562.0,
    -13173297.0,
    231024780.0,
    573258878.0,
    -652170000.0,
    -4403616876.0,
    -3279527306.0,
    14073058626.0,
    30648891327.0,
    -6624796772.0,
    -95722245642.0,
    -96078701196.0,
    109028322063.0,
    320282353203.0,
    146199697428.0,
    -387189815220.0,
    -633630895140.0,
    -110264743632.0,
    705461323810.0,
    844513487238.0,
    81064137654.0,
    -761939181471.0,
    -830471885874.0,
    -179069336982.0,
    469940901930.0,
    600973184715.0,
    294112860570.0,
    -80561292362.0,
    -269734614594.0,
    -244308818556.0,
    -106267475028.0,
    21348472896.0,
    73766318253.0,
    66858686031.0,
    46029568158.0,
    28559655468.0,
    7535899442.0,
    -14060902353.0,
    -21556469214.0,
    -12673945644.0,
    -2047788864.0,
    -816137613.0,
    -4681550744.0,
    -4559915724.0,
    431591055.0,
    4480498396.0,
    4014124188.0,
    973472172.0,
    -1045979676.0,
    -747132066.0,
    743608020.0,
    1636016988.0,
    1018120323.0,
    -412680306.0,
    -1212460690.0,
    -865219830.0,
    -57517404.0,
    424218756.0,
    397613196.0,
    89710308.0,
    -221162141.0,
    -323049084.0,
    -152962236.0,
    114067668.0,
    224632290.0,
    130270470.0,
    -14085442.0,
    -90200460.0,
    -108343881.0,
    -96792996.0,
    -42612492.0,
    39795486.0,
    78972068.0,
    40766466.0,
    -18179838.0,
    -36959362.0,
    -18939834.0,
    3200793.0,
    13334390.0,
    17124096.0,
    23409672.0,
    25458981.0,
    12007926.0,
    -8130300.0,
    -16116851.0,
    -9454803.0,
    -184764.0,
    3590376.0,
    2650530.0,
    493857.0,
    -307504.0,
    59106.0,
    -1080453.0,
    -4277493.0,
    -5802906.0,
    -2898258.0,
    1500931.0,
    3174249.0,
    2155128.0,
    669704.0,
    -208989.0,
    -466761.0,
    -395880.0,
    -319482.0,
    -503574.0,
    -753633.0,
    -295596.0,
    782052.0,
    1220984.0,
    587862.0,
    -180870.0,
    -433270.0,
    -356634.0,
    -188952.0,
    -49834.0,
    24810.0,
    56730.0,
    17580.0,
    66.0,
    125178.0,
    240186.0,
    112044.0,
    -129702.0,
    -188790.0,
    -86982.0,
    1386.0,
    36666.0,
    37626.0,
    23226.0,
    12426.0,
    5226.0,
    1626.0,
    8146.0,
    27903.0,
    26790.0,
    -19316.0,
    -48726.0,
    -19401.0,
    15806.0,
    18594.0,
    10434.0,
    4994.0,
    2274.0,
    2274.0,
    2274.0,
    2274.0,
    2274.0,
    2274.0,
    2274.0,
    -1842.0,
    -11082.0,
    -11514.0,
    -876.0,
    2202.0,
    -1470.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -3306.0,
    -1140.0,
    2109.0,
    2166.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    1140.0,
    380.0
};

static const double hg_skewness19[] = {
    4620.0,
    123900.0,
    580830.0,
    -4036512.0,
    -35934174.0,
    -39034692.0,
    370695876.0,
    1161621180.0,
    -748661952.0,
    -8663993184.0,
    -9317907864.0,
    26316613068.0,
    73098687324.0,
    3359171376.0,
    -222554792622.0,
    -279264313518.0,
    225389086116.0,
    885930992148.0,
    531387940992.0,
    -1061399916750.0,
    -2039731840428.0,
    -511389853854.0,
    2363737882356.0,
    3097657212858.0,
    315167237664.0,
    -3106577574570.0,
    -3433607108010.0,
    -500553810150.0,
    2488824362148.0,
    2887068476898.0,
    1036657074468.0,
    -971940274986.0,
    -1700033311182.0,
    -1180107681558.0,
    -211590872088.0,
    485076262890.0,
    638853225804.0,
    413960805954.0,
    135275437452.0,
    -24704463078.0,
    -97655978586.0,
    -140517898086.0,
    -135084238050.0,
    -66597317934.0,
    10184598492.0,
    33755306160.0,
    12417112932.0,
    -2436196920.0,
    9769052142.0,
    25394955876.0,
    21175712778.0,
    2769221448.0,
    -10156404216.0,
    -8929774884.0,
    -192484110.0,
    5536260444.0,
    3541913118.0,
    -2921470110.0,
    -6826029720.0,
    -4791370416.0,
    46272666.0,
    2885573832.0,
    2383376580.0,
    330000222.0,
    -1234352322.0,
    -1429857234.0,
    -415219494.0,
    898403112.0,
    1394871540.0,
    775375608.0,
    -178660698.0,
    -637467132.0,
    -554493564.0,
    -293964204.0,
    -7988010.0,
    285522414.0,
    405113178.0,
    195496668.0,
    -141983382.0,
    -265536270.0,
    -131736186.0,
    45244146.0,
    117580158.0,
    105740490.0,
    83569254.0,
    64443708.0,
    9012246.0,
    -71523510.0,
    -101063976.0,
    -52972044.0,
    12059562.0,
    35842884.0,
    20060472.0,
    -3870216.0,
    -15415542.0,
    -13906350.0,
    -11705772.0,
    -16908990.0,
    -19573536.0,
    -6478290.0,
    14684766.0,
    23144694.0,
    14149242.0,
    1523238.0,
    -4206522.0,
    -3767094.0,
    -1374126.0,
    160698.0,
    -16854.0,
    -883116.0,
    152202.0,
    3825168.0,
    5826192.0,
    2627010.0,
    -2316996.0,
    -4002030.0,
    -2606094.0,
    -813024.0,
    242676.0,
    599166.0,
    626346.0,
    471630.0,
    255042.0,
    439422.0,
    774084.0,
    275454.0,
    -910722.0,
    -1357950.0,
    -623982.0,
    235878.0,
    515934.0,
    422550.0,
    225942.0,
    61902.0,
    -28818.0,
    -75618.0,
    -62358.0,
    3882.0,
    18678.0,
    -126618.0,
    -260136.0,
    -120222.0,
    146886.0,
    212982.0,
    97926.0,
    -1830.0,
    -41610.0,
    -42630.0,
    -26310.0,
    -14070.0,
    -5910.0,
    -1830.0,
    -1830.0,
    -9090.0,
    -31104.0,
    -29766.0,
    21822.0,
    54618.0,
    21678.0,
    -17778.0,
    -20874.0,
    -11694.0,
    -5574.0,
    -2514.0,
    -2514.0,
    -2514.0,
    -2514.0,
    -2514.0,
    -2514.0,
    -2514.0,
    -2514.0,
    2058.0,
    12324.0,
    12780.0,
    924.0,
    -2496.0,
    1608.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    3660.0,
    1260.0,
    -2340.0,
    -2400.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -1260.0,
    -420.0
};

static const double hg_skewness20[] = {
    5313.0,
    158466.0,
    881094.0,
    -5244975.0,
    -57086640.0,
    -91107252.0,
    568950762.0,
    2248410120.0,
    -439119351.0,
    -16316864190.0,
    -23627960679.0,
    45533629875.0,
    166046758338.0,
    51560482824.0,
    -489802674948.0,
    -764704087701.0,
    395942129481.0,
    2326001951346.0,
    1785999624789.0,
    -2687234841132.0,
    -6234734104782.0,
    -2178790961004.0,
    7357600559658.0,
    10800553240134.0,
    1471613662521.0,
    -11619985496544.0,
    -13428738573624.0,
    -1437090623952.0,
    11492341282098.0,
    12776065371378.0,
    3257607316899.0,
    -6604411383804.0,
    -9091005681138.0,
    -4863919618896.0,
    873715223577.0,
    4050883549563.0,
    3829541821134.0,
    1706770364724.0,
    -296028670473.0,
    -1138060403835.0,
    -1073220980916.0,
    -763464346038.0,
    -448833330210.0,
    -65382243798.0,
    284829377925.0,
    372242674572.0,
    199630536354.0,
    19318167042.0,
    -600811755.0,
    60411799908.0,
    57081691893.0,
    -21836661975.0,
    -82927822116.0,
    -69596682906.0,
    -13026144624.0,
    26371281546.0,
    22290797535.0,
    -7370813580.0,
    -27276028938.0,
    -18683260833.0,
    5625296823.0,
    20111980266.0,
    15248667141.0,
    1572698418.0,
    -7372164384.0,
    -7102886328.0,
    -979850055.0,
    5235729348.0,
    6925992453.0,
    3254606226.0,
    -2051617671.0,
    -4465283796.0,
    -3243489267.0,
    -874397904.0,
    757715313.0,
    1566935481.0,
    1600461876.0,
    552700488.0,
    -978648549.0,
    -1556154057.0,
    -730143825.0,
    426821016.0,
    864292467.0,
    607509852.0,
    226039395.0,
    7954350.0,
    -166770309.0,
    -386320311.0,
    -443362485.0,
    -181151256.0,
    171225123.0,
    289281921.0,
    150113643.0,
    -40409784.0,
    -126688725.0,
    -104550180.0,
    -52007397.0,
    -31080600.0,
    -26814303.0,
    16924419.0,
    92842386.0,
    121258050.0,
    64587993.0,
    -16173102.0,
    -48986652.0,
    -31642704.0,
    -2502663.0,
    12904920.0,
    11900670.0,
    3861744.0,
    1360686.0,
    9553998.0,
    16111935.0,
    4129266.0,
    -18738546.0,
    -27144702.0,
    -15353196.0,
    -227178.0,
    6320088.0,
    5844636.0,
    3370836.0,
    1485624.0,
    363708.0,
    413160.0,
    1445136.0,
    225756.0,
    -4127514.0,
    -6318276.0,
    -2619900.0,
    2827632.0,
    4580748.0,
    2956716.0,
    889500.0,
    -336780.0,
    -777888.0,
    -826968.0,
    -683184.0,
    -326292.0,
    -76110.0,
    -367155.0,
    -784590.0,
    -266874.0,
    1012746.0,
    1488762.0,
    667146.0,
    -294252.0,
    -602376.0,
    -489846.0,
    -261150.0,
    -69288.0,
    33120.0,
    85854.0,
    82488.0,
    42408.0,
    -26040.0,
    -36918.0,
    130770.0,
    284880.0,
    133500.0,
    -165138.0,
    -238632.0,
    -109518.0,
    2316.0,
    46866.0,
    47946.0,
    29586.0,
    15816.0,
    6636.0,
    2046.0,
    2046.0,
    2046.0,
    10086.0,
    34479.0,
    32898.0,
    -24480.0,
    -60846.0,
    -24081.0,
    19866.0,
    23286.0,
    13026.0,
    6186.0,
    2766.0,
    2766.0,
    2766.0,
    2766.0,
    2766.0,
    2766.0,
    2766.0,
    2766.0,
    2766.0,
    -2286.0,
    -13632.0,
    -14112.0,
    -972.0,
    2808.0,
    -1752.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -4032.0,
    -1386.0,
    2583.0,
    2646.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    1386.0,
    462.0
};

static const double hg_skewness21[] = {
    6072.0,
    200376.0,
    1299540.0,
    -6611496.0,
    -88634412.0,
    -188975646.0,
    831058674.0,
    4179278632.0,
    978457806.0,
    -29416145754.0,
    -55315089030.0,
    71133762894.0,
    360188779614.0,
    211014827544.0,
    -1018307613918.0,
    -1983173889750.0,
    499334830082.0,
    5804005677522.0,
    5599471327020.0,
    -6233725021846.0,
    -18129403917294.0,
    -8496332895486.0,
    21346146215174.0,
    35875516831650.0,
    7019391147420.0,
    -40315978841994.0,
    -50001875200962.0,
    -4930171951584.0,
    47930077306446.0,
    53075355108108.0,
    9439153513194.0,
    -35758280139294.0,
    -43612668371736.0,
    -17652520183908.0,
    12707238509086.0,
    25326865241802.0,
    18657682448826.0,
    3964829488212.0,
    -7096278038988.0,
    -9739992427242.0,
    -6453940500984.0,
    -2233724740404.0,
    360416301810.0,
    1696698143454.0,
    2403016623150.0,
    2157193722024.0,
    918518891016.0,
    -318038043732.0,
    -630416936814.0,
    -243862626844.0,
    9723726072.0,
    -170598288858.0,
    -400886476624.0,
    -316554549270.0,
    -7232058918.0,
    212679595812.0,
    190780177686.0,
    25761634542.0,
    -94739666722.0,
    -73141335444.0,
    35514835782.0,
    104483117028.0,
    73793279760.0,
    -6019454724.0,
    -54859310482.0,
    -45860582112.0,
    -6241720908.0,
    26078792512.0,
    30944654256.0,
    10712096172.0,
    -15311097790.0,
    -25854360078.0,
    -16359857124.0,
    -338265362.0,
    8720876268.0,
    9382008120.0,
    6014998182.0,
    561348846.0,
    -5309961426.0,
    -7159739634.0,
    -2779449126.0,
    3461755596.0,
    5698756546.0,
    3373522356.0,
    135741810.0,
    -1425493502.0,
    -1637879076.0,
    -1658190876.0,
    -1413910360.0,
    -259907310.0,
    1232411964.0,
    1663024734.0,
    743068296.0,
    -446924046.0,
    -921012288.0,
    -644386368.0,
    -155081676.0,
    114392218.0,
    151264956.0,
    211797984.0,
    391311058.0,
    445541454.0,
    171689808.0,
    -224627448.0,
    -374102118.0,
    -211190982.0,
    26262640.0,
    137499186.0,
    110497656.0,
    31420908.0,
    -18820188.0,
    -11746638.0,
    11848712.0,
    -17396730.0,
    -96027708.0,
    -127610808.0,
    -60297054.0,
    33424728.0,
    68350062.0,
    45839814.0,
    12489282.0,
    -5181484.0,
    -8027778.0,
    -3860124.0,
    2781400.0,
    3182592.0,
    -9155076.0,
    -18655112.0,
    -5223828.0,
    20085336.0,
    28553632.0,
    15064692.0,
    -1694928.0,
    -8862104.0,
    -8180532.0,
    -5185542.0,
    -2737156.0,
    -875946.0,
    419904.0,
    -80490.0,
    -1503876.0,
    -115740.0,
    4625582.0,
    6914424.0,
    2773524.0,
    -3265610.0,
    -5210526.0,
    -3357384.0,
    -993298.0,
    447192.0,
    965250.0,
    993028.0,
    829848.0,
    489336.0,
    101146.0,
    -72936.0,
    300912.0,
    814636.0,
    301968.0,
    -1091220.0,
    -1626602.0,
    -713418.0,
    354144.0,
    690580.0,
    556338.0,
    292224.0,
    77088.0,
    -37716.0,
    -96738.0,
    -92850.0,
    -60720.0,
    -22296.0,
    47382.0,
    53016.0,
    -139398.0,
    -316308.0,
    -147468.0,
    184458.0,
    265740.0,
    121758.0,
    -2844.0,
    -52434.0,
    -53574.0,
    -33054.0,
    -17664.0,
    -7404.0,
    -2274.0,
    -2274.0,
    -2274.0,
    -2274.0,
    -11134.0,
    -38028.0,
    -36186.0,
    27290.0,
    67410.0,
    26610.0,
    -22070.0,
    -25830.0,
    -14430.0,
    -6830.0,
    -3030.0,
    -3030.0,
    -3030.0,
    -3030.0,
    -3030.0,
    -3030.0,
    -3030.0,
    -3030.0,
    -3030.0,
    -3030.0,
    2526.0,
    15006.0,
    15510.0,
    1020.0,
    -3138.0,
    1902.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    4422.0,
    1518.0,
    -2838.0,
    -2904.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -1518.0,
    -506.0
};

static const double hg_skewness22[] = {
    6900.0,
    250746.0,
    1871694.0,
    -8060316.0,
    -134760960.0,
    -363847614.0,
    1143976203.0,
    7488546654.0,
    4949636610.0,
    -50696553438.0,
    -121675913022.0,
    94438017867.0,
    747448000110.0,
    658610953956.0,
    -1989951398952.0,
    -4891453220775.0,
    -10886436450.0,
    13769880780534.0,
    16497941128869.0,
    -12989434120572.0,
    -50224515544062.0,
    -30553271122176.0,
    57732756194268.0,
    113696150535366.0,
    31449119516772.0,
    -130616107014366.0,
    -177875152357506.0,
    -21007049900607.0,
    184018314770922.0,
    209343677054196.0,
    27000001814850.0,
    -168541702465254.0,
    -193432463348088.0,
    -57687006956910.0,
    90207148310787.0,
    135107310214515.0,
    78370810561416.0,
    -7722743626416.0,
    -59199119326980.0,
    -58485326222121.0,
    -27038931202440.0,
    3504025020324.0,
    16937899548339.0,
    16823274052458.0,
    12386491331748.0,
    6950920563636.0,
    292840850532.0,
    -5206616283822.0,
    -6090797001114.0,
    -2986972172622.0,
    -67057831008.0,
    207556290867.0,
    -773288397108.0,
    -737024651736.0,
    478383766233.0,
    1421180127201.0,
    1178940350712.0,
    202941158112.0,
    -516832280700.0,
    -479560802025.0,
    38549591472.0,
    406685403354.0,
    285602427579.0,
    -109810041510.0,
    -350567979732.0,
    -266182651992.0,
    -18661850763.0,
    156853389111.0,
    161829279042.0,
    45026273214.0,
    -81128323659.0,
    -122702816073.0,
    -65101204920.0,
    23723809422.0,
    67816064247.0,
    54345697959.0,
    18938133120.0,
    -10909077564.0,
    -29174376948.0,
    -29757643434.0,
    -7860404160.0,
    20979918360.0,
    30626286456.0,
    15426706677.0,
    -5534461146.0,
    -14500852746.0,
    -11593393509.0,
    -6102885423.0,
    -2313023166.0,
    1842597744.0,
    6498273981.0,
    7293599322.0,
    2380167960.0,
    -3715872726.0,
    -5714366121.0,
    -3283026033.0,
    220152768.0,
    1993157130.0,
    1823131188.0,
    1181899266.0,
    1102750680.0,
    1049178102.0,
    50634456.0,
    -1460851200.0,
    -1970167320.0,
    -944943066.0,
    501576501.0,
    1116787896.0,
    772103670.0,
    115560324.0,
    -287838177.0,
    -301279416.0,
    -138502062.0,
    -119252682.0,
    -316683534.0,
    -415599198.0,
    -132315132.0,
    311385000.0,
    466107663.0,
    260045076.0,
    -14187924.0,
    -133782636.0,
    -108188310.0,
    -36161934.0,
    22457472.0,
    43036620.0,
    11387517.0,
    -31392060.0,
    -1919268.0,
    89591184.0,
    124283625.0,
    49074240.0,
    -51628194.0,
    -86775798.0,
    -59456655.0,
    -20227392.0,
    2095884.0,
    8551440.0,
    8228364.0,
    3349944.0,
    -4144524.0,
    -2685138.0,
    12788796.0,
    23083392.0,
    7642176.0,
    -20627556.0,
    -30381564.0,
    -15784830.0,
    2686284.0,
    10731282.0,
    9994158.0,
    6351150.0,
    3262560.0,
    1163310.0,
    -516648.0,
    -1220610.0,
    -248052.0,
    1351572.0,
    43704.0,
    -4919646.0,
    -7413468.0,
    -2877732.0,
    3814590.0,
    5920434.0,
    3775968.0,
    1053252.0,
    -581526.0,
    -1140498.0,
    -1161252.0,
    -968040.0,
    -595656.0,
    -228852.0,
    84210.0,
    221040.0,
    -259299.0,
    -912144.0,
    -379728.0,
    1174032.0,
    1773318.0,
    764814.0,
    -413370.0,
    -778194.0,
    -619506.0,
    -325044.0,
    -85302.0,
    42606.0,
    108270.0,
    103824.0,
    67914.0,
    39186.0,
    3660.0,
    -66144.0,
    -65082.0,
    154518.0,
    349380.0,
    162126.0,
    -204846.0,
    -294306.0,
    -134646.0,
    3414.0,
    58314.0,
    59514.0,
    36714.0,
    19614.0,
    8214.0,
    2514.0,
    2514.0,
    2514.0,
    2514.0,
    2514.0,
    12234.0,
    41751.0,
    39630.0,
    -30252.0,
    -74310.0,
    -29265.0,
    24390.0,
    28506.0,
    15906.0,
    7506.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    3306.0,
    -2778.0,
    -16446.0,
    -16974.0,
    -1068.0,
    3486.0,
    -2058.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -4830.0,
    -1656.0,
    3105.0,
    3174.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    1656.0,
    552.0
};

static const double* hg_skewness[HG_MAXDEGREE+1] = {
    hg_skewness0,
    hg_skewness1,
    hg_skewness2,
    hg_skewness3,
    hg_skewness4,
    hg_skewness5,
    hg_skewness6,
    hg_skewness7,
    hg_skewness8,
    hg_skewness9,
    hg_skewness10,
    hg_skewness11,
    hg_skewness12,
    hg_skewness13,
    hg_skewness14,
    hg_skewness15,
    hg_skewness16,
    hg_skewness17,
    hg_skewness18,
    hg_skewness19,
    hg_skewness20,
    hg_skewness21,
    hg_skewness22
};

static const unsigned hg_skewness_len[HG_MAXDEGREE+1] = {
    sizeof(hg_skewness0)/sizeof(double),
    sizeof(hg_skewness1)/sizeof(double),
    sizeof(hg_skewness2)/sizeof(double),
    sizeof(hg_skewness3)/sizeof(double),
    sizeof(hg_skewness4)/sizeof(double),
    sizeof(hg_skewness5)/sizeof(double),
    sizeof(hg_skewness6)/sizeof(double),
    sizeof(hg_skewness7)/sizeof(double),
    sizeof(hg_skewness8)/sizeof(double),
    sizeof(hg_skewness9)/sizeof(double),
    sizeof(hg_skewness10)/sizeof(double),
    sizeof(hg_skewness11)/sizeof(double),
    sizeof(hg_skewness12)/sizeof(double),
    sizeof(hg_skewness13)/sizeof(double),
    sizeof(hg_skewness14)/sizeof(double),
    sizeof(hg_skewness15)/sizeof(double),
    sizeof(hg_skewness16)/sizeof(double),
    sizeof(hg_skewness17)/sizeof(double),
    sizeof(hg_skewness18)/sizeof(double),
    sizeof(hg_skewness19)/sizeof(double),
    sizeof(hg_skewness20)/sizeof(double),
    sizeof(hg_skewness21)/sizeof(double),
    sizeof(hg_skewness22)/sizeof(double)
};

static const double hg_excess0[] = {
    -6.0,
    0.0,
    3.0,
    2.0,
    1.0
};

static const double hg_excess1[] = {
    12.0,
    12.0,
    -4.0,
    -8.0,
    -8.0,
    -4.0
};

static const double hg_excess2[] = {
    30.0,
    42.0,
    -24.0,
    -78.0,
    -50.0,
    -4.0,
    24.0,
    30.0,
    20.0,
    10.0
};

static const double hg_excess3[] = {
    60.0,
    116.0,
    -64.0,
    -324.0,
    -252.0,
    80.0,
    264.0,
    248.0,
    132.0,
    0.0,
    -60.0,
    -80.0,
    -60.0,
    -40.0,
    -20.0
};

static const double hg_excess4[] = {
    105.0,
    276.0,
    -104.0,
    -982.0,
    -940.0,
    416.0,
    1493.0,
    1222.0,
    332.0,
    -420.0,
    -771.0,
    -672.0,
    -435.0,
    -180.0,
    5.0,
    130.0,
    175.0,
    140.0,
    105.0,
    70.0,
    35.0
};

static const double hg_excess5[] = {
    168.0,
    588.0,
    -60.0,
    -2472.0,
    -2992.0,
    1336.0,
    5976.0,
    4940.0,
    -156.0,
    -3848.0,
    -4480.0,
    -3004.0,
    -916.0,
    444.0,
    1212.0,
    1464.0,
    1360.0,
    1028.0,
    588.0,
    280.0,
    -28.0,
    -252.0,
    -336.0,
    -280.0,
    -224.0,
    -168.0,
    -112.0,
    -56.0
};

static const double hg_excess6[] = {
    252.0,
    1148.0,
    282.0,
    -5442.0,
    -8484.0,
    2960.0,
    19722.0,
    17706.0,
    -4244.0,
    -21624.0,
    -20208.0,
    -7712.0,
    4224.0,
    9702.0,
    9116.0,
    6714.0,
    4056.0,
    1654.0,
    -84.0,
    -1186.0,
    -1804.0,
    -2436.0,
    -2548.0,
    -2044.0,
    -1344.0,
    -868.0,
    -392.0,
    84.0,
    448.0,
    588.0,
    504.0,
    420.0,
    336.0,
    252.0,
    168.0,
    84.0
};

static const double hg_excess7[] = {
    360.0,
    2088.0,
    1368.0,
    -10740.0,
    -21816.0,
    3916.0,
    56348.0,
    58596.0,
    -18964.0,
    -92384.0,
    -80344.0,
    -7952.0,
    51492.0,
    64020.0,
    41188.0,
    11304.0,
    -7096.0,
    -14812.0,
    -15448.0,
    -12044.0,
    -8560.0,
    -6760.0,
    -6088.0,
    -3968.0,
    -1336.0,
    384.0,
    1416.0,
    2784.0,
    4104.0,
    4368.0,
    3640.0,
    2592.0,
    1896.0,
    1200.0,
    504.0,
    -192.0,
    -744.0,
    -960.0,
    -840.0,
    -720.0,
    -600.0,
    -480.0,
    -360.0,
    -240.0,
    -120.0
};

static const double hg_excess8[] = {
    495.0,
    3582.0,
    4023.0,
    -19224.0,
    -51480.0,
    -2106.0,
    143256.0,
    179588.0,
    -53457.0,
    -330368.0,
    -293644.0,
    32272.0,
    301719.0,
    305476.0,
    123577.0,
    -61832.0,
    -144304.0,
    -131828.0,
    -82328.0,
    -29828.0,
    7955.0,
    20886.0,
    18113.0,
    13520.0,
    13051.0,
    11340.0,
    7820.0,
    7452.0,
    10581.0,
    11812.0,
    8669.0,
    4224.0,
    1209.0,
    -206.0,
    -2133.0,
    -4572.0,
    -6525.0,
    -7008.0,
    -5997.0,
    -4500.0,
    -3525.0,
    -2550.0,
    -1575.0,
    -600.0,
    375.0,
    1170.0,
    1485.0,
    1320.0,
    1155.0,
    990.0,
    825.0,
    660.0,
    495.0,
    330.0,
    165.0
};

static const double hg_excess9[] = {
    660.0,
    5852.0,
    9652.0,
    -31272.0,
    -112564.0,
    -34088.0,
    329280.0,
    510284.0,
    -99652.0,
    -1034280.0,
    -1006236.0,
    222808.0,
    1344444.0,
    1243508.0,
    225024.0,
    -695088.0,
    -934112.0,
    -615392.0,
    -159648.0,
    157912.0,
    294600.0,
    278652.0,
    173460.0,
    72660.0,
    15804.0,
    -12924.0,
    -35080.0,
    -39756.0,
    -19564.0,
    2572.0,
    5592.0,
    -4460.0,
    -9588.0,
    -8004.0,
    -7528.0,
    -13432.0,
    -19896.0,
    -20668.0,
    -15420.0,
    -8328.0,
    -4068.0,
    -2196.0,
    420.0,
    3780.0,
    7140.0,
    9900.0,
    10680.0,
    9320.0,
    7260.0,
    5940.0,
    4620.0,
    3300.0,
    1980.0,
    660.0,
    -660.0,
    -1760.0,
    -2200.0,
    -1980.0,
    -1760.0,
    -1540.0,
    -1320.0,
    -1100.0,
    -880.0,
    -660.0,
    -440.0,
    -220.0
};

static const double hg_excess10[] = {
    858.0,
    9174.0,
    20504.0,
    -45794.0,
    -229904.0,
    -139070.0,
    689944.0,
    1347826.0,
    -56148.0,
    -2907472.0,
    -3244728.0,
    777786.0,
    5077620.0,
    4628372.0,
    -22056.0,
    -4101636.0,
    -4485122.0,
    -2008642.0,
    720548.0,
    2060478.0,
    2016886.0,
    1277238.0,
    389380.0,
    -256164.0,
    -491162.0,
    -473982.0,
    -400572.0,
    -308060.0,
    -152526.0,
    14496.0,
    86156.0,
    64282.0,
    33866.0,
    37872.0,
    41152.0,
    14206.0,
    -26762.0,
    -42714.0,
    -25796.0,
    -20.0,
    12626.0,
    10060.0,
    8188.0,
    15172.0,
    24608.0,
    31692.0,
    32176.0,
    24244.0,
    14436.0,
    8628.0,
    6220.0,
    2772.0,
    -1716.0,
    -6204.0,
    -10692.0,
    -14454.0,
    -15620.0,
    -13838.0,
    -11088.0,
    -9350.0,
    -7612.0,
    -5874.0,
    -4136.0,
    -2398.0,
    -660.0,
    1078.0,
    2552.0,
    3146.0,
    2860.0,
    2574.0,
    2288.0,
    2002.0,
    1716.0,
    1430.0,
    1144.0,
    858.0,
    572.0,
    286.0
};

static const double hg_excess11[] = {
    1092.0,
    13884.0,
    40008.0,
    -58476.0,
    -441440.0,
    -420076.0,
    1320000.0,
    3322348.0,
    495304.0,
    -7444880.0,
    -9840148.0,
    1820276.0,
    16979240.0,
    16227256.0,
    -2046380.0,
    -18840056.0,
    -18553692.0,
    -4663884.0,
    8841164.0,
    13051056.0,
    9108188.0,
    2676240.0,
    -2246288.0,
    -4435352.0,
    -4125748.0,
    -2548760.0,
    -1129636.0,
    -284680.0,
    323516.0,
    844276.0,
    1009456.0,
    734196.0,
    380660.0,
    234744.0,
    199800.0,
    64448.0,
    -146236.0,
    -258296.0,
    -192508.0,
    -54768.0,
    17344.0,
    6084.0,
    -19164.0,
    412.0,
    45352.0,
    73848.0,
    70432.0,
    37388.0,
    -2844.0,
    -18916.0,
    -12352.0,
    -6996.0,
    -14772.0,
    -27388.0,
    -38604.0,
    -46824.0,
    -46516.0,
    -36224.0,
    -23076.0,
    -15384.0,
    -12356.0,
    -7920.0,
    -2076.0,
    3768.0,
    9612.0,
    15456.0,
    20436.0,
    22088.0,
    19804.0,
    16224.0,
    13988.0,
    11752.0,
    9516.0,
    7280.0,
    5044.0,
    2808.0,
    572.0,
    -1664.0,
    -3588.0,
    -4368.0,
    -4004.0,
    -3640.0,
    -3276.0,
    -2912.0,
    -2548.0,
    -2184.0,
    -1820.0,
    -1456.0,
    -1092.0,
    -728.0,
    -364.0
};

static const double hg_excess12[] = {
    1365.0,
    20384.0,
    73190.0,
    -58902.0,
    -800727.0,
    -1083084.0,
    2291043.0,
    7675504.0,
    2849775.0,
    -17499840.0,
    -28049831.0,
    2189408.0,
    51413793.0,
    54071644.0,
    -9397926.0,
    -74198502.0,
    -70333432.0,
    -5783544.0,
    54059649.0,
    63596638.0,
    31318454.0,
    -7259700.0,
    -28079642.0,
    -28923780.0,
    -18048330.0,
    -4624552.0,
    4389234.0,
    7126998.0,
    6577811.0,
    5585868.0,
    4281057.0,
    2100616.0,
    -59372.0,
    -895908.0,
    -696756.0,
    -680972.0,
    -1206699.0,
    -1540804.0,
    -1141838.0,
    -348246.0,
    151579.0,
    153276.0,
    11811.0,
    52140.0,
    250266.0,
    375312.0,
    316215.0,
    130936.0,
    -59415.0,
    -131568.0,
    -70853.0,
    2268.0,
    1167.0,
    -49788.0,
    -91125.0,
    -104564.0,
    -89528.0,
    -43920.0,
    4613.0,
    23204.0,
    12240.0,
    2444.0,
    12327.0,
    28788.0,
    43379.0,
    56100.0,
    65067.0,
    65098.0,
    52013.0,
    34836.0,
    24891.0,
    21154.0,
    15561.0,
    8112.0,
    663.0,
    -6786.0,
    -14235.0,
    -21684.0,
    -28119.0,
    -30368.0,
    -27495.0,
    -22932.0,
    -20111.0,
    -17290.0,
    -14469.0,
    -11648.0,
    -8827.0,
    -6006.0,
    -3185.0,
    -364.0,
    2457.0,
    4914.0,
    5915.0,
    5460.0,
    5005.0,
    4550.0,
    4095.0,
    3640.0,
    3185.0,
    2730.0,
    2275.0,
    1820.0,
    1365.0,
    910.0,
    455.0
};

static const double hg_excess13[] = {
    1680.0,
    29148.0,
    127180.0,
    -26104.0,
    -1376608.0,
    -2510852.0,
    3532164.0,
    16685780.0,
    10434124.0,
    -37862124.0,
    -75215380.0,
    -4976160.0,
    142784140.0,
    171313328.0,
    -26644980.0,
    -260908136.0,
    -251537768.0,
    10199192.0,
    256616564.0,
    268603428.0,
    83907496.0,
    -108355752.0,
    -178497928.0,
    -132135368.0,
    -41802956.0,
    32485336.0,
    64792208.0,
    57601728.0,
    32816616.0,
    12037036.0,
    264508.0,
    -7619224.0,
    -13012192.0,
    -12776220.0,
    -7517612.0,
    -3029628.0,
    -2586148.0,
    -3538772.0,
    -2282272.0,
    826228.0,
    2855392.0,
    2454760.0,
    1096300.0,
    686064.0,
    1292536.0,
    1764956.0,
    1347888.0,
    321208.0,
    -603600.0,
    -914768.0,
    -581324.0,
    -103140.0,
    9404.0,
    -198128.0,
    -375060.0,
    -357340.0,
    -198696.0,
    14492.0,
    193684.0,
    222668.0,
    106380.0,
    -16104.0,
    -27272.0,
    42284.0,
    100692.0,
    127096.0,
    132440.0,
    111364.0,
    58056.0,
    -2996.0,
    -25580.0,
    -8460.0,
    5188.0,
    -7152.0,
    -28188.0,
    -46784.0,
    -62940.0,
    -76656.0,
    -88176.0,
    -88672.0,
    -72324.0,
    -50364.0,
    -37764.0,
    -33224.0,
    -26292.0,
    -16968.0,
    -7644.0,
    1680.0,
    11004.0,
    20328.0,
    29652.0,
    37800.0,
    40768.0,
    37212.0,
    31500.0,
    28000.0,
    24500.0,
    21000.0,
    17500.0,
    14000.0,
    10500.0,
    7000.0,
    3500.0,
    0.0,
    -3500.0,
    -6580.0,
    -7840.0,
    -7280.0,
    -6720.0,
    -6160.0,
    -5600.0,
    -5040.0,
    -4480.0,
    -3920.0,
    -3360.0,
    -2800.0,
    -2240.0,
    -1680.0,
    -1120.0,
    -560.0
};

static const double hg_excess14[] = {
    2040.0,
    40728.0,
    211818.0,
    78018.0,
    -2246668.0,
    -5371712.0,
    4560888.0,
    34242972.0,
    31382924.0,
    -75186284.0,
    -190075780.0,
    -45498988.0,
    366319664.0,
    515219370.0,
    -44130682.0,
    -836475788.0,
    -858820372.0,
    90923142.0,
    1047903532.0,
    1038914380.0,
    162718224.0,
    -694701854.0,
    -883450978.0,
    -477091798.0,
    59814600.0,
    379403484.0,
    411223310.0,
    253938938.0,
    54480080.0,
    -74357650.0,
    -109886452.0,
    -97981094.0,
    -77018812.0,
    -48931300.0,
    -11862216.0,
    16764014.0,
    21031180.0,
    10328424.0,
    5168470.0,
    10605190.0,
    15250480.0,
    10675672.0,
    1736130.0,
    -2250064.0,
    396636.0,
    3699136.0,
    2906284.0,
    -1187352.0,
    -4849344.0,
    -5524648.0,
    -3394116.0,
    -677356.0,
    318144.0,
    -519704.0,
    -1403380.0,
    -1159038.0,
    -126980.0,
    907174.0,
    1464460.0,
    1331508.0,
    602796.0,
    -131542.0,
    -319388.0,
    -16626.0,
    296912.0,
    350984.0,
    250416.0,
    99304.0,
    -94412.0,
    -267408.0,
    -282496.0,
    -111088.0,
    58596.0,
    67104.0,
    -31732.0,
    -113528.0,
    -150732.0,
    -164856.0,
    -163500.0,
    -140944.0,
    -77472.0,
    -2470.0,
    24008.0,
    1674.0,
    -16692.0,
    -1518.0,
    24888.0,
    48174.0,
    68340.0,
    85386.0,
    102432.0,
    116946.0,
    118048.0,
    97930.0,
    70368.0,
    54678.0,
    49236.0,
    40770.0,
    29280.0,
    17790.0,
    6300.0,
    -5190.0,
    -16680.0,
    -28170.0,
    -39660.0,
    -49800.0,
    -53620.0,
    -49280.0,
    -42240.0,
    -37960.0,
    -33680.0,
    -29400.0,
    -25120.0,
    -20840.0,
    -16560.0,
    -12280.0,
    -8000.0,
    -3720.0,
    560.0,
    4840.0,
    8640.0,
    10200.0,
    9520.0,
    8840.0,
    8160.0,
    7480.0,
    6800.0,
    6120.0,
    5440.0,
    4760.0,
    4080.0,
    3400.0,
    2720.0,
    2040.0,
    1360.0,
    680.0
};

static const double hg_excess15[] = {
    2448.0,
    55760.0,
    340368.0,
    318084.0,
    -3479136.0,
    -10769628.0,
    3947220.0,
    66487232.0,
    83444748.0,
    -135499668.0,
    -453698940.0,
    -199061480.0,
    870994872.0,
    1469256648.0,
    33059712.0,
    -2474300736.0,
    -2807261032.0,
    320821320.0,
    3838627116.0,
    3787566264.0,
    137036824.0,
    -3413724420.0,
    -3804887760.0,
    -1409036952.0,
    1276536468.0,
    2409616724.0,
    1899790912.0,
    649990764.0,
    -444382204.0,
    -921437068.0,
    -808257024.0,
    -447291904.0,
    -144268420.0,
    38415336.0,
    160893512.0,
    225135540.0,
    190887804.0,
    88013164.0,
    10188272.0,
    1027812.0,
    16612132.0,
    4385224.0,
    -29254332.0,
    -43273640.0,
    -23676328.0,
    3370944.0,
    9903608.0,
    -3497792.0,
    -17826408.0,
    -19152752.0,
    -8380404.0,
    3889284.0,
    7884412.0,
    2891568.0,
    -2998740.0,
    -2881708.0,
    2041612.0,
    6609876.0,
    7874296.0,
    5804848.0,
    1871724.0,
    -1598364.0,
    -2486856.0,
    -978888.0,
    701796.0,
    922996.0,
    48468.0,
    -824800.0,
    -1362356.0,
    -1587240.0,
    -1259708.0,
    -310796.0,
    622932.0,
    773324.0,
    234956.0,
    -279132.0,
    -417604.0,
    -335452.0,
    -211440.0,
    -68528.0,
    123020.0,
    317484.0,
    323484.0,
    104212.0,
    -102444.0,
    -108876.0,
    20132.0,
    127956.0,
    179028.0,
    196612.0,
    199428.0,
    201656.0,
    178624.0,
    104556.0,
    14320.0,
    -19628.0,
    8904.0,
    32940.0,
    14528.0,
    -18108.0,
    -46824.0,
    -71620.0,
    -92496.0,
    -113372.0,
    -134248.0,
    -152232.0,
    -154096.0,
    -129664.0,
    -95616.0,
    -76368.0,
    -69920.0,
    -59712.0,
    -45744.0,
    -31776.0,
    -17808.0,
    -3840.0,
    10128.0,
    24096.0,
    38064.0,
    52032.0,
    64464.0,
    69280.0,
    64048.0,
    55488.0,
    50320.0,
    45152.0,
    39984.0,
    34816.0,
    29648.0,
    24480.0,
    19312.0,
    14144.0,
    8976.0,
    3808.0,
    -1360.0,
    -6528.0,
    -11152.0,
    -13056.0,
    -12240.0,
    -11424.0,
    -10608.0,
    -9792.0,
    -8976.0,
    -8160.0,
    -7344.0,
    -6528.0,
    -5712.0,
    -4896.0,
    -4080.0,
    -3264.0,
    -2448.0,
    -1632.0,
    -816.0
};

static const double hg_excess16[] = {
    2907.0,
    74970.0,
    530349.0,
    798240.0,
    -5095251.0,
    -20436240.0,
    -1666479.0,
    122241052.0,
    202715885.0,
    -215292570.0,
    -1025075670.0,
    -682491252.0,
    1917913530.0,
    3972858858.0,
    634479819.0,
    -6799276436.0,
    -8777007901.0,
    649812476.0,
    12883130619.0,
    13199621464.0,
    -517278562.0,
    -14415800420.0,
    -15007532678.0,
    -3359138048.0,
    8706580934.0,
    12087431164.0,
    7134017752.0,
    -238449700.0,
    -5033245737.0,
    -5745711874.0,
    -3562655936.0,
    -718519392.0,
    1092119032.0,
    1592976624.0,
    1432418900.0,
    1098614500.0,
    617605130.0,
    30581152.0,
    -370667152.0,
    -389858848.0,
    -203680632.0,
    -110477232.0,
    -163004375.0,
    -189850860.0,
    -84400555.0,
    63824732.0,
    116035630.0,
    55918144.0,
    -23032417.0,
    -42727146.0,
    -4258414.0,
    41186496.0,
    49198653.0,
    18035362.0,
    -16634776.0,
    -21051904.0,
    2188474.0,
    25016268.0,
    28722933.0,
    15631512.0,
    -2659806.0,
    -15334210.0,
    -16226668.0,
    -7294932.0,
    1846605.0,
    3286584.0,
    -1667184.0,
    -6454596.0,
    -7584176.0,
    -6097774.0,
    -3104473.0,
    1115024.0,
    4746509.0,
    5112252.0,
    2217911.0,
    -753184.0,
    -1453109.0,
    -547648.0,
    395627.0,
    884092.0,
    1187883.0,
    1369746.0,
    1037617.0,
    -29260.0,
    -1057549.0,
    -1110066.0,
    -358183.0,
    348020.0,
    566805.0,
    478290.0,
    317453.0,
    183288.0,
    59825.0,
    -146068.0,
    -363039.0,
    -367976.0,
    -105315.0,
    152796.0,
    159481.0,
    -6092.0,
    -145683.0,
    -209354.0,
    -230913.0,
    -233880.0,
    -241775.0,
    -249438.0,
    -227318.0,
    -142496.0,
    -31009.0,
    11660.0,
    -24140.0,
    -54884.0,
    -32803.0,
    6988.0,
    41929.0,
    72020.0,
    97261.0,
    122502.0,
    147743.0,
    172984.0,
    194949.0,
    197746.0,
    168419.0,
    126936.0,
    103629.0,
    96066.0,
    83895.0,
    67116.0,
    50337.0,
    33558.0,
    16779.0,
    0.0,
    -16779.0,
    -33558.0,
    -50337.0,
    -67116.0,
    -82161.0,
    -88128.0,
    -81889.0,
    -71604.0,
    -65433.0,
    -59262.0,
    -53091.0,
    -46920.0,
    -40749.0,
    -34578.0,
    -28407.0,
    -22236.0,
    -16065.0,
    -9894.0,
    -3723.0,
    2448.0,
    8619.0,
    14178.0,
    16473.0,
    15504.0,
    14535.0,
    13566.0,
    12597.0,
    11628.0,
    10659.0,
    9690.0,
    8721.0,
    7752.0,
    6783.0,
    5814.0,
    4845.0,
    3876.0,
    2907.0,
    1938.0,
    969.0
};

static const double hg_excess17[] = {
    3420.0,
    99180.0,
    804492.0,
    1679736.0,
    -7000608.0,
    -36953964.0,
    -19242684.0,
    212610060.0,
    458211704.0,
    -278099312.0,
    -2195797532.0,
    -2033281372.0,
    3887622372.0,
    10194118944.0,
    3273358260.0,
    -17417076540.0,
    -26205426864.0,
    -118899596.0,
    40084611576.0,
    44203966044.0,
    -2943202252.0,
    -54713491024.0,
    -55739190264.0,
    -5923221272.0,
    44446081548.0,
    53000646620.0,
    22621594988.0,
    -14389603572.0,
    -32131432568.0,
    -26939807492.0,
    -9922621972.0,
    5781249568.0,
    12701025248.0,
    11136243492.0,
    6112235792.0,
    1837530604.0,
    -872675408.0,
    -2670253276.0,
    -3408792628.0,
    -2635302744.0,
    -1016697056.0,
    66771416.0,
    127073976.0,
    -101358104.0,
    115772340.0,
    613828732.0,
    775356820.0,
    422481964.0,
    -46241936.0,
    -218485324.0,
    -77646092.0,
    116683492.0,
    145799500.0,
    2118052.0,
    -153049628.0,
    -166943460.0,
    -43180280.0,
    84171084.0,
    109219532.0,
    43650212.0,
    -37982808.0,
    -79244988.0,
    -65573968.0,
    -17316600.0,
    24701624.0,
    28738836.0,
    714340.0,
    -26127080.0,
    -29799580.0,
    -15635552.0,
    2039932.0,
    17253704.0,
    26654964.0,
    23954192.0,
    9303052.0,
    -5291012.0,
    -8343752.0,
    -1871532.0,
    4380996.0,
    5775136.0,
    4305272.0,
    2487780.0,
    130196.0,
    -3568504.0,
    -6837184.0,
    -6358652.0,
    -2213260.0,
    1705468.0,
    2738760.0,
    1630040.0,
    293004.0,
    -455352.0,
    -761188.0,
    -1036232.0,
    -1326156.0,
    -1013956.0,
    181396.0,
    1365528.0,
    1408104.0,
    430156.0,
    -491136.0,
    -770520.0,
    -638316.0,
    -417240.0,
    -246912.0,
    -158616.0,
    -57704.0,
    150656.0,
    401620.0,
    420200.0,
    104524.0,
    -212180.0,
    -218032.0,
    -8648.0,
    163176.0,
    241400.0,
    267484.0,
    270528.0,
    279632.0,
    294796.0,
    310200.0,
    290812.0,
    189340.0,
    53492.0,
    756.0,
    44980.0,
    83556.0,
    57348.0,
    9412.0,
    -32604.0,
    -68700.0,
    -98876.0,
    -129052.0,
    -159228.0,
    -189404.0,
    -219580.0,
    -246072.0,
    -249988.0,
    -215148.0,
    -165216.0,
    -137316.0,
    -128524.0,
    -114156.0,
    -94212.0,
    -74268.0,
    -54324.0,
    -34380.0,
    -14436.0,
    5508.0,
    25452.0,
    45396.0,
    65340.0,
    85284.0,
    103284.0,
    110568.0,
    103200.0,
    90972.0,
    83676.0,
    76380.0,
    69084.0,
    61788.0,
    54492.0,
    47196.0,
    39900.0,
    32604.0,
    25308.0,
    18012.0,
    10716.0,
    3420.0,
    -3876.0,
    -11172.0,
    -17784.0,
    -20520.0,
    -19380.0,
    -18240.0,
    -17100.0,
    -15960.0,
    -14820.0,
    -13680.0,
    -12540.0,
    -11400.0,
    -10260.0,
    -9120.0,
    -7980.0,
    -6840.0,
    -5700.0,
    -4560.0,
    -3420.0,
    -2280.0,
    -1140.0
};

static const double hg_excess18[] = {
    3990.0,
    129314.0,
    1191832.0,
    3204186.0,
    -8869486.0,
    -63978328.0,
    -62062836.0,
    348519960.0,
    974619016.0,
    -201489990.0,
    -4462746212.0,
    -5486872948.0,
    7142822286.0,
    24849868202.0,
    12564947620.0,
    -41606709690.0,
    -74636378630.0,
    -8411601956.0,
    116411822628.0,
    142338422142.0,
    -7075042724.0,
    -190827200316.0,
    -197889792352.0,
    -4965262476.0,
    193978909440.0,
    212957633492.0,
    61053499260.0,
    -106483094052.0,
    -163404033114.0,
    -104762233908.0,
    -4638212460.0,
    65710380270.0,
    79054668174.0,
    49625028024.0,
    10041971748.0,
    -15199900422.0,
    -22458746262.0,
    -20395216916.0,
    -15160758172.0,
    -7401076302.0,
    1293148404.0,
    6405828128.0,
    5788736388.0,
    2645274344.0,
    1279286240.0,
    2112132450.0,
    2495390232.0,
    953929336.0,
    -1169849280.0,
    -1923038696.0,
    -1079179384.0,
    78041904.0,
    413814422.0,
    -113095042.0,
    -739739988.0,
    -755261974.0,
    -148971204.0,
    481487820.0,
    607283708.0,
    261084826.0,
    -147424032.0,
    -307481728.0,
    -195915328.0,
    32049360.0,
    190382750.0,
    171481616.0,
    16928004.0,
    -122322842.0,
    -135941582.0,
    -48705774.0,
    46352664.0,
    97329998.0,
    104807790.0,
    72504368.0,
    8028240.0,
    -49231986.0,
    -55291158.0,
    -15277132.0,
    23003112.0,
    28425416.0,
    11229226.0,
    -6260154.0,
    -16911252.0,
    -25096428.0,
    -30013278.0,
    -23370222.0,
    -4243676.0,
    13338486.0,
    16354662.0,
    7824710.0,
    -1024344.0,
    -4199386.0,
    -3249862.0,
    -1694286.0,
    -935500.0,
    424558.0,
    3930114.0,
    7394698.0,
    6705636.0,
    1648956.0,
    -3261622.0,
    -4452554.0,
    -2765352.0,
    -762608.0,
    367582.0,
    736908.0,
    814968.0,
    1075340.0,
    1422126.0,
    1118104.0,
    -248388.0,
    -1684620.0,
    -1725848.0,
    -493248.0,
    652692.0,
    978538.0,
    803612.0,
    524994.0,
    308104.0,
    201962.0,
    153168.0,
    91238.0,
    -127952.0,
    -441222.0,
    -477848.0,
    -103350.0,
    279276.0,
    282778.0,
    28848.0,
    -179922.0,
    -274796.0,
    -305990.0,
    -309024.0,
    -319418.0,
    -337172.0,
    -362286.0,
    -388240.0,
    -366524.0,
    -246336.0,
    -82804.0,
    -18560.0,
    -72450.0,
    -120068.0,
    -89248.0,
    -32112.0,
    17884.0,
    60740.0,
    96456.0,
    132172.0,
    167888.0,
    203604.0,
    239320.0,
    275036.0,
    306636.0,
    311872.0,
    270864.0,
    211404.0,
    178344.0,
    168204.0,
    151392.0,
    127908.0,
    104424.0,
    80940.0,
    57456.0,
    33972.0,
    10488.0,
    -12996.0,
    -36480.0,
    -59964.0,
    -83448.0,
    -106932.0,
    -128250.0,
    -137028.0,
    -128402.0,
    -114000.0,
    -105450.0,
    -96900.0,
    -88350.0,
    -79800.0,
    -71250.0,
    -62700.0,
    -54150.0,
    -45600.0,
    -37050.0,
    -28500.0,
    -19950.0,
    -11400.0,
    -2850.0,
    5700.0,
    14250.0,
    22040.0,
    25270.0,
    23940.0,
    22610.0,
    21280.0,
    19950.0,
    18620.0,
    17290.0,
    15960.0,
    14630.0,
    13300.0,
    11970.0,
    10640.0,
    9310.0,
    7980.0,
    6650.0,
    5320.0,
    3990.0,
    2660.0,
    1330.0
};

static const double hg_excess19[] = {
    4620.0,
    166404.0,
    1728944.0,
    5723788.0,
    -9960452.0,
    -106393416.0,
    -153905752.0,
    533974472.0,
    1965304248.0,
    331754428.0,
    -8601201692.0,
    -13704507536.0,
    11459053276.0,
    57607258132.0,
    41221895736.0,
    -92390425016.0,
    -202715282512.0,
    -49604460788.0,
    316701668688.0,
    440180713940.0,
    1898990360.0,
    -619165896264.0,
    -676303587132.0,
    10639801216.0,
    760504434220.0,
    805132363376.0,
    138222171244.0,
    -567236631600.0,
    -728606067864.0,
    -351285441600.0,
    153566688624.0,
    423844728296.0,
    377173185520.0,
    148995174204.0,
    -72012653564.0,
    -171716160376.0,
    -152415974552.0,
    -84305759928.0,
    -24239041408.0,
    16114790612.0,
    42261707568.0,
    49912774388.0,
    34824229044.0,
    9833485028.0,
    -4909970964.0,
    -4236770948.0,
    86054996.0,
    -2314903820.0,
    -8686559980.0,
    -10384066648.0,
    -5155929396.0,
    1430660060.0,
    3688787876.0,
    1339332152.0,
    -1804813104.0,
    -2276879164.0,
    113557636.0,
    2741957368.0,
    3123997044.0,
    1276492764.0,
    -814336836.0,
    -1513818652.0,
    -811396640.0,
    293301096.0,
    880287460.0,
    633328236.0,
    -125913468.0,
    -729998444.0,
    -716844236.0,
    -213293332.0,
    287130076.0,
    466798860.0,
    367133060.0,
    144495928.0,
    -107300460.0,
    -282415012.0,
    -255373932.0,
    -46619540.0,
    146456132.0,
    164122540.0,
    46039192.0,
    -68139152.0,
    -110488004.0,
    -106223320.0,
    -86897368.0,
    -44548364.0,
    26563484.0,
    86180260.0,
    85097836.0,
    31879836.0,
    -17961428.0,
    -29625732.0,
    -12848340.0,
    6270888.0,
    14549320.0,
    16278020.0,
    20346356.0,
    25082940.0,
    18344456.0,
    -2874852.0,
    -22962480.0,
    -25664024.0,
    -13172076.0,
    94180.0,
    5534532.0,
    4788012.0,
    2476360.0,
    1382176.0,
    1549172.0,
    551816.0,
    -3391184.0,
    -7720852.0,
    -7251136.0,
    -1234348.0,
    4686648.0,
    5913460.0,
    3587612.0,
    1007856.0,
    -456800.0,
    -967876.0,
    -974412.0,
    -896668.0,
    -1062856.0,
    -1483672.0,
    -1234352.0,
    355900.0,
    2054456.0,
    2070116.0,
    561784.0,
    -818092.0,
    -1209280.0,
    -990388.0,
    -643768.0,
    -371404.0,
    -233200.0,
    -184356.0,
    -204492.0,
    -169024.0,
    97432.0,
    483516.0,
    543320.0,
    104196.0,
    -351816.0,
    -359924.0,
    -55408.0,
    195348.0,
    309112.0,
    346036.0,
    348960.0,
    360724.0,
    381328.0,
    410772.0,
    449056.0,
    480080.0,
    455904.0,
    314812.0,
    120060.0,
    42772.0,
    107656.0,
    165612.0,
    129668.0,
    62212.0,
    3276.0,
    -47140.0,
    -89036.0,
    -130932.0,
    -172828.0,
    -214724.0,
    -256620.0,
    -298516.0,
    -340412.0,
    -377736.0,
    -384508.0,
    -336640.0,
    -266508.0,
    -227688.0,
    -216076.0,
    -196560.0,
    -169140.0,
    -141720.0,
    -114300.0,
    -86880.0,
    -59460.0,
    -32040.0,
    -4620.0,
    22800.0,
    50220.0,
    77640.0,
    105060.0,
    132480.0,
    157500.0,
    167960.0,
    157940.0,
    141120.0,
    131180.0,
    121240.0,
    111300.0,
    101360.0,
    91420.0,
    81480.0,
    71540.0,
    61600.0,
    51660.0,
    41720.0,
    31780.0,
    21840.0,
    11900.0,
    1960.0,
    -7980.0,
    -17920.0,
    -27020.0,
    -30800.0,
    -29260.0,
    -27720.0,
    -26180.0,
    -24640.0,
    -23100.0,
    -21560.0,
    -20020.0,
    -18480.0,
    -16940.0,
    -15400.0,
    -13860.0,
    -12320.0,
    -10780.0,
    -9240.0,
    -7700.0,
    -6160.0,
    -4620.0,
    -3080.0,
    -1540.0
};

static const double hg_excess20[] = {
    5313.0,
    211596.0,
    2461332.0,
    9739962.0,
    -8834438.0,
    -170281168.0,
    -335184435.0,
    751319522.0,
    3776018250.0,
    2041116712.0,
    -15682285832.0,
    -32094042264.0,
    14378349287.0,
    127077645160.0,
    121563088477.0,
    -189018557996.0,
    -525164617834.0,
    -208941042684.0,
    807981191523.0,
    1305529805220.0,
    115012392650.0,
    -1882196498768.0,
    -2229767175893.0,
    38110844722.0,
    2743565730050.0,
    2908342075432.0,
    252970810155.0,
    -2559678703598.0,
    -2978770759665.0,
    -1032554851560.0,
    1269819446943.0,
    2186478441884.0,
    1513541115832.0,
    168230374908.0,
    -843999043098.0,
    -1073938335866.0,
    -688824771201.0,
    -148483864076.0,
    202515701916.0,
    312083089240.0,
    289410269821.0,
    209868267864.0,
    88123251799.0,
    -39197050188.0,
    -104105722726.0,
    -84572665932.0,
    -33780836587.0,
    -12550678816.0,
    -23377031065.0,
    -27878985252.0,
    -6530240217.0,
    21446140124.0,
    29495079175.0,
    14949026656.0,
    -3746883295.0,
    -9431227176.0,
    -881033187.0,
    9991885656.0,
    11537152966.0,
    3131790580.0,
    -6321977162.0,
    -8821560652.0,
    -4362309673.0,
    1474470706.0,
    3947597131.0,
    2274554988.0,
    -1273179119.0,
    -3632422752.0,
    -3100441840.0,
    -465036556.0,
    1927828287.0,
    2473677892.0,
    1473441054.0,
    123052244.0,
    -839772557.0,
    -1204437684.0,
    -841899977.0,
    85838980.0,
    868780521.0,
    852275922.0,
    176415627.0,
    -453205736.0,
    -601189819.0,
    -401493492.0,
    -153257827.0,
    51360620.0,
    253064657.0,
    387725012.0,
    313414093.0,
    50654192.0,
    -178281603.0,
    -201876720.0,
    -69006105.0,
    64849300.0,
    111148183.0,
    89508244.0,
    59515610.0,
    42835952.0,
    10301601.0,
    -57981822.0,
    -119481899.0,
    -113274344.0,
    -42717955.0,
    26984220.0,
    47093711.0,
    26786800.0,
    785160.0,
    -10781954.0,
    -9028964.0,
    -5849876.0,
    -11124617.0,
    -20046434.0,
    -16010791.0,
    6728444.0,
    29881754.0,
    31888156.0,
    15090851.0,
    -1984584.0,
    -8809141.0,
    -7976114.0,
    -5010120.0,
    -2705920.0,
    -1883371.0,
    -2108362.0,
    -1123122.0,
    3520304.0,
    8925233.0,
    8402994.0,
    1143313.0,
    -5892188.0,
    -7241131.0,
    -4374266.0,
    -1190437.0,
    640760.0,
    1303749.0,
    1298476.0,
    995983.0,
    684984.0,
    892129.0,
    1541554.0,
    1358155.0,
    -444620.0,
    -2427303.0,
    -2441948.0,
    -647879.0,
    1004472.0,
    1467979.0,
    1196952.0,
    770685.0,
    432856.0,
    266535.0,
    206952.0,
    242739.0,
    309336.0,
    260835.0,
    -60588.0,
    -531093.0,
    -619992.0,
    -110497.0,
    437154.0,
    450821.0,
    89288.0,
    -208821.0,
    -343858.0,
    -387167.0,
    -389868.0,
    -403081.0,
    -426806.0,
    -461043.0,
    -505792.0,
    -550541.0,
    -587250.0,
    -560482.0,
    -396176.0,
    -166455.0,
    -74492.0,
    -151784.0,
    -221460.0,
    -179853.0,
    -100892.0,
    -32001.0,
    26820.0,
    75571.0,
    124322.0,
    173073.0,
    221824.0,
    270575.0,
    319326.0,
    368077.0,
    416828.0,
    460527.0,
    469066.0,
    413609.0,
    331596.0,
    286383.0,
    273170.0,
    250677.0,
    218904.0,
    187131.0,
    155358.0,
    123585.0,
    91812.0,
    60039.0,
    28266.0,
    -3507.0,
    -35280.0,
    -67053.0,
    -98826.0,
    -130599.0,
    -162372.0,
    -191499.0,
    -203840.0,
    -192283.0,
    -172788.0,
    -161315.0,
    -149842.0,
    -138369.0,
    -126896.0,
    -115423.0,
    -103950.0,
    -92477.0,
    -81004.0,
    -69531.0,
    -58058.0,
    -46585.0,
    -35112.0,
    -23639.0,
    -12166.0,
    -693.0,
    10780.0,
    22253.0,
    32802.0,
    37191.0,
    35420.0,
    33649.0,
    31878.0,
    30107.0,
    28336.0,
    26565.0,
    24794.0,
    23023.0,
    21252.0,
    19481.0,
    17710.0,
    15939.0,
    14168.0,
    12397.0,
    10626.0,
    8855.0,
    7084.0,
    5313.0,
    3542.0,
    1771.0
};

static const double hg_excess21[] = {
    6072.0,
    266156.0,
    3444980.0,
    15952056.0,
    -2937792.0,
    -262506512.0,
    -671545104.0,
    932622844.0,
    6935465580.0,
    6420241560.0,
    -26900101328.0,
    -71062168092.0,
    7159234056.0,
    266695630952.0,
    330394545940.0,
    -349191345240.0,
    -1298271895068.0,
    -743030525732.0,
    1929923868180.0,
    3709999978752.0,
    726611265072.0,
    -5381545965492.0,
    -7090129549788.0,
    -117668277860.0,
    9228321748056.0,
    10119575076644.0,
    369458029380.0,
    -10349761087428.0,
    -11452964541800.0,
    -2677099543668.0,
    7101715858452.0,
    9907180263996.0,
    5312886163992.0,
    -1511049530508.0,
    -5531021388696.0,
    -5218156318612.0,
    -2206803793116.0,
    852419805684.0,
    2281409325248.0,
    2074427920632.0,
    1175114600900.0,
    338214467208.0,
    -260488304484.0,
    -641368159912.0,
    -716713789036.0,
    -458067349152.0,
    -86668730540.0,
    111402275368.0,
    86352763188.0,
    16734083716.0,
    44162635968.0,
    124565257596.0,
    138613161032.0,
    57765418108.0,
    -39486083280.0,
    -71578495844.0,
    -32248077220.0,
    22050081168.0,
    36109248580.0,
    4433501768.0,
    -33974043804.0,
    -41307341860.0,
    -16405671508.0,
    13063673640.0,
    23125944492.0,
    12162416640.0,
    -5583803148.0,
    -15282444176.0,
    -11283213284.0,
    1165022268.0,
    11310590300.0,
    12124005396.0,
    5447797956.0,
    -1899491964.0,
    -5447415060.0,
    -5117854296.0,
    -2344185028.0,
    1452568704.0,
    4063400100.0,
    3407689804.0,
    76037568.0,
    -2863673736.0,
    -3187815168.0,
    -1514960992.0,
    224595792.0,
    1093553092.0,
    1371319160.0,
    1336958820.0,
    779671424.0,
    -273907036.0,
    -1102513428.0,
    -1024892180.0,
    -233902536.0,
    491556720.0,
    664946320.0,
    403982020.0,
    84589152.0,
    -85486608.0,
    -177049204.0,
    -315882492.0,
    -436104856.0,
    -345853808.0,
    -31929780.0,
    259100164.0,
    302018688.0,
    132232236.0,
    -50857204.0,
    -117365860.0,
    -81144768.0,
    -21874372.0,
    906528.0,
    -9659004.0,
    909048.0,
    65242592.0,
    133664448.0,
    125140532.0,
    36470824.0,
    -50382168.0,
    -73398284.0,
    -46404912.0,
    -13320996.0,
    3656548.0,
    5902048.0,
    1468968.0,
    -212328.0,
    8668356.0,
    22447896.0,
    20013412.0,
    -6704024.0,
    -34150932.0,
    -36000044.0,
    -15656004.0,
    4741848.0,
    12849352.0,
    11672312.0,
    7699284.0,
    4078064.0,
    1598224.0,
    859740.0,
    1811428.0,
    1195648.0,
    -4138992.0,
    -10440232.0,
    -9808452.0,
    -1221024.0,
    7144560.0,
    8769564.0,
    5298252.0,
    1413140.0,
    -860052.0,
    -1681428.0,
    -1640476.0,
    -1239616.0,
    -663864.0,
    -336268.0,
    -749108.0,
    -1625520.0,
    -1540656.0,
    478728.0,
    2802036.0,
    2860820.0,
    747648.0,
    -1209960.0,
    -1752748.0,
    -1420404.0,
    -901596.0,
    -499952.0,
    -301836.0,
    -230028.0,
    -271680.0,
    -363720.0,
    -430356.0,
    -365064.0,
    19916.0,
    587604.0,
    712356.0,
    114176.0,
    -536628.0,
    -556880.0,
    -131508.0,
    219648.0,
    378484.0,
    428868.0,
    431220.0,
    445960.0,
    473088.0,
    512604.0,
    564508.0,
    616412.0,
    668316.0,
    711360.0,
    681868.0,
    491916.0,
    223264.0,
    114900.0,
    206100.0,
    288964.0,
    241128.0,
    149412.0,
    69496.0,
    1380.0,
    -54936.0,
    -111252.0,
    -167568.0,
    -223884.0,
    -280200.0,
    -336516.0,
    -392832.0,
    -449148.0,
    -505464.0,
    -556224.0,
    -566776.0,
    -502964.0,
    -407796.0,
    -355524.0,
    -340576.0,
    -314820.0,
    -278256.0,
    -241692.0,
    -205128.0,
    -168564.0,
    -132000.0,
    -95436.0,
    -58872.0,
    -22308.0,
    14256.0,
    50820.0,
    87384.0,
    123948.0,
    160512.0,
    197076.0,
    230736.0,
    245168.0,
    231924.0,
    209484.0,
    196328.0,
    183172.0,
    170016.0,
    156860.0,
    143704.0,
    130548.0,
    117392.0,
    104236.0,
    91080.0,
    77924.0,
    64768.0,
    51612.0,
    38456.0,
    25300.0,
    12144.0,
    -1012.0,
    -14168.0,
    -27324.0,
    -39468.0,
    -44528.0,
    -42504.0,
    -40480.0,
    -38456.0,
    -36432.0,
    -34408.0,
    -32384.0,
    -30360.0,
    -28336.0,
    -26312.0,
    -24288.0,
    -22264.0,
    -20240.0,
    -18216.0,
    -16192.0,
    -14168.0,
    -12144.0,
    -10120.0,
    -8096.0,
    -6072.0,
    -4048.0,
    -2024.0
};

static const double hg_excess22[] = {
    6900.0,
    331476.0,
    4748074.0,
    25317974.0,
    11997716.0,
    -389607384.0,
    -1265327042.0,
    908270526.0,
    12200000736.0,
    16367864004.0,
    -42927278252.0,
    -149603199516.0,
    -34218472204.0,
    531732173698.0,
    839714233428.0,
    -554816594794.0,
    -3063628329956.0,
    -2361983688280.0,
    4292648516106.0,
    10097089889108.0,
    3271924517844.0,
    -14494848577928.0,
    -21718309934244.0,
    -1728874432304.0,
    29162724448376.0,
    34035116321554.0,
    712020388802.0,
    -38504738428172.0,
    -42049170593120.0,
    -6173795806510.0,
    33178613261024.0,
    41172508716478.0,
    16645285046872.0,
    -14720586205384.0,
    -28980925545348.0,
    -21545705383920.0,
    -3663609377088.0,
    10655549779936.0,
    14419605323452.0,
    9523848239556.0,
    2268628458372.0,
    -2594657801092.0,
    -4276731065020.0,
    -4078978864394.0,
    -2903931325388.0,
    -1052753829600.0,
    773524388898.0,
    1584709288110.0,
    1186535623012.0,
    415565372468.0,
    111586000864.0,
    263086828628.0,
    321553081256.0,
    25318173848.0,
    -355602830518.0,
    -452650366536.0,
    -223302005808.0,
    71646566892.0,
    173564849000.0,
    53600766574.0,
    -112445266200.0,
    -146577635618.0,
    -34467056792.0,
    96802465974.0,
    130305297112.0,
    61587360114.0,
    -30508595144.0,
    -72426943576.0,
    -47284155420.0,
    10896632924.0,
    52321050564.0,
    49228059598.0,
    13228106440.0,
    -21642325162.0,
    -32519178598.0,
    -21922099104.0,
    -4001654132.0,
    11051593476.0,
    17483885194.0,
    11594272864.0,
    -3189435596.0,
    -15045737574.0,
    -14560328998.0,
    -4416290416.0,
    5229653592.0,
    8333661026.0,
    6563854230.0,
    3637913962.0,
    570424672.0,
    -2832021552.0,
    -5000602244.0,
    -3831758486.0,
    -8672480.0,
    3192882362.0,
    3478791488.0,
    1522475730.0,
    -535380384.0,
    -1401448206.0,
    -1328894402.0,
    -1158674720.0,
    -1089970668.0,
    -605119802.0,
    488905226.0,
    1443472510.0,
    1379294348.0,
    395277268.0,
    -576769842.0,
    -837435582.0,
    -475721020.0,
    1989372.0,
    233852334.0,
    210284360.0,
    158324916.0,
    258613062.0,
    408552512.0,
    332023630.0,
    -28548856.0,
    -377387592.0,
    -418834276.0,
    -196910736.0,
    33122980.0,
    118612240.0,
    86086616.0,
    20339898.0,
    -21765768.0,
    -16160948.0,
    19712168.0,
    21803436.0,
    -51367896.0,
    -136268260.0,
    -127992020.0,
    -23886392.0,
    76426604.0,
    100661122.0,
    65777304.0,
    23044546.0,
    -1534052.0,
    -9486554.0,
    -7960332.0,
    -1700056.0,
    611088.0,
    -11231776.0,
    -29125260.0,
    -26524162.0,
    5311000.0,
    38359744.0,
    41300916.0,
    17610792.0,
    -6621704.0,
    -16432836.0,
    -14982896.0,
    -9857284.0,
    -5079528.0,
    -1627932.0,
    453668.0,
    268440.0,
    -1561364.0,
    -1167670.0,
    4647048.0,
    11819710.0,
    11205960.0,
    1231482.0,
    -8659452.0,
    -10536846.0,
    -6313080.0,
    -1598382.0,
    1131588.0,
    2076946.0,
    2017044.0,
    1532098.0,
    817492.0,
    196920.0,
    17212.0,
    608106.0,
    1777932.0,
    1845532.0,
    -449752.0,
    -3219210.0,
    -3332860.0,
    -865714.0,
    1431444.0,
    2060396.0,
    1656476.0,
    1045992.0,
    572768.0,
    338948.0,
    253332.0,
    301476.0,
    409052.0,
    501732.0,
    565536.0,
    479016.0,
    20944.0,
    -657900.0,
    -811264.0,
    -114516.0,
    651636.0,
    679572.0,
    183148.0,
    -227076.0,
    -412380.0,
    -470564.0,
    -472428.0,
    -488772.0,
    -519596.0,
    -564900.0,
    -624684.0,
    -684468.0,
    -744252.0,
    -804036.0,
    -854100.0,
    -821752.0,
    -603600.0,
    -291842.0,
    -165256.0,
    -271950.0,
    -369556.0,
    -314898.0,
    -209112.0,
    -117046.0,
    -38700.0,
    25926.0,
    90552.0,
    155178.0,
    219804.0,
    284430.0,
    349056.0,
    413682.0,
    478308.0,
    542934.0,
    607560.0,
    666102.0,
    678928.0,
    605958.0,
    496296.0,
    436266.0,
    419444.0,
    390126.0,
    348312.0,
    306498.0,
    264684.0,
    222870.0,
    181056.0,
    139242.0,
    97428.0,
    55614.0,
    13800.0,
    -28014.0,
    -69828.0,
    -111642.0,
    -153456.0,
    -195270.0,
    -237084.0,
    -275724.0,
    -292468.0,
    -277380.0,
    -251712.0,
    -236716.0,
    -221720.0,
    -206724.0,
    -191728.0,
    -176732.0,
    -161736.0,
    -146740.0,
    -131744.0,
    -116748.0,
    -101752.0,
    -86756.0,
    -71760.0,
    -56764.0,
    -41768.0,
    -26772.0,
    -11776.0,
    3220.0,
    18216.0,
    33212.0,
    47104.0,
    52900.0,
    50600.0,
    48300.0,
    46000.0,
    43700.0,
    41400.0,
    39100.0,
    36800.0,
    34500.0,
    32200.0,
    29900.0,
    27600.0,
    25300.0,
    23000.0,
    20700.0,
    18400.0,
    16100.0,
    13800.0,
    11500.0,
    9200.0,
    6900.0,
    4600.0,
    2300.0
};

static const double* hg_excess[HG_MAXDEGREE+1] = {
    hg_excess0,
    hg_excess1,
    hg_excess2,
    hg_excess3,
    hg_excess4,
    hg_excess5,
    hg_excess6,
    hg_excess7,
    hg_excess8,
    hg_excess9,
    hg_excess10,
    hg_excess11,
    hg_excess12,
    hg_excess13,
    hg_excess14,
    hg_excess15,
    hg_excess16,
    hg_excess17,
    hg_excess18,
    hg_excess19,
    hg_excess20,
    hg_excess21,
    hg_excess22
};

static const unsigned hg_excess_len[HG_MAXDEGREE+1] = {
    sizeof(hg_excess0)/sizeof(double),
    sizeof(hg_excess1)/sizeof(double),
    sizeof(hg_excess2)/sizeof(double),
    sizeof(hg_excess3)/sizeof(double),
    sizeof(hg_excess4)/sizeof(double),
    sizeof(hg_excess5)/sizeof(double),
    sizeof(hg_excess6)/sizeof(double),
    sizeof(hg_excess7)/sizeof(double),
    sizeof(hg_excess8)/sizeof(double),
    sizeof(hg_excess9)/sizeof(double),
    sizeof(hg_excess10)/sizeof(double),
    sizeof(hg_excess11)/sizeof(double),
    sizeof(hg_excess12)/sizeof(double),
    sizeof(hg_excess13)/sizeof(double),
    sizeof(hg_excess14)/sizeof(double),
    sizeof(hg_excess15)/sizeof(double),
    sizeof(hg_excess16)/sizeof(double),
    sizeof(hg_excess17)/sizeof(double),
    sizeof(hg_excess18)/sizeof(double),
    sizeof(hg_excess19)/sizeof(double),
    sizeof(hg_excess20)/sizeof(double),
    sizeof(hg_excess21)/sizeof(double),
    sizeof(hg_excess22)/sizeof(double)
};

namespace npstat {
    SbMomentsBigGamma::SbMomentsBigGamma(const unsigned maxdeg)
        : maxdeg_(maxdeg)
    {
        if (maxdeg_ < 2)
            maxdeg_ = 2;
        if (maxdeg_ > HG_MAXDEGREE)
            maxdeg_ = HG_MAXDEGREE;
        if (!coeffs_checked)
        {
            check_coeffs(hg_skewness, hg_skewness_len);
            check_coeffs(hg_excess, hg_excess_len);
            coeffs_checked = true;
        }
    }

    void SbMomentsBigGamma::setMaxDeg(const unsigned maxdeg)
    {
        maxdeg_ = maxdeg;
        if (maxdeg_ < 2)
            maxdeg_ = 2;
        if (maxdeg_ > HG_MAXDEGREE)
            maxdeg_ = HG_MAXDEGREE;
    }

    void SbMomentsBigGamma::getParameters(
        const long double gamma, const long double delta,
        long double *w, long double *q) const
    {
        assert(w);
        assert(q);
        assert(delta > 0.0L);
        assert(gamma > 0.0L);
        *w = expl(1.0L/delta/delta);
        *q = expl(-gamma/delta);
    }

    void SbMomentsBigGamma::getGammaDelta(
        const long double w, const long double q,
        long double *gamma, long double *delta) const
    {
        assert(gamma);
        assert(delta);
        assert(w > 1.0L);
        assert(q > 0.0L);
        *delta = sqrtl(1.0L/logl(w));
        *gamma = -*delta*logl(q);
    }

    unsigned SbMomentsBigGamma::bestDeg(const long double w,
                                        const long double q) const
    {
        return best_poly_degree(hg_skewness, hg_skewness_len, w, q);
    }

    bool SbMomentsBigGamma::calculate(
        const long double w, const long double q,
        long double *mean, long double *var,
        long double *skew, long double *kurt,
        long double *dskewdp0, long double *dskewdp1,
        long double *dkurtdp0, long double *dkurtdp1) const
    {
        assert(skew);
        assert(kurt);
        assert(dskewdp0);
        assert(dskewdp1);
        assert(dkurtdp0);
        assert(dkurtdp1);

        if (mean)
            *mean = calculate_mean(maxdeg_, w, q);
        if (var)
            *var = calculate_variance(maxdeg_, w, q);

        long double b1;
        calculate_series(hg_skewness, hg_skewness_len, maxdeg_, w, q,
                         &b1, dskewdp0, dskewdp1);
        calculate_series(hg_excess, hg_excess_len, maxdeg_, w, q,
                         kurt, dkurtdp0, dkurtdp1);

        // This is a "big gamma" calculator, so skew should not be 0
        if (b1 <= 0.0L)
            goto fail;
        *skew = sqrtl(b1);
        *dskewdp0 /= (2.0L * *skew);
        *dskewdp1 /= (2.0L * *skew);
        *kurt += 3.0L;

        if (mean)
            if (my_inf_nan(*mean))
                goto fail;
        if (var)
            if (my_inf_nan(*var))
                goto fail;
        if (my_inf_nan(*skew))
            goto fail;
        if (my_inf_nan(*kurt))
            goto fail;
        if (my_inf_nan(*dskewdp0))
            goto fail;
        if (my_inf_nan(*dskewdp1))
            goto fail;
        if (my_inf_nan(*dkurtdp0))
            goto fail;
        if (my_inf_nan(*dkurtdp1))
            goto fail;

        return true;

    fail:
        if (mean)
            *mean = 0.L;
        if (var)
            *var = 0.L;
        *skew = 0.L;
        *kurt = 0.L;
        *dskewdp0 = 0.L;
        *dskewdp1 = 0.L;
        *dkurtdp0 = 0.L;
        *dkurtdp1 = 0.L;
    
        return false;
    }
}
