#include <cmath>
#include <cfloat>
#include <climits>
#include <stdexcept>

#include "kstest/kstest.hh"

#include "npstat/nm/SpecialFunctions.hh"

#include "npstat/stat/UnbinnedGOFTests1D.hh"
#include "npstat/stat/StatUtils.hh"

namespace {
    double cvm_G(const double y)
    {
        return -std::exp(-y)*(npstat::besselK(0.25, y) + npstat::besselK(0.75, y));
    }

    double cvm_H(const double y)
    {
        return std::exp(-y)*(npstat::besselK(1.25, y) - 3.0*npstat::besselK(0.75, y) -
                             2.0*npstat::besselK(0.25, y));
    }

    double cvm_Ak(const unsigned k, const double x)
    {
        const double fourkp1 = 4U*k + 1U;
        const double fourkp3 = 4U*k + 3U;
        const double fourkp5 = 4U*k + 5U;
        const double sixteenx = 16.0*x;

        return 7*std::pow(fourkp1, 1.5)*cvm_G(fourkp1*fourkp1/sixteenx) +
              16*std::pow(fourkp3, 1.5)*cvm_G(fourkp3*fourkp3/sixteenx) +
               7*std::pow(fourkp5, 1.5)*cvm_G(fourkp5*fourkp5/sixteenx);
    }

    double cvm_Bk(const unsigned k, const double x)
    {
         const double fourkp1 = 4U*k + 1U;
         return std::pow(fourkp1, 2.5)*cvm_H(fourkp1*fourkp1/(16.0*x));
    }

    double cvm_Ck(const unsigned k, const double x)
    {
         const double fourkp5 = 4U*k + 5U;
         return 24*std::pow(fourkp5, 2.5)*cvm_H(fourkp5*fourkp5/(16.0*x));
    }

    double cvm_phi1_partA(const double x)
    {
        const unsigned maxIter = 1000;

        long double sum = 0.0L;
        long double gammaTerm = 0.886226925452758013649084L;
        bool converged = false;
        for (unsigned k=0; k<maxIter && !converged; ++k)
        {
            const long double delta = gammaTerm*cvm_Ak(k, x);
            sum += delta;
            gammaTerm *= (k + 1.5L);
            gammaTerm /= (k + 1.0L);
            if (k > 10U && std::abs(delta) < std::abs(sum)*DBL_EPSILON)
                converged = true;
        }
        if (!converged) throw std::runtime_error(
            "In cvm_phi1_partA: series did not converge");
        sum /= std::pow(M_PI, 1.5);
        sum /= (576*std::pow(x, 1.5));
        return sum;
    }

    double cvm_phi1_partB(const double x)
    {
        const unsigned maxIter = 1000;

        long double sum = 0.0L;
        long double gammaTerm = 1.77245385090551602729817L;
        bool converged = false;
        for (unsigned k=0; k<maxIter && !converged; ++k)
        {
            const long double delta = gammaTerm*cvm_Bk(k, x);
            sum += delta;
            gammaTerm *= (k + 0.5L);
            gammaTerm /= (k + 1.0L);
            if (k > 10U && std::abs(delta) < std::abs(sum)*DBL_EPSILON)
                converged = true;
        }
        if (!converged) throw std::runtime_error(
            "In cvm_phi1_partB: series did not converge");
        sum /= std::pow(M_PI, 1.5);
        sum /= (2304*std::pow(x, 2.5));
        return sum;
    }

    double cvm_phi1_partC(const double x)
    {
        const unsigned maxIter = 1000;

        long double sum = 0.0L;
        long double gammaTerm = 1.32934038817913702047363L;
        bool converged = false;
        for (unsigned k=0; k<maxIter && !converged; ++k)
        {
            const long double delta = gammaTerm*cvm_Ck(k, x);
            sum += delta;
            gammaTerm *= (k + 2.5L);
            gammaTerm /= (k + 1.0L);
            if (k > 10U && std::abs(delta) < std::abs(sum)*DBL_EPSILON)
                converged = true;
        }
        if (!converged) throw std::runtime_error(
            "In cvm_phi1_partC: series did not converge");
        sum /= std::pow(M_PI, 1.5);
        sum /= (2304*std::pow(x, 2.5));
        return sum;
    }

    double cvm_Vinf(const double x)
    {
        const unsigned maxIter = 1000;

        long double sum = 0.0L;
        long double gammaTerm = 1.77245385090551602729817L;
        bool converged = false;
        for (unsigned k=0; k<maxIter && !converged; ++k)
        {
            const double fourkP1 = 4*k+1;
            const double q = fourkP1*fourkP1/(16*x);
            const long double delta = gammaTerm*std::sqrt(fourkP1)*std::exp(-q)*
                                      npstat::besselK(0.25, q);
            sum += delta;
            gammaTerm *= (k + 0.5L);
            gammaTerm /= (k + 1.0L);
            if (k > 10U && std::abs(delta) < std::abs(sum)*DBL_EPSILON)
                converged = true;
        }
        if (!converged) throw std::runtime_error(
            "In cvm_Vinf: series did not converge");
        sum /= (std::sqrt(x)*pow(M_PI, 1.5));
        return sum;
    }
}

namespace npstat {
    double KSTest1D::analyticPValue(const double stat,
                                    const unsigned long sz) const
    {
        if (!sz) throw std::invalid_argument(
            "In npstat::KSTest1D::analyticPValue: sample size must be positive");
        assert(sz <= UINT_MAX);
        return kstest::KSfbar(sz, stat);
    }

    double ADTest1D::analyticPValue(const double stat,
                                    const unsigned long sz) const
    {
        if (!sz) throw std::invalid_argument(
            "In npstat::ADTest1D::analyticPValue: sample size must be positive");
        return 1.0 - AD(sz, stat);
    }

    // The algorithm for the following comes from S. Csorgo and J.J. Faraway,
    // "The Exact and Asymptotic Distributions of Cramer-von Mises Statistics",
    // Journal of the Royal Statistical Society, Series B (Methodological),
    // Vol. 58, No. 1 (1996), pp. 221-234.
    double CvMTest1D::analyticPValue(const double x,
                                     const unsigned long sz) const
    {
        if (!sz) throw std::invalid_argument(
            "In npstat::CvMTest1D::analyticPValue: sample size must be positive");
        const double vinf = cvm_Vinf(x);
        const double cdf = vinf + (vinf/12.0 + cvm_phi1_partA(x) +
                                   cvm_phi1_partB(x) + cvm_phi1_partC(x))/sz;
        return 1.0 - cdf;
    }
}
