#include <cassert>
#include <algorithm>

#include "npstat/nm/LinearMapper1d.hh"

#include "npstat/stat/scanMultivariateDensityAsWeight.hh"
#include "npstat/stat/DensityScanND.hh"

namespace npstat {
    CPP11_auto_ptr<ArrayND<double> > scanMultivariateDensityAsWeight(
        const AbsDistributionND& kernel, const unsigned* maxDim,
        const double* bandwidthSet, const double* stepSize,
        const unsigned dim)
    {
        if (!dim) throw std::invalid_argument(
            "In npstat::scanMultivariateDensityAsWeight: empty input arrays");
        if (kernel.dim() != dim)
            throw std::invalid_argument(
                "In npstat::scanMultivariateDensityAsWeight: "
                "incompatible argument dimensionalities");
        assert(maxDim);
        assert(bandwidthSet);
        assert(stepSize);

        // Make sure inputs make sense and calculate normalization
        long double ldnorm = 1.0L;
        for (unsigned idim=0; idim<dim; ++idim)
        {
            if (maxDim[idim] % 2U == 0U) throw std::invalid_argument(
                "In npstat::scanMultivariateDensityAsWeight: maximum number "
                "of steps must be odd in each dimension");
            if (!stepSize[idim]) throw std::invalid_argument(
                "In npstat::scanMultivariateDensityAsWeight: step size in each "
                "dimension must not be zero");
            if (bandwidthSet[idim] <= 0.0) throw std::invalid_argument(
                "In npstat::scanMultivariateDensityAsWeight: bandwidth "
                "must be positive in each dimension");
            ldnorm *= bandwidthSet[idim];
        }

        std::vector<LinearMapper1d> maps;
        maps.reserve(dim);
        for (unsigned idim=0; idim<dim; ++idim)
        {
            const unsigned middle = maxDim[idim]/2U;
            const double a = stepSize[idim]/bandwidthSet[idim];
            maps.push_back(LinearMapper1d(a, -a*middle));
        }

        CPP11_auto_ptr<ArrayND<double> > arr(new ArrayND<double>(maxDim, dim));
        DensityScanND<LinearMapper1d> scanner(kernel, maps, 1.0L/ldnorm);
        arr->functorFill(scanner);

        // See if we can truncate the array
        const ArrayRange& non0 = arr->nonZeroRange();
        const unsigned long lenNon0 = non0.rangeSize();
        const unsigned long arrLen = arr->length();
        assert(lenNon0);
        assert(lenNon0 <= arrLen);
        if (lenNon0 != arrLen)
        {
            bool improved = false;
            ArrayRange r(dim);
            for (unsigned idim=0; idim<dim; ++idim)
            {
                const unsigned delMin = non0[idim].min();
                const unsigned ua = arr->span(idim);
                const unsigned ub = non0[idim].max();
                assert(ua >= ub);
                const unsigned delMax = ua - ub;
                const unsigned del = std::min(delMin, delMax);
                if (del)
                    improved = true;
                r[idim] = Interval<unsigned>(del, ua-del);
            }
            if (improved)
            {
                CPP11_auto_ptr<ArrayND<double> > newarr(new ArrayND<double>(r.shape()));
                unsigned tmp[CHAR_BIT*sizeof(unsigned long)];
                r.lowerLimits(tmp, dim);
                arr->exportSubrange(tmp, dim, &*newarr);
                return newarr;
            }
        }

        return arr;
    }
}
