#ifndef NPSTAT_SCANSYMMETRICDENSITYASWEIGHT_HH_
#define NPSTAT_SCANSYMMETRICDENSITYASWEIGHT_HH_

/*!
// \file scanSymmetricDensityAsWeight.hh
//
// \brief Scan a symmetric multivariate density for subsequent use as
//        a tabulated weight
//
// Author: I. Volobouev
//
// December 2011
*/

#include <stdexcept>

#include "geners/CPP11_auto_ptr.hh"
#include "npstat/nm/ArrayND.hh"
#include "npstat/stat/AbsDistributionND.hh"

namespace npstat {
    /**
    // Utility function for scanning multivariate kernels and returning
    // ArrayND of minimal size which encloses the complete density support
    // region. It is assumed that the density is symmetric under all possible
    // mirror reflections (i.e., whenever the sign of any "x" component
    // changes, the density does not change).
    //
    // Arguments are as follows:
    //
    // kernel            -- the density to scan. Assumed to be even in
    //                      any coordinate.
    //
    // maxOctantDim      -- maximum number of steps to make in each dimension
    //                      while scanning the hyperoctant
    //
    // bandwidthSet      -- density function bandwidth values in each dimension
    //
    // stepSize          -- scan step size in each dimension
    //
    // arrayLength       -- number of elements in each of the arrays
    //                      maxOctantDim, bandwidthSet, and stepSize
    //
    // fillOneOctantOnly -- set "true" to scan one octant only (including
    //                      central grid points), "false" to scan the complete
    //                      support region of the density. In the latter case,
    //                      array size in the n-th dimension can become
    //                      as large as 2*maxOctantDim[n] - 1.
    */
    CPP11_auto_ptr<ArrayND<double> > scanSymmetricDensityAsWeight(
        const AbsDistributionND& kernel, const unsigned* maxOctantDim,
        const double* bandwidthSet, const double* stepSize,
        unsigned arrayLength, bool fillOneOctantOnly = true);

#ifdef SWIG
    inline ArrayND<double>* scanSymmetricDensityAsWeight2(
        const AbsDistributionND& kernel,
        const unsigned* maxOctantDim, unsigned len1,
        const double* bandwidthSet, unsigned len2,
        const double* stepSize, unsigned len3,
        const bool fillOneOctantOnly = true)
    {
        const unsigned dim = kernel.dim();
        if (!dim || len1 != dim || len2 != dim || len3 != dim)
            throw std::invalid_argument(
                "In npstat::scanSymmetricDensityAsWeight2: "
                "incompatible argument dimensions");
        CPP11_auto_ptr<ArrayND<double> > p = scanSymmetricDensityAsWeight(
            kernel, maxOctantDim, bandwidthSet, stepSize, dim, fillOneOctantOnly);
        return p.release();
    }
#endif // SWIG
}

#endif // NPSTAT_SCANSYMMETRICDENSITYASWEIGHT_HH_
