"""
This module provides a number of useful wrapper classes and functions
related to local regression
"""

__author__ = "Igor Volobouev (i.volobouev@ttu.edu)"
__version__ = "1.0"
__date__  = "Oct 9 2019"

import sys, npstat, math
import numpy as np

class LocalQuadraticLeastSquaresND:
   def __init__(self, polyDegree, kernel, coords, values,
                errors=None, bandwidthValues=None):
      if isinstance(kernel, npstat.AbsDistribution1D):
         weight = npstat.ProductDistributionND(kernel)
      else:
         weight = kernel
      wdim = weight.dim()
      if wdim == 1:
         if type(coords).__name__ == 'Point1DVector':
            self._coords = coords
         else:
            self._coords = npstat.makePointsFromFlat1D(coords, 1, (0,))
      else:
         self._coords = coords
      self._values = np.array(values, np.double)
      if errors is None:
         self._errors = np.array((), np.double)
      else:
         self._errors = np.array(errors, np.double)
      cl = npstat.__dict__["LocalQuadraticLeastSquares{:d}D".format(wdim)]
      if bandwidthValues is None:
         bandwidthValues = [1.0]*wdim
      self._fitter = cl(polyDegree, weight, bandwidthValues,
                        self._coords, self._values, self._errors)

   def __call__(self, *args):
      assert len(args) == self._fitter.dim(), \
         "Incompatible number of input coordinates"
      return self._fitter.fit(args)

   def dim(self):
      return self._fitter.dim()

   def polyDegree(self):
      return self._fitter.polyDegree()

   def nPoints(self):
      return self._fitter.nPoints()

   def nTermsFitted(self):
      return self._fitter.nTermsFitted()

   def hasErrors(self):
      return self._fitter.hasErrors()

   def chiSquare(self):
      return self._fitter.chiSquare()

   def lscvCriterion(self):
      return self._fitter.lscvCriterion()

   def getBandwidth(self, n):
      return self._fitter.getBandwidth(n)

   def setBandwidth(self, n, bw):
      self._fitter.setBandwidth(n, bw)

   def setBandwidthValues(self, bw):
      assert self._fitter.dim() == len(bw), \
         "Incompatible number of bandwidth values"
      for dim, h in enumerate(bw):
         self._fitter.setBandwidth(dim, h)      

   def setPointCoords(self, coords):
      wdim = self._fitter.dim()
      if wdim == 1:
         if type(coords).__name__ == 'Point1DVector':
            self._coords = coords
         else:
            self._coords = npstat.makePointsFromFlat1D(coords, 1, (0,))
      else:
         self._coords = coords
      self._fitter.setPointCoords(self._coords)

   def setValues(self, values):
      self._values = np.array(values, np.double)
      self._fitter.setValues(self._values)

   def setErrors(self, errors):
      if errors is None:
         self._errors = np.array((), np.double)
      else:
         self._errors = np.array(errors, np.double)
      self._fitter.setErrors(self._errors)

   def fit(self, coords):
      return self._fitter.fit(coords)

   def fitWithGradient(self, coords):
      wdim = self._fitter.dim()
      grad = np.empty((wdim,), np.double)
      value = self._fitter.fitWithGradient(coords, grad)
      return value, grad

   def fitWithHessian(self, coords):
      wdim = self._fitter.dim()
      grad = np.empty((wdim,), np.double)
      hess = npstat.DoubleMatrix(wdim, wdim)
      value = self._fitter.fitWithHessian(coords, grad, hess)
      return value, grad, hess

   def _selectBestBandwidthFactor(self, bwref, bwfactors, criterionCalculator):
      if bwref is None:
         bwref = [1.0]*self._fitter.dim()
      assert len(bwfactors) > 0, "The set of bandwidth factors is empty"
      cvBest = sys.float_info.max
      bestSet = None
      for bw in bwfactors:
         bwset = [bw*ref for ref in bwref]
         self.setBandwidthValues(bwset)
         cv = criterionCalculator(self)
         if cv < cvBest:
            cvBest = cv
            bestSet = bwset
      self.setBandwidthValues(bestSet)
      return bestSet

   def selectLSCVBandwidth(self, bwset, bwref=None):
      return self._selectBestBandwidthFactor(
         bwref, bwset, LocalQuadraticLeastSquaresND.lscvCriterion)

   def fitNDoF(self, entropic=True):
      eR, tR = self._fitter.smootherMatrix().timesT().symPSDefEffectiveRank()
      if entropic:
         return eR
      else:
         return tR

   def AIC(self, entropic=True):
      nDoF = self.fitNDoF(entropic)
      chisq = self.chiSquare()
      if self._fitter.hasErrors():
         # The chi-square is the "normal" chi-square
         logli = -chisq/2.0
      else:
         # The chi-square is the residual sum of squares.
         # We can think that we are estimating the homoscedastic
         # error in addition to the regression curve itself.
         nDoF = nDoF + 1
         logli = -0.5*self.nPoints()*math.log(chisq/self.nPoints())
      return npstat.aicc(nDoF, logli, self.nPoints())

   def _EAIC(self):
      return self.AIC(True)

   def _TAIC(self):
      return self.AIC(False)

   def selectEAICBandwidth(self, bwset, bwref=None):
     return self._selectBestBandwidthFactor(
         bwref, bwset, LocalQuadraticLeastSquaresND._EAIC)

   def selectTAICBandwidth(self, bwset, bwref=None):
     return self._selectBestBandwidthFactor(
         bwref, bwset, LocalQuadraticLeastSquaresND._TAIC)
